/*
** 
**  Copyright (c) 1998 by Larry Plona
** 
**  This distribution  of XPipeman is copyright by Larry Plona.
**  The  package  and  its  component  files  may  be  freely
**  distributed as long as  all copyright notices remain intact
**  and this distribution is not sold  as  commercial software.
** 
**  No warranty of any kind is attached with this software.  It
**  is supplied as is.
** 
** ----------------------------------------------------------------
*/

/*
** C++ Class Colores. See "colores.h" for description.
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "colores.h"

extern int errno;
extern char *sys_errlist[];

#define TO16(x) ((65535*(x))/255)

Colores::Colores( Display *dpy, Colormap cmap, XVisualInfo *vinfo )
{
    count = 0;         /* Entries in colormap. */
    colors = NULL;     /* Array of X11 colors. */
    display = dpy;     /* X11 display.         */
    x_colormap = cmap; /* X11 colormap.        */
    this->vinfo = vinfo;
}

/*
** loadStandard()
**
** Initialize colormap with some hardcoded defaults.
**
** FUTURE. These defaults ought to be application specified.
*/
int Colores::loadStandard( void )
{
    if( vinfo->c_class == PseudoColor )
    {
	if( !standard() )
	    return 0;
	return flush();
    }
    return 1;
}

/*
** load()
**
** Given an Xpm buffer, use its colormap as the new map for the object.
** The standard colors remain in their positions. The buffer's colormap
** is appended to the standard colors.
*/
int Colores::load( const char *buf )
{
    char *colorname;
    int rc;
    int i;
    int start;
    XColor color;
    XpmImage xpmimage;

    if( vinfo->c_class != PseudoColor )
	return 1;

    /*
    ** Initial pixels in all colormaps are the same.
    */
    if( (start=standard()) == 0 )
	return 0;

    /*
    ** Get the Xpm colormap from the Xpm buffer. The actual image
    ** is ignored.
    */
    rc = XpmCreateXpmImageFromBuffer((char*)buf,&xpmimage,NULL);
    if( rc != XpmSuccess )
    {
	fprintf( stdout, "Error. Bad colormap.xpm\n" );
	fprintf( stdout, "XpmCreateImageFromBuffer. %s.\n",
	    XpmGetErrorString(rc) );
	return 0;
    }

    /*
    ** FUTURE. Is there a reason not to allocate 256 up front?
    */
    /*
    if( !setSize(xpmimage.ncolors+start) )
	return 0;
    */
    setSize( 256 );

    for( i=0; i<xpmimage.ncolors; ++i )
    {
	/*
	** Use the symbolic color name if it exists, otherwise use
	** the normal color name. Symbolic names are usually something
	** like "foregroundShadowColor" while regular color names
	** are like "red" or "#CB2B00FF3164".
	*/
	colorname = xpmimage.colorTable[i].symbolic;
	if( xpmimage.colorTable[i].symbolic == NULL )
	    colorname = xpmimage.colorTable[i].c_color;
	if( xpmimage.colorTable[i].c_color == NULL )
	{
	    fprintf( stdout, "Error. Bad color in colormap.\n" );
	    fprintf( stdout, "       XPM index is \"%s\".\n",
		xpmimage.colorTable[i].string );
	    return 0;
	}
	/*
	** Get an XColor record for this string. The color has to
	** be recognized by the X11 server or there is an error.
	** X11 servers either get their colors from a file named
	** "X11/lib/rgb.txt" or from a color server. The X11
	** record contains the red, green and blue components of
	** the color.
	*/
	if( !XParseColor(display,x_colormap,colorname,&color) )
	{
	    fprintf( stdout, "ERROR. Bad XParseColor, \"%s\".\n",
		colorname );
	    return 0;
	}
	/*
	fprintf( stdout, "TMP load \"%s\" %x,%x,%x\n",
	    colorname, color.red, color.green, color.blue );
	*/
	colors[start].flags = DoRed|DoGreen|DoBlue;
	colors[start].red = color.red;
	colors[start].green = color.green;
	colors[start].blue = color.blue;
	++start;
    }

    XpmFreeXpmImage( &xpmimage );

    /* Add the new colors to the X11 colormap for the object. */
    return flush();
}

/*
** CreatePixmapFromData
**
** Essentially just a wrapper around the routine from the Xpm library.
** However, this function will supply the Colores color allocator
** mapping the colors of the data to the current colormap.
**
** The attributes field may be null. If it is not NULL, it is the
** applications responsibility to call XpmFreeAttributes.
**
** The pix argument may be NULL if only the mask or attributes are
** desired.
*/
int Colores::createPixmapFromData( const char **data, Pixmap *pix,
   Pixmap *mask, XpmAttributes *attr )
{
    int rc;
    Pixmap dummyPix;
    XpmAttributes nattr;

    /* For their convenience, allow caller to not supply attributes. */
    nattr.valuemask = 0;
    if( attr == NULL )
	attr = &nattr;

    attr->valuemask |= XpmColormap;
    attr->colormap = x_colormap;

    if( vinfo->c_class == PseudoColor )
    {
	attr->valuemask |= XpmAllocColor|XpmColorClosure;
	attr->alloc_color = xpm_alloc;
	attr->color_closure = this;
    }

    /* A NULL pixmap crashes some versions of Xpm. */
    if( pix == NULL )
	pix = &dummyPix;

    rc = XpmCreatePixmapFromData(display,DefaultRootWindow(display),
	(char**)data,pix,mask,attr);

    return rc;
}

/*
** See createPixmapFromData() for comments.
*/
int Colores::createImageFromData( const char **data, XImage **img,
   XImage **mask, XpmAttributes *attr )
{
    int rc;
    XpmAttributes nattr;

    /* For their convenience, allow caller to not supply attributes. */
    nattr.valuemask = 0;
    if( attr == NULL )
	attr = &nattr;

    attr->valuemask |= XpmColormap;
    attr->colormap = x_colormap;

    if( vinfo->c_class == PseudoColor )
    {
	attr->valuemask |= XpmAllocColor|XpmColorClosure;
	attr->alloc_color = xpm_alloc;
	attr->color_closure = (void*)this;
    }

    rc = XpmCreateImageFromData(display,(char**)data,img,mask,attr);

    if( nattr.valuemask == 0 )
	XpmFreeAttributes( &nattr );

    return rc;
}

/*
** See createPixmapFromData() for comments.
*/
int Colores::createPixmapFromBuffer( const char *buf, Pixmap *pix,
   Pixmap *mask, XpmAttributes *attr )
{
    int rc;
    Pixmap dummyPix;
    XpmAttributes nattr;

    /* For their convenience, allow caller to not supply attributes. */
    nattr.valuemask = 0;
    if( attr == NULL )
	attr = &nattr;

    attr->valuemask |= XpmColormap;
    attr->colormap = x_colormap;

    if( vinfo->c_class == PseudoColor )
    {
	attr->valuemask |= XpmAllocColor|XpmColorClosure;
	attr->alloc_color = xpm_alloc;
	attr->color_closure = (void*)this;
    }

    /* A NULL pixmap crashes some versions of Xpm. */
    if( pix == NULL )
	pix = &dummyPix;

    rc = XpmCreatePixmapFromBuffer(display,DefaultRootWindow(display),
	(char*)buf,pix,mask,attr);

    if( nattr.valuemask == 0 )
	XpmFreeAttributes( &nattr );

    return rc;
}

/*
** See createPixmapFromData() for comments.
*/
int Colores::createImageFromBuffer( const char *buf, XImage **img,
   XImage **mask, XpmAttributes *attr )
{
    int rc;
    XpmAttributes nattr;

    /* For their convenience, allow caller to not supply attributes. */
    nattr.valuemask = 0;
    if( attr == NULL )
	attr = &nattr;

    attr->valuemask |= XpmColormap;
    attr->colormap = x_colormap;

    if( vinfo->c_class == PseudoColor )
    {
	attr->valuemask |= XpmAllocColor|XpmColorClosure;
	attr->alloc_color = xpm_alloc;
	attr->color_closure = (void*)this;
    }

    rc = XpmCreateImageFromBuffer(display,(char*)buf,img,mask,attr);

    if( nattr.valuemask == 0 )
	XpmFreeAttributes( &nattr );

    return rc;
}

/*
** See createPixmapFromData() for comments.
*/
int Colores::readFileToImage( const char *path, XImage **img,
   XImage **mask, XpmAttributes *attr )
{
    int rc;
    XpmAttributes nattr;

    /* For their convenience, allow caller to not supply attributes. */
    nattr.valuemask = 0;
    if( attr == NULL )
	attr = &nattr;

    attr->valuemask |= XpmColormap;
    attr->colormap = x_colormap;

    if( vinfo->c_class == PseudoColor )
    {
	attr->valuemask |= XpmAllocColor|XpmColorClosure;
	attr->alloc_color = xpm_alloc;
	attr->color_closure = (void*)this;
    }

    rc = XpmReadFileToImage(display,(char*)path,img,mask,attr);

    if( nattr.valuemask == 0 )
	XpmFreeAttributes( &nattr );

    return rc;
}

/*
** flush()
**
** Store the current array of X11 color records in the X11 colormap
** for the object. The display will not reflect colormap changes
** until this routine is called.
*/
int Colores::flush( void )
{
    if( vinfo->c_class != PseudoColor )
	return 1;

    //fputs( "TMP colores flush\n", stdout );
    if( XStoreColors(display,x_colormap,colors,count) != 0 )
    {
	/*
	** The X server supplied with UnixWare 2 returns the number of
	** colors stored instead of an X error code. To avoid ifdefs,
	** ignore the return code.
	*/
#if 0
	fprintf( stdout, "ERROR. XStoreColors, %d.\n", rc );
	return 0;
#endif
    }
    return 1;
}

/*
** setSize()
**
** Specify the number of X11 color records to maintain for this object.
** This is almost always 256.
**
** Pseudo color only.
*/
int Colores::setSize( int size )
{
    int i;

    if( vinfo->c_class != PseudoColor )
	return 1;

    if( count > size )
	return 1;

    if( colors == NULL )
    {
	if( (colors=(XColor*)malloc(sizeof(XColor)*size)) == NULL )
	{
	    fprintf( stdout, "ERROR. Bad malloc. %s.\n",
		sys_errlist[errno] );
	    return 0;
	}
	count = size;
	for( i=0; i<size; ++i )
	    colors[i].pixel = i;
	return 1;
    }

    if( (colors=(XColor*)realloc(colors,sizeof(XColor)*size))
	== NULL )
    {
	fprintf( stdout, "ERROR. Bad realloc. %s.\n",
	    sys_errlist[errno] );
	return 0;
    }
    for( i=0; i<size; ++i )
	colors[i].pixel = i;
    count = size;

    return 1;
}

/*
** standard()
**
** Fills colormap with the "standard" colors used by all colormaps
** which may be loaded over the time of the object. The first 4
** entries in the colormap are arbitrarily reserved and filled with
** hardcoded values.
**
** NOTE. These colors are not "standard" by any definition but my own.
**       This member is just a hack for my own convenience.
*/
int Colores::standard( void )
{
    if( vinfo->c_class != PseudoColor )
	return 0;

    if( !setSize(4) )
	return 0;

    /* FUTURE. Pixels values harcoded by xface (XFACE_PIXEL_*). */
    colors[0].flags = DoRed|DoGreen|DoBlue;
    colors[0].pixel = 0;
    colors[0].red = 0;
    colors[0].green = 0;
    colors[0].blue = 0;
    colors[1].flags = DoRed|DoGreen|DoBlue;
    colors[1].pixel = 1;
    colors[1].red = 0xffff;
    colors[1].green = 0xffff;
    colors[1].blue = 0xffff;
    colors[2].flags = DoRed|DoGreen|DoBlue;
    colors[2].pixel = 2;
    colors[2].red = 0xcb2b;
    colors[2].green = 0xcb2b;
    colors[2].blue = 0xcb2b;
    colors[3].flags = DoRed|DoGreen|DoBlue;
    colors[3].pixel = 3;
    colors[3].red = 0xab2b;
    colors[3].green = 0xab2b;
    colors[3].blue = 0xab2b;

    return 4; // Number of standard colors.
}

/*
** Colores::xpm_alloc()
**
** Used as color allocation function by Xpm routines. This is a static
** member, a Colores object must be passed in as client data.
**
** Pseudo color only.
**
** This routine will always return the nearest color.
**
** The createPixmapFromBuffer family of member functions makes direct
** use of xpm_alloc uncommon. To use xpm_alloc as an Xpm color allocated,
** an Xpm attributes record must be passed to the Xpm routine with
** the following values.
**
**    attr->valuemask |= XpmColormap|XpmAllocColor|XpmColorClosure;
**    attr->colormap = colores.x_colormap;
**    attr->alloc_color = Colores::xpm_alloc;
**    attr->color_closure = (void*)colores;
*/
int Colores::xpm_alloc( Display *dpy, Colormap cmap, char *name,
    XColor *target, void *data )
{
    int i;
    XColor *p;
    Colores *c = (Colores*)data;

    if( name != NULL )
    {
	target->pixel = 0;
	target->red = target->green = target->blue = 0;
	if( !XParseColor(dpy,cmap,name,target) )
	{
	    fprintf( stdout, "ERROR. Bad XParseColor on \"%s\".\n", name );
	    return -1;
	}
	/*
	fprintf( stdout, "TMP Xpm \"%s\" %x,%x,%x\n",
	    name, target->red, target->green, target->blue );
	*/
    }

    for( i=0; i<c->count; ++i )
    {
	/*
	fprintf( stdout, "TMP Xpm testing against %x,%x,%x %d entries\n",
	    p->red, p->green, p->blue, c->count );
	*/
	p = &(c->colors[i]);
	if( (target->red==p->red) &&
	    (target->green==p->green) &&
	    (target->blue==p->blue) )
	{
	    target->pixel = i;
	    return 1;
	}
    }

    target->pixel = c->nearest( target->red, target->green, target->blue );

    /*
    fprintf( stdout, "TMP nearest \"%s\" %d,%d,%d -> %d,%d,%d\n",
	name, target->red, target->green, target->blue,
	c->colors[target->pixel].red,
	c->colors[target->pixel].green,
	c->colors[target->pixel].blue );
    */

    /*
    if( name != NULL )
    {
	fprintf( stdout, "ERROR. Color \"%s\" is not in colormap.\n", name );
    }
    else
    {
	fprintf( stdout, "ERROR. Color 0x%x,0x%x,0x%x not in colormap.\n",
	    target->red, target->green, target->blue );
    }
    */

    return 1;
}

/*
** Colores::nearest
**
** Find the nearest color in the current colormap to given RGB values.
**   (r-r1)^2 + (g-g1)^2 + (b-b1)^2
**
** Pseudo color only.
*/
Pixel Colores::nearest( unsigned int tR, unsigned int tG, unsigned int tB )
{
    unsigned long nearest = (unsigned long)1e9;
    int pixel = 0;
    long dist;
    int i;

    //fprintf( stdout, "TMP Need nearest to %x %x %x.\n", tR, tG, tB );
    for( i=0; i<256; ++i )
    {
        dist =
	    ((colors[i].red-tR)*(colors[i].red-tR)) +
	    ((colors[i].green-tG)*(colors[i].green-tG)) +
            ((colors[i].blue-tB)*(colors[i].blue-tB));
	/*
	fprintf( stdout, "TMP    %06d :  %04x %04x %04x.\n",
	    dist, colors[i].red, colors[i].green, colors[i].blue );
	*/
        if( dist < nearest )
        {
	    nearest = dist;
	    pixel = (Pixel)i;
	}
    }
    return pixel;
}

/*
** Effect routines.
*/

/*
** Ramp effect adapted from code appearing in Gems II by Ken Musgrave.
** Copyright 1985.
*/

#define ERANDOM(x) ((25173*(x)+13849)&65535)

/*
** effectRampInit()
**
** Begin a palette shift. A study (settings) record is passed in to
** specify the nature of the palette shift.
**
** Call effectRamp() to perform one shift.
**
** A state record is returned. It must be given as an argument each
** time a shift is made.
*/
struct colores_effect_ramp *Colores::effectRampInit(
    struct colores_study_ramp *study )
{
    int i;
    struct colores_effect_ramp *rec;

    if( (rec=(struct colores_effect_ramp*)malloc(
	sizeof(struct colores_effect_ramp))) == NULL )
    {
	fprintf( stdout, "ERROR. Bad malloc. %s.\n", sys_errlist[errno] );
	return NULL;
    }
    rec->r.y2 = 0.0;
    rec->g.y2 = 0.0;
    rec->b.y2 = 0.0;
    rec->r.xsteps = 0.0;
    rec->g.xsteps = 0.0;
    rec->b.xsteps = 0.0;
    rec->r.xcount = 0;
    rec->g.xcount = 0;
    rec->b.xcount = 0;
    rec->study = *study;

    if( rec->study.maxsteps < 1 )
	rec->study.maxsteps = 10;

    if( rec->study.flagBegin == COLORES_RAMP_BLACK )
    {
	for( i=rec->study.pixelBegin; i<=rec->study.pixelEnd; ++i )
	{
	    /* All pixels start at black. */
	    colors[i].flags = DoRed|DoGreen|DoBlue;
	    colors[i].pixel = i;
	    colors[i].red = 0;
	    colors[i].green = 0;
	    colors[i].blue = 0;
	}
    }
    if( rec->study.flagBegin == COLORES_RAMP_GRAY )
    {
	for( i=rec->study.pixelBegin; i<=rec->study.pixelEnd; ++i )
	{
	    /* Grayscale start. */
	    colors[i].flags = DoRed|DoGreen|DoBlue;
	    colors[i].pixel = i;
	    colors[i].red = i;
	    colors[i].green = i;
	    colors[i].blue = i;
	}
    }
    if( rec->study.flagBegin == COLORES_RAMP_ZEBRA )
    {
	// FUTURE. The plus four method may cross the end pixel.
	for( i=rec->study.pixelBegin; i<=rec->study.pixelEnd; i+=4 )
	{
	    /* Zebra start. */
	    colors[i].flags = DoRed|DoGreen|DoBlue;
	    colors[i].pixel = i;
	    colors[i+1].flags = DoRed|DoGreen|DoBlue;
	    colors[i+1].pixel = i+1;
	    colors[i+2].flags = DoRed|DoGreen|DoBlue;
	    colors[i+2].pixel = i+2;
	    colors[i+3].flags = DoRed|DoGreen|DoBlue;
	    colors[i+3].pixel = i+3;
	    colors[i].red = colors[i+1].red = colors[i].green =
		colors[i+1].green = colors[i].blue = colors[i+1].blue =
		(i*4/5);
	    colors[i+2].red = colors[i+3].red = colors[i+2].green =
		colors[i+3].green = colors[i+2].blue = colors[i+3].blue =
		(i*4/5+51);
	}
    }
    /*
    ** For COLORES_RAMP_KEEP, do nothing.
    */
    flush();

    return rec;
}

int Colores::effectRamp( struct colores_effect_ramp *rec )
{
    int i;
    XColor hold;
    unsigned int pixelBegin = (unsigned int)rec->study.pixelBegin;
    unsigned int pixelEnd = (unsigned int)rec->study.pixelEnd;

    /*
    ** Shift the pallette down. Pixels underneath the starting pixel
    ** are preserved.
    */
    hold = colors[pixelBegin];
    for( i=pixelBegin; i<pixelEnd-1; ++i )
    {
	colors[i].pixel = i;
	colors[i].flags = DoRed|DoGreen|DoBlue;
	colors[i].red = colors[i+1].red;
	colors[i].green = colors[i+1].green;
	colors[i].blue = colors[i+1].blue;
    }

    /*
    ** Fill in the empty pixel created by the shift.
    */
    if( rec->study.flagRamp == COLORES_RAMP_STATIC )
    {
	colors[pixelEnd-1] = hold;
	colors[pixelEnd-1].pixel = pixelEnd-1;
	colors[pixelEnd-1].flags = DoRed|DoGreen|DoBlue;
    }
    else
    {
	colors[pixelEnd-1].pixel = pixelEnd-1;
	colors[pixelEnd-1].flags = DoRed|DoGreen|DoBlue;
	colors[pixelEnd-1].red = ramp_component( rec, &rec->r );
	if( rec->study.flagRamp == COLORES_RAMP_GRAYS )
	{
	    colors[pixelEnd-1].green = colors[pixelEnd-1].red;
	    colors[pixelEnd-1].blue = colors[pixelEnd-1].red;
	}
	else
	{
	    /* COLORES_RAMP_FULL. */
	    colors[pixelEnd-1].green = ramp_component( rec, &rec->r );
	    colors[pixelEnd-1].blue = ramp_component( rec, &rec->r );
	}
    }
    /*
    fprintf( stdout, "TMP cl effect new %x,%x,%x 0-%x,%x,%x\n",
	colors[count-1].red,
	colors[count-1].green,
	colors[count-1].blue,
	colors[5].red,
	colors[5].green,
	colors[5].blue );
    */

    flush();

    return 1;
}

/*
** ramp_component()
**
** Used internally by the ramp effect. Randomly select a value
** for the component (r, g, b).
*/
int Colores::ramp_component( struct colores_effect_ramp *ramp,
    struct colores_effect_ramp_component *rec )
{
    int temp;

    if( rec->xcount >= (int)rec->xsteps )
    {
	rec->y1 = rec->y2;
	ramp->study.seed = ERANDOM(ramp->study.seed);
	rec->y2 = (float)255 * (ramp->study.seed/65535.0);
	ramp->study.seed = ERANDOM(ramp->study.seed);
	rec->xsteps = (ramp->study.maxsteps*(ramp->study.seed/65535.0));
	if( rec->xsteps != 0 )
	    rec->inc = (rec->y2-rec->y1)/rec->xsteps;
	else
	    rec->inc = 0;
	rec->xcount = 0;
    }
    temp = (int)rec->y1;
    rec->y1 += rec->inc;
    rec->xcount++;
    return TO16(temp);
}
