#ifndef COLORES_H
#define COLORES_H 1

/*
** 
**  Copyright (c) 1998 by Larry Plona
** 
**  This distribution  of XPipeman is copyright by Larry Plona.
**  The  package  and  its  component  files  may  be  freely
**  distributed as long as  all copyright notices remain intact
**  and this distribution is not sold  as  commercial software.
** 
**  No warranty of any kind is attached with this software.  It
**  is supplied as is.
** 
** ----------------------------------------------------------------
*/

/*
** C++ class Colores.
**
** The Colores class manages an X11 colormap, adding the ability to
** load maps from XPM format file, find color matches and providing
** some colormap effects such as palette shifting.
**
** An Xpm file is used as a colormap (its image is irrelevant). Other
** Xpm files (or Xpm data or Xpm buffers) may be used to create pixmaps.
** Those pixmaps will use the colormap of the Colores object. Colors
** not in the colormap will be approximated.
**
** An alarm driven palette shift may be performed on the colormap.
** Several settings are available for controlling the effect.
**
** This is class is independant of XPipeman. It requires only Xlib.
**
** Typical use.
**
** {
**     void *effect;
**     Colores *c;
**     struct colores_study_ramp study;
**
**     c = new Colores( my_display, my_x_colormap, my_x_visual );
**     c->load( buf );   ** buf is Xpm an buffer. See Xpm package
**
**     ** Create a pixmap from an included xpm file. **
**
**     c->createPixmapFromData(pixels_xpm,&pix);
**
**     ** Set up a palette shift. **
**
**     study.pixelBegin = 6;
**     study.pixelEnd = 12;
**     study.flagBegin = COLORES_RAMP_KEEP;
**     study.flagRamp = COLORES_RAMP_STATIC;
**     effect = c->effectRampInit(&study);
**
**    ** set up an X or C++ timeout which will call the following. **
**    c->effectRamp( effect );
** }
*/

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/xpm.h>

/*
** flagBegin setting for the Ramp effect.
**
** BLACK - Start colormap as all black.
** GRAY  - Start colormap as ordered shades of gray.
** ZEBRA - Start colormap with color stripes.
** KEEP  - Keep colormap as it is.
*/
#define COLORES_RAMP_BLACK  0
#define COLORES_RAMP_GRAY   1
#define COLORES_RAMP_ZEBRA  2
#define COLORES_RAMP_KEEP   3

/*
** flagRamp setting of the ramp effect. Specifies what to do with the
** empty color entry created by the shift.
**
** STATIC - Fill with the entry pushed out by the shift.
** FULL   - Randomly select new r, g and b components.
** GRAYS  - Randomly select a gray.
*/
#define COLORES_RAMP_FULL   0
#define COLORES_RAMP_GRAYS  1
#define COLORES_RAMP_STATIC 2

/* Study (settings) for the ramp effect. */
struct colores_study_ramp
{
    unsigned char flagBegin; //COLORES_RAMP_*
    unsigned char flagRamp;  //COLORES_RAMP_*
    unsigned char maxsteps;
    int pixelBegin;  // Shift between these two pixels.
    int pixelEnd;
    int seed;
};

/* State information for an instance of the ramp effect. */
struct colores_effect_ramp_component
{
    float y1;
    float y2;
    float inc;
    float xsteps;
    int xcount;
};

/* State information for an instance of the ramp effect. */
struct colores_effect_ramp
{
    struct colores_study_ramp study;
    struct colores_effect_ramp_component r;
    struct colores_effect_ramp_component g;
    struct colores_effect_ramp_component b;
};

class Colores
{
    Display *display;
    int standard( void );
    int flush( void );
    XColor *colors;
    int count;
    int ramp_component( struct colores_effect_ramp*,
       struct colores_effect_ramp_component* );
public:
    static int xpm_alloc( Display*, Colormap, char*, XColor*, void* );
    XVisualInfo *vinfo;
    Colormap x_colormap;
    Colores( Display*, Colormap, XVisualInfo* );
    int load( const char *buf );
    int loadStandard( void );
    int setSize( int );
    Pixel nearest( unsigned int, unsigned int, unsigned int );
    /* Effects. */
    struct colores_effect_ramp *effectRampInit( struct colores_study_ramp* );
    int effectRamp( struct colores_effect_ramp* );
    /* Xpm Routines. */
    int createPixmapFromData( const char**, Pixmap*, Pixmap*,
	XpmAttributes* );
    int createPixmapFromBuffer( const char*, Pixmap*, Pixmap*,
	XpmAttributes* );
    int createImageFromData( const char**, XImage**, XImage**,
	XpmAttributes* );
    int createImageFromBuffer( const char*, XImage**, XImage**,
	XpmAttributes* );
    int readFileToImage( const char*, XImage**, XImage**, XpmAttributes* );
};

#endif
