/*
** 
**  Copyright (c) 1998 by Larry Plona
** 
**  This distribution  of XPipeman is copyright by Larry Plona.
**  The  package  and  its  component  files  may  be  freely
**  distributed as long as  all copyright notices remain intact
**  and this distribution is not sold  as  commercial software.
** 
**  No warranty of any kind is attached with this software.  It
**  is supplied as is.
** 
** ----------------------------------------------------------------
*/

/*
** A few notes.
**
** Byte ordering is big-endian in the glob header and indices. After
** an open, the count and version fields of the glob handle are in
** host order (the indices in the handle remain big-endian).
*/

#include<sys/stat.h>
#include<netinet/in.h>
#include<stdio.h>
#include<unistd.h>
#include<stdlib.h>
#include<errno.h>
#include<fcntl.h>
#include"glob.h"

/* Not always in errno.h. */
extern char *sys_errlist[];

/*
** Glob constructor for existing glob file.
*/
Glob::Glob( const char *globfile )
{
    int rc;
    int len;
    struct globhdr hdr;

    fd = -1;
    idxs = NULL;
    if( (fd=open(globfile,O_RDONLY)) == -1 )
	return;

    if( (rc=read(fd,&hdr,sizeof(hdr))) == -1 )
    {
	fprintf( stdout, "ERROR. Bad read of \"%s\"\". %s\n",
	    globfile, sys_errlist[errno] );
	close( fd );
	fd = -1;
	return;
    }
    if( rc < sizeof(hdr) )
    {
	fputs( "ERROR. Short glob header.\n", stdout );
	close( fd );
	fd = -1;
	return;
    }

    count = ntohs(hdr.version);
    //fprintf( stdout, "Glob Version %d.\n", count );
    count = ntohl(hdr.count);
    //fprintf( stdout, "Component count is %d.\n", count );

    len = sizeof(struct globidx)*count;

    if( (idxs=(struct globidx*)malloc((size_t)len)) == NULL )
    {
	fputs( "ERROR. Bad malloc.\n", stdout );
	close( fd );
	fd = -1;
	return;
    }

    if( (rc=read(fd,idxs,len)) == -1 )
    {
	fprintf( stdout, "ERROR. Bad read of \"%s\"\". %s\n",
	    globfile, sys_errlist[errno] );
	close( fd );
	free( (void*)idxs );
	fd = -1;
	return;
    }

#if 0
    {
	int i;
	char component[GLOB_MAX_NAME+1];
	char type[GLOB_MAX_TYPE+1];

	for( i=0; i<count; ++i )
	{
	    memcpy( component, gd->idxs[i].name, GLOB_MAX_NAME );
	    component[GLOB_MAX_NAME] = '\0';
	    /*fprintf( stdout, "-- %d \"%s\"\n", i, component );*/
	    rc = ntohl(gd->idxs[i].begin);
	    /*fprintf( stdout, "   Start: %d\n", rc );*/
	    rc = ntohl(gd->idxs[i].length);
	    /*fprintf( stdout, "   Length: %d\n", rc );*/
	}
    }
#endif
}

Glob::~Glob( void )
{
    if( fd == -1 )
	return;
    if( idxs != NULL )
	free( (void*)idxs  );
    close( fd );
}

int Glob::initialized( void )
{
    return (fd!=-1) ? 1 : 0;
}

/*
** Glob::dump()
**
** Debug routine. Dump contents of a single Glob index record.
*/
void Glob::dump( int idx )
{
    char name[GLOB_MAX_NAME+1];
    char type[GLOB_MAX_TYPE+1];

    memcpy( name, idxs[idx].name, GLOB_MAX_NAME );
    name[GLOB_MAX_NAME] = '\0';

    memcpy( type, idxs[idx].type, GLOB_MAX_TYPE );
    type[GLOB_MAX_NAME] = '\0';

    fprintf( stdout, "Component: \"%s\" (%s). Start %d. Length %d.\n",
	name, type, ntohl(idxs[idx].begin), ntohl(idxs[idx].length) );
}

/*
** Glob::extract()
**
** Return a pointer to a block of memory containing component data
** from an open glob. The desired component's name is passed in.
** If the address of an int is also given, the length of the component
** will be placed into it.
*/
char *Glob::extract( const char *component, int *len )
{
    char *buf;
    int i;

    if( !initialized() )
    {
	//fputs( "TMP glob is not valid\n", stdout );
	return NULL;
    }
    if( idxs == NULL )
    {
	//fputs( "TMP no idxs\n", stdout );
	return NULL;
    }
    if( count < 1 )
    {
	//fputs( "TMP empty glob\n", stdout );
	return NULL;
    }

    for( i=0; i<count; ++i )
    {
	if( !strncmp(idxs[i].name,component,GLOB_MAX_NAME) )
	    break;
    }
    if( i >= count )
    {
	return NULL;
    }

    if( lseek(fd,ntohl(idxs[i].begin),SEEK_SET) == -1 )
    {
	/*
	** When running UW7 binaries on UW2 (via UDK), lseek fails
	** with EBADF.
	*/
/*
	fprintf( stdout, "ERROR. Bad lseek to byte %d (fd%d). %s.\n",
	    ntohl(idxs[i].begin), fd, sys_errlist[errno] );
	return NULL;
*/
    }

    i = ntohl(idxs[i].length);
    if( len != NULL )
	*len = i;
    if( (buf=(char*)malloc((size_t)i)) == NULL )
	return NULL;

    if( read(fd,buf,i) == -1 )
    {
	fprintf( stdout, "ERROR. Bad read. %s\n",
	    sys_errlist[errno] );
	return NULL;
    }

    return buf;
}
