/*
** 
**  Copyright (c) 1998 by Larry Plona
** 
**  This distribution  of XPipeman is copyright by Larry Plona.
**  The  package  and  its  component  files  may  be  freely
**  distributed as long as  all copyright notices remain intact
**  and this distribution is not sold  as  commercial software.
** 
**  No warranty of any kind is attached with this software.  It
**  is supplied as is.
** 
** ----------------------------------------------------------------
*/

/*
** globber.
**
** globber <indir> <glob> < <component-list>
**
** Add files listed in <component-list> (one file per line) into the
** named glob. Files will have <indir> preappened to them.
*/
#include<stdio.h>
#include<unistd.h>
#include<stdlib.h>
#include<sys/stat.h>
#include<netinet/in.h>
#include<errno.h>
#include<malloc.h>
#include<fcntl.h>
#include<ctype.h>
#include"glob.h"

/* Not always in errno.h. */
extern char *sys_errlist[];

void usage( void )
{
    fputs( "usage: globber <indir> <glob> < <component-list>\n", stdout );
}

static void cleanup( void );
static int copy( int fdout, const char *filename );
static int sizeup( const char *filename, int *len );

static char *globfile = NULL;
static char *components[GLOB_MAX_COMPONENTS];
static int numcomponents = 0;

main( int argc, char *argv[] )
{
    char *indir = NULL;
    char *p, *d;
    char component[GLOB_MAX_COMPONENTS];
    struct globidx idxrecs[GLOB_MAX_COMPONENTS];
    char filename[256];
    char line[256];
    int i;

    globfile = NULL;

    for( i=1; i<argc; ++i )
    {
	if( argv[i][0] == '-' )
	{
	    fprintf( stdout, "ERROR. Unrecognized option \"%s\".\n",
		argv[i] );
	    usage();
	    exit( 1 );
	}
	if( indir == NULL )
	{
	    indir = argv[i];
	    continue;
	}
	if( globfile == NULL )
	{
	    globfile = argv[i];
	    continue;
	}
	fprintf( stdout, "ERROR. Unrecognized argument \"%s\".\n",
	    argv[i] );
	usage();
	exit( 1 );
    }

    if( indir == NULL )
    {
	fputs( "ERROR. Specify an input directory.\n", stdout );
	usage();
	exit( 1 );
    }

    if( globfile == NULL )
    {
	fputs( "ERROR. Specify a glob file.\n", stdout );
	usage();
	exit( 1 );
    }

    while( 1 )
    {
	if( fgets(line,256,stdin) == NULL )
	    break;
	/* Skip leading whitespace. */
	for( p=line; isspace(*p); ++p );
	component[0] = '\0';
	for( d=component; !isspace(*p); ++p,++d )
	    *d = *p;
	*d = '\0';
	if( component[0] == '\0' )
	    continue;

	components[numcomponents] = strdup(component);
	++numcomponents;
	if( numcomponents > GLOB_MAX_COMPONENTS )
	{
	    fprintf( stdout, "ERROR. Max components is %d.\n",
		GLOB_MAX_COMPONENTS );
	    exit( 1 );
	}
    }

    {
	int fd;
	int i;
	int loc = 0;
	int len;
	struct globhdr hdr;

	if( (fd=open(globfile,O_WRONLY|O_CREAT,0664)) == -1 )
	{
	    fprintf( stdout, "ERROR. Bad open on \"%s\". %s.\n",
		globfile, sys_errlist[errno] );
	    cleanup();
	    exit( 1 );
	}

	hdr.magic1 = '%';
	hdr.magic2 = '@';
	hdr.version = htons((short)1);
	hdr.count = htonl(numcomponents);
	if( write(fd,(char*)&hdr,sizeof(hdr)) == -1 )
	{
	    fprintf( stdout, "ERROR. Bad write. %s.\n", sys_errlist[errno] );
	    cleanup();
	    exit( 1 );
	}

	loc = sizeof(struct globhdr);
	loc += sizeof(struct globidx) * numcomponents;
	for( i=0; i<numcomponents; ++i )
	{
	    sprintf( filename, "%s/%s", indir, components[i] );
	    if( !sizeup(filename,&len) )
	    {
		fprintf( stdout, "ERROR. Failed to size \"%s\".\n",
		    filename );
		cleanup();
		exit( 1 );
	    }
	    memset( idxrecs[i].type, 0, 16 );
	    memcpy( idxrecs[i].name, components[i], GLOB_MAX_NAME );
	    idxrecs[i].begin = htonl(loc);
	    idxrecs[i].length = htonl(len);
	    /* glob_dumpidx( &idxrecs[i] ); */
	    loc += len;
	}

	if( write(fd,(char*)(&idxrecs),
	    sizeof(struct globidx)*numcomponents) == -1 )
	{
	    fprintf( stdout, "ERROR. Bad write. %s.\n", sys_errlist[errno] );
	    cleanup();
	    exit( 1 );
	}

	for( i=0; i<numcomponents; ++i )
	{
	    sprintf( filename, "%s/%s", indir, components[i] );
	    if( !copy(fd,filename) )
	    {
		fprintf( stdout, "ERROR. Failed to copy \"%s\".\n",
		    filename );
		cleanup();
		exit( 1 );
	    }
	}

	close( fd );
    }

    for( i=0; i<numcomponents; ++i )
	free( components[i] );

    exit( 0 );
}

static int copy( int fdout, const char *filename )
{
    int fdin;
    int rc;
    char buf[2048];

    if( (fdin=open(filename,O_RDONLY)) == -1 )
    {
	fprintf( stdout, "ERROR. Failed to open \"%s\".\n", filename );
	return 0;
    }

    while( (rc=read(fdin,buf,2048)) != 0 )
    {
	if( rc == -1 )
	{
	    fprintf( stdout, "ERROR. Bad read. %s.\n", sys_errlist[errno] );
	    close( fdin );
	    return 0;
	}

	if( write(fdout,buf,rc) == -1 )
	{
	    fprintf( stdout, "ERROR. Bad write. %s.\n", sys_errlist[errno] );
	    close( fdin );
	    return 0;
	}
    }

    close( fdin );

    return 1;
}

static int sizeup( const char *filename, int *len )
{
    struct stat statrec;

    if( stat(filename,&statrec) == -1 )
    {
	fprintf( stdout, "ERROR. Bad stat on \"%s\". %s.\n",
	    filename, sys_errlist[errno] );
	return 0;
    }
    *len = statrec.st_size;
    return 1;
}

static void cleanup( void )
{
    int i;

    for( i=0; i<numcomponents; ++i )
	free( components[i] );
    unlink( globfile );
}
