/*
** 
**  Copyright (c) 1998 by Larry Plona
** 
**  This distribution  of XPipeman is copyright by Larry Plona.
**  The  package  and  its  component  files  may  be  freely
**  distributed as long as  all copyright notices remain intact
**  and this distribution is not sold  as  commercial software.
** 
**  No warranty of any kind is attached with this software.  It
**  is supplied as is.
** 
** ----------------------------------------------------------------
*/

/*
** C++ Class Imago. See "imago.h" for description.
*/

#include "imago.h"

Imago::Imago( Display *ndisplay, XVisualInfo *nvinfo, Window nwin )
{
    constructor( ndisplay, nvinfo, nwin, 0 );
}

Imago::Imago( Display *ndisplay, XVisualInfo *nvinfo, Window nwin, int local )
{
    constructor( ndisplay, nvinfo, nwin, local );
}

/*
** Create a new Imago object. Several X11 variables must be supplied.
** An XImage is not created at this time.
*/
void Imago::constructor( Display *ndisplay, XVisualInfo *nvinfo, Window nwin,
    int local )
{
    XGCValues gcv;

    display = ndisplay;
    vinfo = nvinfo;
    win = nwin;
    useShm = local;

    if( (vinfo->c_class==PseudoColor) && (vinfo->depth!=8) )
    {
	fputs( "ERROR. PseudoColor visual must be 8 bit.\n", stdout );
	return;
    }

    if( (vinfo->c_class==TrueColor) && (vinfo->depth!=16) &&
	(vinfo->depth!=24) && (vinfo->depth!=32) )
    {
	fputs( "ERROR. TrueColor visual must be 24/32 or 16 bit.\n", stdout );
	return;
    }

    /* We need an X11 GC to use XPutImage(). */
    gcv.foreground = 0;
    gcv.background = 1;
    gcv.function = GXcopy;
    gc = XCreateGC( display, win, GCForeground|GCBackground|GCFunction,
	&gcv );

    imgs[0] = imgs[1] = NULL;
    imgDraw = imgDisp = 0;

    /*
    ** Due to maximum size limits on shared memory, only 8 and 16 bit
    ** depths can take advantage of the MIT-SHM extension.
    */
    if( (vinfo->depth!=8) && (vinfo->depth!=16) )
	useShm = 0;
    if( useShm )
	useShm = XShmQueryExtension(display);
    if( useShm )
	fputs( "Using MIT-SHM extension.\n", stdout );
}

Imago::~Imago()
{
    XFreeGC( display, gc );
    if( imgs[0] != NULL )
    {
	if( useShm )
	{
	    XShmDetach( display, &xshminfo[0] );
	    shmctl( xshminfo[0].shmid, IPC_RMID, 0 );
	}
	XDestroyImage( imgs[0] );
    }
    if( imgs[1] != NULL )
    {
	if( useShm )
	{
	    XShmDetach( display, &xshminfo[1] );
	    shmctl( xshminfo[1].shmid, IPC_RMID, 0 );
	}
	XDestroyImage( imgs[1] );
    }
}

/*
** resize()
**
** This routine creates an XImage for the object.
*/
int Imago::resize( int w, int h )
{
    if( w<1 || h<1 )
	return 0;

    if( imgs[0] != NULL )
	XDestroyImage( imgs[0] );
    if( imgs[1] != NULL )
	XDestroyImage( imgs[1] );

    width = w;
    height = h;

    if( !createImage(width,height,0) )
	return 0;

    imgDraw = imgDisp = 0;

    dirtycnt = 0;

    return 1;
}

/*
** flush()
**
** Run through the list of dirty areas and redraw them on the X11 window.
** We assume that the list of areas has already been optimised.
*/
void Imago::flush( void )
{
    int i;
    int w, h;

    if( imgDisp != imgDraw )
    {
	/*
	fputs( "TMP im fl disp != draw\n", stdout );
	*/
	return;
    }

    for( i=0; i<dirtycnt; ++i )
    {
	// fprintf( stdout, "TMP flush %d,%d %d,%d\n",
	// areas[i].x1, areas[i].y1, areas[i].x2, areas[i].y2 );

	w = (areas[i].x2-areas[i].x1) + 1;
	h = (areas[i].y2-areas[i].y1) + 1;
	if( useShm )
	{
	    XShmPutImage( display, win, gc, imgs[imgDraw],
		areas[i].x1, areas[i].y1, areas[i].x1, areas[i].y1, w, h,
		True );
	}
	else
	{
	    XPutImage( display, win, gc, imgs[imgDraw],
		areas[i].x1, areas[i].y1, areas[i].x1, areas[i].y1, w, h );
	}
    }
    /*
    fprintf( stdout, "TMP im flushed %d\n", dirtycnt );
    */
    dirtycnt = 0;
}

/*
** flush()
**
** Draw a given portion of the image onto the X11 window.
**
** FUTURE. Delete the redrawn area from the list of dirty areas.
*/
void Imago::flush( int x, int y, int w, int h )
{
    if( useShm )
    {
	XShmPutImage( display, win, gc, imgs[imgDraw], x, y, x, y, w, h,
	    True );
    }
    else
	XPutImage( display, win, gc, imgs[imgDisp], x, y, x, y, w, h );
    XFlush( display );
}

/*
** back()
**
** Set the current XImage for drawing and the currently visible XImage.
** If the secondary XImage is referenced and it does not exist, it is
** created.
**
** 00 - Buffer 1.
** 01 - Buffer 2.
** 02 - set draw
** 04 - set disp
*/
int Imago::back( unsigned int flags )
{
    if( ((flags&1)==1) && imgs[1]==NULL )
    {
	if( !createImage(width,height,1) )
	    return 0;
    }

    if( flags&2 )
    {
	imgDraw = flags&1;
	return 1;
    }

    if( flags&4 )
    {
	imgDisp = flags&1;
	dirty( 0, 0, width, height );
	return 1;
    }

    return 1;
}

/*
** dirty()
**
** Add a region to the list of dirty regions. Perform some minimal
** optimisation to coalesce neighboring areas. If a certain amount
** of the image becomes dirty, mark the whole image as dirty.
**
** Collection is important for TrueColor visuals. The amount of data
** (four bytes per pixel) is quite large and has a visible effect
** on performance even on fast machines.
*/
void Imago::dirty( int x, int y, int w, int h )
{
    int i;
    int x2, y2;
    unsigned int covered = 0;
    XSegment *a;

    x2 = x + (w-1);
    y2 = y + (h-1);
    /*
    fprintf( stdout, "TMP im dirty %d,%d %d,%d\n",
	x, y, x2, y2 );
	fflush( stdout );
    */

    for( i=0; i<dirtycnt; ++i )
    {
	a = &areas[i];

	/* See if an existing area covers this one. */
	if( (a->x1<=x) && (a->x2>=x2) && (a->y1<=y) && (a->y2>=y2) )
	{
	    // fputs( "TMP already covered\n", stdout );
	    return;
	}

	/* See if this area covers an existing one. */
	if( (a->x1>=x) && (a->x2<=x2) && (a->y1>=y) && (a->y2<=y2) )
	{
	    a->x1 = x;
	    a->x2 = x2;
	    a->y1 = y;
	    a->y2 = y2;
	    // fprintf( stdout, "TMP recover dirty %d,%d %d %d\n", x, y, w, h );
	    return;
	}

	/*
	** Determine if new area is a horizontal extension of an existing
	** area.
	*/
	if( (a->y1==y) && (a->y2==y2) )
	{
	    if( (x<=a->x1) && (x2>=(a->x1-1)) )
	    {
		a->x1 = x;
		a->x2 = (x2>a->x2)?x2:a->x2;
		return;
	    }

	    if( (x2>=a->x2) && (x<=(a->x2+1)) )
	    {
		a->x1 = (x<a->x1)?x:a->x1;
		a->x2 = x2;
		return;
	    }
	}

	/*
	** Determine if new area is a vertical extension of an existing
	** area.
	*/
	if( (a->x1==x) && (a->x2==x2) )
	{
	    if( (y<=a->y1) && (y2>=(a->y1-1)) )
	    {
		a->y1 = y;
		a->y2 = (y2>a->y2)?y2:a->y2;
		return;
	    }

	    if( (y2>=a->y2) && (y<=(a->y2+1)) )
	    {
		a->y1 = (y<a->y1)?y:a->y1;
		a->y2 = y2;
		return;
	    }
	}

	covered += ((a->x2-a->x1)*(a->y2-a->y1));
    }

    /*
    fprintf( stdout, "TMP covered %d%% %d\n",
	((covered*100)/(width*height)), covered );
    */

    /*
    ** Refresh everything if we have no more dirty records or if
    ** over half of the image is dirty.
    */
    if( dirtycnt >= AN_MAX_AREAS )
    {
	dirtycnt = 1;
	areas[0].x1 = 0;
	areas[0].x2 = width-1;
	areas[0].y1 = 0;
	areas[0].y2 = height-1;
	/*
	fputs( "TMP maxing dirty\n", stdout ); fflush( stdout );
	*/
	return;
    }

    //fprintf( stdout, "TMP new dirty %d,%d %d %d\n", x, y, w, h );
    areas[dirtycnt].x1 = x;
    areas[dirtycnt].x2 = (x+w)-1;
    areas[dirtycnt].y1 = y;
    areas[dirtycnt].y2 = (y+h)-1;
    ++dirtycnt;
}

/*
** copy()
**
** Copy given XImage onto the XImage of the object. Width and
** height of the given image are supplied.
*/
void Imago::copy( XImage *im, int x, int y, int w, int h )
{
    char *ibuf = imgs[imgDraw]->data;
    int i, j;
    int dstart;
    int sidx, didx;
    int dscanwidth = imgs[imgDraw]->bytes_per_line;
    int sscanwidth = im->bytes_per_line;

    /*
    fprintf( stdout, "TMP cp %d,%d %d %d\n", x, y, w, h );
    */

    if( x<0 || x>=width || y<0 || y>=height
	|| w<0 | w>width || h<0 || h>height )
    {
	fprintf( stdout, "ERROR. Bad copy, %d,%d w%d h%d.\n",
	    x, y, w, h );
	fflush( stdout );
	return;
    }

    if( vinfo->depth == 8 )
	dstart = x+(y*dscanwidth);
    else if( vinfo->depth == 16 )
	dstart = (x*2)+(y*dscanwidth);
    else
	dstart = (x*4)+(y*dscanwidth);

    if( vinfo->depth == 8 )
    {
	for( i=0; i<h; ++i )
	{
	    for( j=0; j<w; ++j )
		ibuf[dstart+j+(i*dscanwidth)] = im->data[j+(i*sscanwidth)];
	}
    }
    else if( vinfo->depth == 16 )
    {
	for( i=0; i<h; ++i )
	{
	    for( j=0; j<w; ++j )
	    {
		sidx = (j*2)+(i*sscanwidth);
		didx = dstart+(j*2)+(i*dscanwidth);
		ibuf[didx+0] = im->data[sidx+0];
		ibuf[didx+1] = im->data[sidx+1];
	    }
	}
    }
    else /* 24 bit (in 32 bit value). */
    {
	for( i=0; i<h; ++i )
	{
	    for( j=0; j<w; ++j )
	    {
		sidx = (j*4)+(i*sscanwidth);
		didx = dstart+(j*4)+(i*dscanwidth);
		ibuf[didx+0] = im->data[sidx+0];
		ibuf[didx+1] = im->data[sidx+1];
		ibuf[didx+2] = im->data[sidx+2];
		ibuf[didx+3] = im->data[sidx+3];
	    }
	}
    }

    dirty( x, y, w, h );
}

/*
** Imago::copy()
**
** Copy given XImage onto XImage of object. Width, height and
** source coordinates of given image are supplied.
*/
void Imago::copy( XImage *im, int sx, int sy, int w, int h, int dx, int dy )
{
    char *ibuf = imgs[imgDraw]->data;
    int i, j;
    int didx, sidx;
    int dstart, sstart;
    int dscanwidth = imgs[imgDraw]->bytes_per_line;
    int sscanwidth = im->bytes_per_line;

    /*
    fprintf( stdout, "TMP cp %d,%d %d %d\n", x, y, w, h );
    */
    if( vinfo->depth == 8 )
    {
	dstart = dx+(dy*dscanwidth);
	sstart = sx+(sy*sscanwidth);
    }
    else if( vinfo->depth == 16 )
    {
	dstart = (dx*2)+(dy*dscanwidth);
	sstart = (sx*2)+(sy*sscanwidth);
    }
    else
    {
	dstart = (dx*4)+(dy*dscanwidth);
	sstart = (sx*4)+(sy*sscanwidth);
    }

    if( vinfo->depth == 8 )
    {
	for( i=0; i<h; ++i )
	{
	    for( j=0; j<w; ++j )
	    {
		ibuf[dstart+j+(i*dscanwidth)] = im->data[sstart+j+
		    (i*sscanwidth)];
	    }
	}
    }
    else if( vinfo->depth == 16 )
    {
	for( i=0; i<h; ++i )
	{
	    for( j=0; j<w; ++j )
	    {
		didx = dstart+(j*2)+(i*dscanwidth);
		sidx = sstart+(j*2)+(i*sscanwidth);
		/* High byte is unused. */
		ibuf[didx+0] = im->data[sidx+0];
		ibuf[didx+1] = im->data[sidx+1];
		ibuf[didx+2] = im->data[sidx+2];
		ibuf[didx+3] = im->data[sidx+3];
	    }
	}
    }
    else
    {
	for( i=0; i<h; ++i )
	{
	    for( j=0; j<w; ++j )
	    {
		didx = dstart+(j*4)+(i*dscanwidth);
		sidx = sstart+(j*4)+(i*sscanwidth);
		/* High byte is unused. */
		ibuf[didx+0] = im->data[sidx+0];
		ibuf[didx+1] = im->data[sidx+1];
		ibuf[didx+2] = im->data[sidx+2];
		ibuf[didx+3] = im->data[sidx+3];
	    }
	}
    }

    dirty( dx, dy, w, h );
}

/*
** copy()
**
*/
void Imago::copy( XImage *im, XImage *mask, int x, int y, int w, int h )
{
    int i, j;
    int b;
    int sidx;
    unsigned long p;
    int sscanwidth = im->bytes_per_line;

    /*
    fprintf( stdout, "TMP cp %d,%d %d %d\n", x, y, w, h );
    */

    for( i=0; i<h; ++i )
    {
	for( j=0; j<w; ++j )
	{
	    /* Mask is 1 bit color, eight pixels per byte. */
	    b = mask->data[(j/8)+(i*mask->bytes_per_line)];
	    b &= (1<<(j%8));

	    if( b == 0 )
		continue;

	    if( vinfo->depth == 8 )
	    {
		p = im->data[j+(i*im->bytes_per_line)];
		XPutPixel( imgs[imgDraw], x+j, y+i, p );
	    }
	    else if( vinfo->depth == 16 )
	    {
		sidx = (j*2) + (i*sscanwidth);
		p = (unsigned char)im->data[sidx+0]<<8;
		p |= (unsigned char)im->data[sidx+1]<<0;
		XPutPixel( imgs[imgDraw], x+j, y+i, p );
	    }
	    else
	    {
		sidx = (j*4) + (i*sscanwidth);
		p  = (unsigned char)im->data[sidx+0]<<16;
		p |= (unsigned char)im->data[sidx+1]<<8;
		p |= (unsigned char)im->data[sidx+2]<<0;
		XPutPixel( imgs[imgDraw], x+j, y+i, p );
	    }
	}
    }
    dirty( x, y, w, h );
}

/*
** Return a new XImage made from a portion of the current XImage.
*/
XImage *Imago::subImage( int x, int y, int w, int h )
{
    return XSubImage( imgs[imgDraw], x, y, w, h );
}

/*
** Imago::clear()
**
** Clear a portion of the image with a given color.
*/
void Imago::clear( unsigned long c, int dx, int dy, int w, int h )
{
    int i, j;
    unsigned char *ibuf8 = (unsigned char*)imgs[imgDraw]->data;
    unsigned short *ibuf16 = (unsigned short*)imgs[imgDraw]->data;
    unsigned int *ibuf32 = (unsigned int*)imgs[imgDraw]->data;

    if( vinfo->depth == 8 )
    {
	for( i=0; i<h; ++i )
	{
	    for( j=0; j<w; ++j )
		ibuf8[(dx+j)+((dy+i)*width)] = c;
	}
    }
    else if( vinfo->depth == 16 )
    {
	for( i=0; i<h; ++i )
	{
	    for( j=0; j<w; ++j )
		ibuf16[(dx+j)+((dy+i)*width)] = c;
	}
    }
    else
    {
	for( i=0; i<h; ++i )
	{
	    for( j=0; j<w; ++j )
		ibuf32[(dx+j)+((dy+i)*width)] = c;
	}
    }

    dirty( dx, dy, w, h );
}

/*
** drawflag should be set when the created image will be used by an
** X server (rather than just internally). If not set, X shared
** memory will not be used.
*/
int Imago::createImage( int width, int height, int which )
{
    char *buf;
    int size;
    int pad;

    if( vinfo->depth == 8 )
    {
	size = width*height;
	pad = width;
    }
    else if( vinfo->depth == 16 )
    {
	size = width*height*2;
	pad = width*2;
    }
    else
    {
	size = width*height*4;
	pad = width*4;
    }

    /*
    ** Convert pad to either 32, 16 or 8; the largest integral bit size
    ** for a single horizontal line of pixels.
    */
    if( (pad%4) == 0 )
	pad = 32;
    else if( (pad%2) == 0 )
	pad = 16;
    else
	pad = 8;

    if( useShm )
    {
	/*
	shmevent = XShmGetEventBase(display) + ShmCompletion;
	*/

	imgs[which] = XShmCreateImage( display, vinfo->visual,
	    vinfo->depth, ZPixmap, 0, &xshminfo[which], width, height );

	if( !createShm(which) )
	{
            fputs( "ERROR. Failed to create shared memory.\n", stdout );
	    return 0;
	}

        if( !XShmAttach(display,&xshminfo[which]) )
	{
	    shmctl( xshminfo[which].shmid, IPC_RMID, 0 );
            fputs( "ERROR. Failed to attach shared memory to X server.\n",
		stdout );
	    return 0;
	}

    }
    else
    {
	buf = (char*)malloc( size );

	imgs[which] = XCreateImage( display, vinfo->visual,
	    vinfo->depth, ZPixmap, 0, buf, width, height, pad, 0 );
    }

    return 1;
}

int Imago::createShm( int which )
{
    const int keybase = ('p'<<24)|('i'<<16)|('p'<<8)|('e'<<0);
    const int shmsize = imgs[which]->bytes_per_line*height;
    int i;
    int id;

    for( i=0; i<4; ++i )
    {
	/* Determine if this key exists. */
	if( (id=shmget((key_t)(keybase+i),shmsize,0777)) == -1 )
	{
	    /* Try to create this key. */
	    if( (id=shmget((key_t)(keybase+i),shmsize,IPC_CREAT|0777)) == -1 )
	    {
		fprintf( stdout,
		    "Could not get shared memory.\n"
		    "Bad shmget(). %s.\n",
		    sys_errlist[errno] );
		continue;
	    }
	    fprintf( stdout, "Shared memory id:%d, key:0x%x.\n",
		id, keybase+i );
	    break;
	}
    }

    if( i >= 4 )
    {
	fprintf( stdout,
	    "Could not find any shared memory segments.\n"
	    "Try deleteing keys 0x%x to 0x%x.\n",
	    keybase, keybase+4 );
	return 0;
    }

    xshminfo[which].shmid = id;
    imgs[which]->data = xshminfo[which].shmaddr = (char*)shmat( id, 0, 0 );

    return 1;
}

/*
** Wave effect.
**
** Scott Scriven. Copyright 1997. scriven@CS.colostate.edu
** Drops from top view.
**
** Not fully implemented. Can be run in current state.
*/

int Imago::wave_sintab[2048];
int Imago::wave_costab[2048];
const float Imago::pi = 3.14159265358979323846;
#define FSINMAX 2047
#define SINFIX 16
#define FSINBITS 16

/*
** Create a wave effect. Return a state information record which must
** be passed to each call of effectWave(). The bounds of the area to
** be affected are passed in.
*/
struct imago_effect_wave *Imago::effectWaveInit( int ex, int ey,
    int ew, int eh )
{
    char *ibuf;
    const int size = ew * eh; // For convenience.
    struct imago_effect_wave *rec;

    if( (rec=(struct imago_effect_wave*)malloc(
	sizeof(struct imago_effect_wave))) == NULL )
    {
	fprintf( stdout, "ERROR. Bad malloc. %s.\n", sys_errlist[errno] );
	return NULL;
    }

    rec->x = ex;
    rec->y = ey;
    rec->w = ew;
    rec->h = eh;

    if( (ibuf=(char*)malloc(size)) == NULL )
    {
	fprintf( stdout, "ERROR. Bad malloc. %s.\n", sys_errlist[errno] );
	return 0;
    }
    rec->imgBuf = XCreateImage( display, vinfo->visual,
	vinfo->depth, ZPixmap, 0, ibuf, ew, eh, 32, 0 );
    if( (rec->htab[0]=(int*)malloc(size*sizeof(int))) == NULL )
    {
	fprintf( stdout, "ERROR. Bad malloc. %s.\n", sys_errlist[errno] );
	return 0;
    }
    if( (rec->htab[1]=(int*)malloc(size*sizeof(int))) == NULL )
    {
	fprintf( stdout, "ERROR. Bad malloc. %s.\n", sys_errlist[errno] );
	return 0;
    }
    rec->imgBg = XSubImage( imgs[imgDraw], ex, ey, ew, eh );

    rec->page = 0;
    // 
    // FUTURE. Pass in.
    rec->radius = 40;
    rec->pheight = 256;
    rec->xang = rand()%2048;
    rec->yang = rand()%2048;
    rec->swirlangle = rand()%2048;
    rec->density = 2; //4.

    memset( rec->htab[0], 0, size*sizeof(int) );
    memset( rec->htab[1], 0, size*sizeof(int) );

    {
	int i;
	double angle;

	for(i=0; i<2048; i++)
	{
	    angle = (float)i * (pi/1024.0);
	    wave_sintab[i] = (int)(sin(angle) * (float)0x10000);
	    wave_costab[i] = (int)(cos(angle) * (float)0x10000);
	}
    }

    return rec;
}

/*
** Run on iteration of a wave effect. A state info record from a previous
** call to effectWaveInit() is passed in. Generally, this routine would
** be called from a timeout callback.
*/
int Imago::effectWave( struct imago_effect_wave *rec )
{
    int ew = rec->w;
    int eh = rec->h;

    // Big Splash.
    if( rand()%20 == 0 )
    {
	rec->radius = 10+rand()%70;
	//rec->pheight = ((rand()%5)+1)*256;
	rec->pheight = -1400;
	//fprintf( stdout, "TMP ph %d r %d\n", rec->pheight, rec->radius );
	wave_hblob( rec, -1, -1, rec->radius/2, -rec->pheight*6 );
    }

#if 0
    // Surfer.
    {
	int lx, ly;

        lx = (ew/2)
          + ((
             (wave_costab[rec->swirlangle&FSINMAX]) * (25)
            ) >> 16);
        ly = (eh/2)
          + ((
             (wave_sintab[rec->swirlangle&FSINMAX]) * (25)
            ) >> 16);
        rec->swirlangle += 50;
	wave_hblob( rec, lx, ly, rec->radius/3, rec->pheight );
    }
#endif

#if 0
    // Rain drops.
    {
	int lx, ly;

	lx = rand()%(ew-2) + 1;
	ly = rand()%(eh-2) + 1;
	rec->htab[rec->page][ly*ew+lx] = rand()%(rec->pheight<<2);
    }
#endif

    wave_draw( rec );

    if( 1 )
    {
	wave_calcwater( rec );
    }
    else
    {
	memcpy( rec->htab[rec->page^1], rec->htab[rec->page],
	    (ew*eh)*sizeof(int) );
    }

    rec->page ^= 1;

#if 0
    {
	int x;

	for(x=0; x<20; x++)
	{
	    XPutPixel( rec->imgBuf, x<<1, 0, 0 );
	    XPutPixel( rec->imgBuf, (x<<1)+1, 0, 0 );
	}
    }
#endif

    return 1;
}

void Imago::wave_draw( struct imago_effect_wave *rec )
{
    int dx, dy;
    int x, y;
    unsigned char c;
    int ew = rec->w;
    int eh = rec->h;
    int offset=ew+1;
    int *ptr = rec->htab[rec->page];

    for( y=(eh-1)*ew; offset<y; offset+=2 )
    {
	for( x=offset+ew-2; offset<x; ++offset )
	{
	    dx = ptr[offset] - ptr[offset+1];
	    dy = ptr[offset] - ptr[offset+ew];
	    c = rec->imgBg->data[offset+ew*(dy>>3)+(dx>>3)];

	    rec->imgBuf->data[offset] = c;
	    ++offset;
	    dx = ptr[offset] - ptr[offset+1];
	    dy = ptr[offset] - ptr[offset+ew];
	    c = rec->imgBg->data[offset + ew*(dy>>3) + (dx>>3)];
	    rec->imgBuf->data[offset] = c;
	}
    }

    // Draw the imago like a sprite...
    copy( rec->imgBuf, rec->x, rec->y, ew, eh );
}

void Imago::wave_hblob( struct imago_effect_wave *rec, int lx, int ly,
    int radius, int lh )
{
    int rquad;
    int cx, cy, cyq;
    int left, top, right, bottom;
    int ew = rec->w;
    int eh = rec->h;

    rquad = radius * radius;

    // Make a randomly-placed blob...
    if( lx < 0 )
	lx = 1+radius+rand()%(ew-2*radius-1);
    if( ly < 0 )
	ly = 1+radius+rand()%(eh-2*radius-1);

    left = -radius; right = radius;
    top = -radius; bottom = radius;

    // Perform edge clipping...
    if( lx-radius < 1 )
	left -= (lx-radius-1);
    if( ly - radius < 1 )
	top  -= (ly-radius-1);
    if( lx+radius > ew-1 )
	right -= (lx+radius-ew+1);
    if( ly + radius > eh-1)
	bottom-= (ly+radius-eh+1);

    for(cy = top; cy < bottom; cy++)
    {
	cyq = cy*cy;
	for(cx = left; cx < right; cx++)
	{
	    if(cx*cx + cyq < rquad)
	    {
		rec->htab[rec->page][ew*(cy+ly) + (cx+lx)] += lh;
	    }
	}
    }
}

void Imago::wave_calcwater( struct imago_effect_wave *rec )
{
    int ew = rec->w;
    int eh = rec->h;
    int newh;
    int count = rec->w+1;
    int *newptr = rec->htab[rec->page^1];
    int *oldptr = rec->htab[rec->page];
    int lx, ly;

    for( ly=(eh-1)*ew; count<ly; count+=2 )
    {
	for( lx=count+ew-2; count<lx; ++count )
	{
	    newh = ( (oldptr[count+ew]
		+ oldptr[count-ew]
		+ oldptr[count+1]
		+ oldptr[count-1]
		+ oldptr[count-ew-1]
		+ oldptr[count-ew+1]
		+ oldptr[count+ew-1]
		+ oldptr[count+ew+1]
		)>>2 )
		- newptr[count];
	    newptr[count] = newh - (newh>>rec->density);
	}
    }
}

/*
** Fire effect.
** Scott Scriven. Copyright 1996. scriven@CS.colostate.edu.
** Doesn't work on top of other images. Problems in port limit color use.
*/

struct imago_effect_fire *Imago::effectFireInit( int ex, int ey,
    int ew, int eh )
{
    char *ibuf;
    const int size = ew * eh; // For convenience.
    struct imago_effect_fire *rec;

    if( (rec=(struct imago_effect_fire*)malloc(
	sizeof(struct imago_effect_fire))) == NULL )
    {
	fprintf( stdout, "ERROR. Bad malloc. %s.\n", sys_errlist[errno] );
	return NULL;
    }

    rec->x = ex;
    rec->y = ey;
    rec->w = ew;
    rec->h = eh;

    if( (ibuf=(char*)malloc(size)) == NULL )
    {
	fprintf( stdout, "ERROR. Bad malloc. %s.\n", sys_errlist[errno] );
	return 0;
    }
    rec->img = XCreateImage( display, vinfo->visual,
	vinfo->depth, ZPixmap, 0, ibuf, ew, eh, 32, 0 );

    for( int i=0; i<size; ++i )
	rec->img->data[i] = 0xff;

    return rec;
}

int Imago::effectFire( struct imago_effect_fire *rec )
{
    unsigned int c = 6;
    unsigned char *buf = (unsigned char*)rec->img->data;
    int ew = rec->w;
    int eh = rec->h;
    int offset;

    for( int i=ew*(eh-1); i<ew*eh; ++i )
    {
	buf[i] = c;
	if( rand()%6 == 0 )
	{
	    //c = (rand()&1)*255; // Highest or lowest.
	    c = (rand()%249)+6; // Highest to lowest.
	    if( c < 6 )
		c = 6;
	    //fprintf( stdout, "TMP pixel %d\n", c );
	}
    }

    for( offset=0; offset<ew*(eh-1); ++offset )
    {
	// Add three pixels below current pixel.
	c= (unsigned char)buf[offset]
	    +(unsigned char)buf[offset+ew+1]
	    +(unsigned char)buf[offset+ew-1]
	    +(unsigned char)buf[offset+ew];
	c = (c+2) >> 2;
	c &= 0xfe;          // Prevent overflow (optimized)
	if( c < 6 )
	    c = 6;
#if 0
	if( (offset>=ew*(eh-2)) && (offset<(ew*(eh-2))+4) )
	{
	    unsigned int x;

	    x = (unsigned char)buf[offset]+(unsigned char)buf[offset+ew+1]
		+(unsigned char)buf[offset+ew-1]
		+(unsigned char)buf[offset+ew];
	    fprintf( stdout, "TMP %u -> %u %u %u -> %u\n",
		(unsigned char)buf[offset],
		(unsigned char)buf[offset+ew+1],
		(unsigned char)buf[offset+ew-1],
		(unsigned char)buf[offset+ew], x );
	    fprintf( stdout, "TMP %u -> %u\n",
		(x+2)>>2, ((x+2)>>2)&0xfe );
	}
#endif
	buf[offset] = c+1;
    }

    copy( rec->img, rec->x, rec->y, ew, eh );

    return 1;
}

/*
** Melt effect. One frame "melts" into the other. The melt starts at
** the top and moves downward in columns of four pixels. Each column
** moves downward a random length on each iteration.
*/

struct imago_effect_melt *Imago::effectMeltInit( int ex, int ey,
    int ew, int eh )
{
    struct imago_effect_melt *rec;

    if( (rec=(struct imago_effect_melt*)malloc(
	sizeof(struct imago_effect_melt))) == NULL )
    {
	fprintf( stdout, "ERROR. Bad malloc. %s.\n", sys_errlist[errno] );
	return NULL;
    }

    /* Table of depths for each column. */
    if( (rec->htab=(int*)malloc(
	sizeof(int)*(ew/4))) == NULL )
    {
	fprintf( stdout, "ERROR. Bad malloc. %s.\n", sys_errlist[errno] );
	return NULL;
    }

    rec->x = ex;
    rec->y = ey;
    rec->w = ew;
    rec->h = eh;

    memset( rec->htab, 0, (ew/4)*sizeof(int) );

    // The page the screen will melt from.
    rec->page1 = (unsigned char*)imgs[imgDisp]->data;
    // The page the screen will melt to.
    rec->page2 = (unsigned char*)imgs[imgDraw]->data;

    return rec;
}

int Imago::effectMelt( struct imago_effect_melt *rec )
{
    unsigned int j, k;
    unsigned int a;
    unsigned int min, max;
    unsigned char *page1 = rec->page1;
    unsigned char *page2 = rec->page2;
    unsigned short *page1s = (unsigned short*)rec->page1;
    unsigned short *page2s = (unsigned short*)rec->page2;
    unsigned int *page1i = (unsigned int*)rec->page1;
    unsigned int *page2i = (unsigned int*)rec->page2;

    /*
    ** Loop through each column.  The columns are four pixels wide.
    */

    /* Highest point of melt line. Initialize at bottom. */
    min = rec->h-1;

    max = 0;

    for( j=0; j<rec->w/4; ++j )
    {
	/*
	** Keep track of minimum y starting point. Make sure that it
	** does not exceed the height of the image.
	*/
	if( rec->htab[j] < min )
	    min = rec->htab[j];

	/* Has this column has reached the bottom? */
	if( rec->htab[j] >= rec->h )
	    continue;

	/*
	** Offset to top of column.
	**
	** Note that this calculation works for BOTH 8, 16 and 32 bit
	** depths since htab is interpreted either as an array of
	** bytes, shorts or ints.
	*/
	a = (j*4) + (rec->w*rec->htab[j]);

	// Drip from 8 to 15 pixels.
	k = (rand()%8)+8;

	// Make sure that the drip does not exceed the image height.
	if( rec->htab[j]+k >= rec->h )
	{
	    /*
	    fprintf( stdout, "TMP drip %d -> %d\n",
		k, rec->h-rec->htab[j] );
	    */
	    k = (rec->h-rec->htab[j]) - 1;
	}

	// Update starting y position for next iteration.
	rec->htab[j] += k;

	// Record maximum y position.
	if( rec->htab[j] > max )
	    max = rec->htab[j];

	/*
	** Draw the current drip.
	*/
	// FUTURE. We're assuming effect size and image size are equal!
	if( vinfo->depth == 8 )
	{
	    while( k-- )
	    {
		page1[a+0] = page2[a+0];
		page1[a+1] = page2[a+1];
		page1[a+2] = page2[a+2];
		page1[a+3] = page2[a+3];
		a += rec->w;
	    }
	}
	else if( vinfo->depth == 16 )
	{
	    while( k-- )
	    {
		page1s[a+0] = page2s[a+0];
		page1s[a+1] = page2s[a+1];
		page1s[a+2] = page2s[a+2];
		page1s[a+3] = page2s[a+3];
		a += rec->w;
	    }
	}
	else
	{
	    while( k-- )
	    {
		page1i[a+0] = page2i[a+0];
		page1i[a+1] = page2i[a+1];
		page1i[a+2] = page2i[a+2];
		page1i[a+3] = page2i[a+3];
		a += rec->w;
	    }
	}
    }

    /*
    fprintf( stdout, "TMP min %u max %u %d\n", min, max, (max-min) );
    fflush( stdout );
    */

    // If the lowest starting point is off the image, its done.
    if( min >= rec->h )
	return 2;

    // Clip maximum drip height.
    // FUTURE. Should not be able to happen.
    if( max >= rec->h )
	max = rec->h-1;

    /*
    fprintf( stdout, "TMP dirty %d,%d w%dh%d\n",
	rec->x, rec->y+min, rec->w, max-min );
    */

    dirty( rec->x, rec->y+min, rec->w, max-min );

    if( min >= (rec->h-1) )
	return 2;

    return 1;
}

int Imago::effectMeltFinis( struct imago_effect_melt *rec )
{
    free( rec->htab );
    free( rec );
    return 1;
}

/*
** Fire2
**
** Adapted from Pawn Production. Copyright 1997.
*/

struct imago_effect_fire2 *Imago::effectFire2Init( int ex, int ey,
    int ew, int eh )
{
    const int size = ew*eh; // For convenience.
    struct imago_effect_fire2 *rec;

    if( (rec=(struct imago_effect_fire2*)malloc(
	sizeof(struct imago_effect_fire2))) == NULL )
    {
	fprintf( stdout, "ERROR. Bad malloc. %s.\n", sys_errlist[errno] );
	return NULL;
    }

    rec->x = ex;
    rec->y = ey;
    rec->w = ew;
    rec->h = eh;

    rec->img = XSubImage( imgs[imgDraw], ex, ey, ew, eh );

    memset( rec->img->data, 0, size );

    return rec;
}

int Imago::effectFire2( struct imago_effect_fire2 *rec )
{
    unsigned char *buf = (unsigned char*)rec->img->data;
    unsigned int c;
    int eh = rec->h;
    int ew = rec->w;
    int x, y;

    // Bottom row is a seed.
    c = ew*(eh-1);
    for( x=0; x<ew; ++x )
    {
	buf[c+x] = (char)(rand()%255);
    }

    // For each row except the last.
    for( y=1; y<(eh-2); ++y )
    {
	for( x=0; x<ew; ++x )
	{
	    //the pixel above
	    c = buf[x+(y-1)*ew];
	    //the pixels below
	    c += buf[(x-1)+(y+1)*ew];
	    c += buf[x+(y+1)*ew];
	    c += buf[(x+1)+(y+1)*ew];
	    c /= 4;
	    buf[x+(y*ew)] = (unsigned char)c;
	}
    }
    copy( rec->img, rec->x, rec->y, rec->w, rec->h );

    return 1;
}

int Imago::effectFire2Finis( struct imago_effect_fire2 *rec )
{
    XDestroyImage( rec->img );
    free( rec );
    return 1;
}

/*
** Stellae
**
** Horizontally scrolling starfield.
*/
struct imago_effect_stellae *Imago::effectStellaeInit( int ex, int ey,
    int ew, int eh )
{
    int i;
    struct imago_effect_stellae *rec;

    if( (rec=(struct imago_effect_stellae*)malloc(
	sizeof(struct imago_effect_stellae))) == NULL )
    {
	fprintf( stdout, "ERROR. Bad malloc. %s.\n", sys_errlist[errno] );
	return NULL;
    }

    rec->x = ex;
    rec->y = ey;
    rec->w = ew;
    rec->h = eh;

    if( (rec->map=(unsigned short*)calloc(eh,sizeof(unsigned short))) == NULL )
    {
	fprintf( stdout, "ERROR. Bad calloc. %s.\n", sys_errlist[errno] );
	return NULL;
    }

    for( i=0; i<eh; ++i )
	rec->map[i] = (rand()%ew);

    /* Area to black. */
    clear( 0, ex, ey, ew, eh );

    return rec;
}

int Imago::effectStellae( struct imago_effect_stellae *rec )
{
    unsigned i = 0;
    unsigned pos;
    unsigned int start = rec->x+(rec->y*width);
    unsigned char *buf8 = (unsigned char*)imgs[imgDraw]->data;
    unsigned short *buf16 = (unsigned short*)imgs[imgDraw]->data;
    unsigned int *buf32 = (unsigned int*)imgs[imgDraw]->data;

    if( vinfo->depth == 8 )
    {
	/*
	** Pseudo color. Assume 0 is black. Assume 1, 2, 3 and 4
	** are visible against a black background.
	*/
	for( i=0; i<rec->h; ++i )
	{
	    pos = start + (i*width);
	    buf8[pos+rec->map[i]] = 0;
	    rec->map[i] += 1;
	    if( rec->map[i] > rec->w )
		rec->map[i] = 0;
	    buf8[pos+rec->map[i]] = 1;

	    ++i;
	    pos += width;
	    buf8[pos+rec->map[i]] = 0;
	    rec->map[i] += 2;
	    if( rec->map[i] > rec->w )
		rec->map[i] = 0;
	    buf8[pos+rec->map[i]] = 2;

	    ++i;
	    pos += width;
	    buf8[pos+rec->map[i]] = 0;
	    rec->map[i] += 3;
	    if( rec->map[i] > rec->w )
		rec->map[i] = 0;
	    buf8[pos+rec->map[i]] = 3;

	    ++i;
	    pos += width;
	    buf8[pos+rec->map[i]] = 0;
	    rec->map[i] += 4;
	    if( rec->map[i] > rec->w )
		rec->map[i] = 0;
	    buf8[pos+rec->map[i]] = 4;
	}
    }
    else if( vinfo->depth == 16 )
    {
	for( i=0; i<rec->h; ++i )
	{
	    pos = start + (i*width);
	    buf16[pos+rec->map[i]] = 0;
	    rec->map[i] += 1;
	    if( rec->map[i] > rec->w )
		rec->map[i] = 0;
	    buf16[pos+rec->map[i]] = 0xffff;

	    ++i;
	    pos += width;
	    buf16[pos+rec->map[i]] = 0;
	    rec->map[i] += 2;
	    if( rec->map[i] > rec->w )
		rec->map[i] = 0;
	    buf16[pos+rec->map[i]] = 0xffff;

	    ++i;
	    pos += width;
	    buf16[pos+rec->map[i]] = 0;
	    rec->map[i] += 3;
	    if( rec->map[i] > rec->w )
		rec->map[i] = 0;
	    buf16[pos+rec->map[i]] = 0xffff;

	    ++i;
	    pos += width;
	    buf16[pos+rec->map[i]] = 0;
	    rec->map[i] += 4;
	    if( rec->map[i] > rec->w )
		rec->map[i] = 0;
	    buf16[pos+rec->map[i]] = 0xffff;
	}
    }
    else
    {
	for( i=0; i<rec->h; ++i )
	{
	    pos = start + (i*width);
	    buf32[pos+rec->map[i]] = 0;
	    rec->map[i] += 1;
	    if( rec->map[i] > rec->w )
		rec->map[i] = 0;
	    buf32[pos+rec->map[i]] = 0xbbbbbb;

	    ++i;
	    pos += width;
	    buf32[pos+rec->map[i]] = 0;
	    rec->map[i] += 1;
	    if( rec->map[i] > rec->w )
		rec->map[i] = 0;
	    buf32[pos+rec->map[i]] = 0xcccccc;

	    ++i;
	    pos += width;
	    buf32[pos+rec->map[i]] = 0;
	    rec->map[i] += 3;
	    if( rec->map[i] > rec->w )
		rec->map[i] = 0;
	    buf32[pos+rec->map[i]] = 0xeeeeee;

	    ++i;
	    pos += width;
	    buf32[pos+rec->map[i]] = 0;
	    rec->map[i] += 3;
	    if( rec->map[i] > rec->w )
		rec->map[i] = 0;
	    buf32[pos+rec->map[i]] = 0xffffff;
	}
    }

    dirty( rec->x, rec->y, rec->w, rec->h );

    return 1;
}

int Imago::effectStellaeFinis( struct imago_effect_stellae *rec )
{
    free( rec->map );
    free( rec );
    return 1;
}
