#ifndef IMAGO_H
#define IMAGO_H 1

/*
** 
**  Copyright (c) 1998 by Larry Plona
** 
**  This distribution  of XPipeman is copyright by Larry Plona.
**  The  package  and  its  component  files  may  be  freely
**  distributed as long as  all copyright notices remain intact
**  and this distribution is not sold  as  commercial software.
** 
**  No warranty of any kind is attached with this software.  It
**  is supplied as is.
** 
** ----------------------------------------------------------------
*/

/*
** C++ class Imago.
**
** This class manages an X11 XImage. Members include functions to copy
** other XImages into the object's image, maintaining a list of areas
** which have been changed and flushing changed areas to the X11
** server.
**
** The Imago class also provides several graphical effects which may
** be applied to the image.
**
** This class requires only XLib and the Colores object.
**
** Supported visuals:
**    8 bit PseudoColor
**    16 bit TrueColor
**    24 bit TrueColor (using 32 bit data)
**    32 bit TrueColor (using 24 bit values)
**
** Note that some X servers report 24 bit color as 32 bit (because 24
** bit color is usually implemented as a series of 32 bit numbers). Imago
** only supports the 24-in-32 scheme.
*/

#include <sys/shm.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/XShm.h>

#define AN_MAX_AREAS 32

/* State information for the Wave effect. */
struct imago_effect_wave
{
    int *htab[2];
    int x, y, w, h;
    int xang, yang, swirlangle;
    int radius;
    int pheight;
    int density;
    int page;
    XImage *imgBuf, *imgBg;
};

/* State information for the Fire effect. */
struct imago_effect_fire
{
    XImage *img;
    int x, y, w, h;
};

/* State information for the Melt effect. */
struct imago_effect_melt
{
    unsigned char *page1;
    unsigned char *page2;
    int x, y, w, h;
    int *htab;
};

/* State information for the Fire2 effect. */
struct imago_effect_fire2
{
    XImage *img;
    int x, y, w, h;
};

/* State information for the Stellae effect. */
struct imago_effect_stellae
{
    int x, y, w, h;
    unsigned short *map;
};

/*
** Not provided in all releases of X.
*/
#ifdef NEED_SHMQUERY_PROTO
extern "C" {
extern Bool XShmQueryExtension( Display* );
}
#endif

class Imago
{
    /* X11. */
    Display *display;
    XVisualInfo *vinfo;
    Window win;
    GC gc;
    int shmevent;
    XShmSegmentInfo xshminfo[2];
    /* */
    int useShm;
    int dirtycnt;
    XSegment areas[AN_MAX_AREAS]; /* Interpreted as rectangles, not lines. */
    int width;
    int height;
    /* */
    int createImage( int, int, int );
    int createShm( int );
    void constructor( Display*, XVisualInfo*, Window, int );
    /* Effects. */
    static int wave_sintab[2048];
    static int wave_costab[2048];
    void wave_hblob( struct imago_effect_wave*, int, int, int, int );
    void wave_draw( struct imago_effect_wave* );
    void wave_calcwater( struct imago_effect_wave* );
public:
    static const float pi;
    XImage *imgs[2];
    int imgDraw;
    int imgDisp;
    Imago( Display*, XVisualInfo*, Window, int );
    Imago( Display*, XVisualInfo*, Window );
    ~Imago();
    void dirty( int, int, int, int );
    void flush( void );
    void flush( int, int, int, int );
    void copy( XImage*, int, int, int, int );
    void copy( XImage*, int, int, int, int, int, int );
    void copy( XImage*, XImage*, int, int, int, int );
    XImage *subImage( int, int, int, int );
    void clear( unsigned long, int, int, int, int );
    int resize( int, int );
    int back( unsigned int );
    /* Effects. */
    struct imago_effect_wave *effectWaveInit( int, int, int, int );
    int effectWave( struct imago_effect_wave* );
    struct imago_effect_fire *effectFireInit( int, int, int, int );
    int effectFire( struct imago_effect_fire* );
    struct imago_effect_melt *effectMeltInit( int, int, int, int );
    int effectMelt( struct imago_effect_melt* );
    int effectMeltFinis( struct imago_effect_melt* );
    struct imago_effect_fire2 *effectFire2Init( int, int, int, int );
    int effectFire2( struct imago_effect_fire2* );
    int effectFire2Finis( struct imago_effect_fire2* );
    struct imago_effect_stellae *effectStellaeInit( int, int, int, int );
    int effectStellae( struct imago_effect_stellae* );
    int effectStellaeFinis( struct imago_effect_stellae* );
};

extern int errno;
extern char *sys_errlist[];

#endif
