/*
** 
**  Copyright (c) 1998 by Larry Plona
** 
**  This distribution  of XPipeman is copyright by Larry Plona.
**  The  package  and  its  component  files  may  be  freely
**  distributed as long as  all copyright notices remain intact
**  and this distribution is not sold  as  commercial software.
** 
**  No warranty of any kind is attached with this software.  It
**  is supplied as is.
** 
** ----------------------------------------------------------------
*/

/*
** Applier
**
** applier <colormap>
**
** Read the given Xpm file and extract its colormap. Then replace the colormap
** of all files whose names appear on standard input with the specified map.
**
** This is a blind copy. No attempt is made to adjust the colormap to fit
** the previous map.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/xpm.h>

extern int errno;
extern char *sys_errlist[];

#define TO16(x) ((65535*(x))/255)

static void usage( void );
static const char *findSignifer( XpmColor*, int );
static int getMaster( const char*, XpmImage*, XpmInfo* );
static int doMap( const char*, XpmImage* );

static const int maxprint = 92;
static const char *printable =
    " .XoO+@#$%&*=-;:>,<1234567890qwertyuipasdfghjklzxcvbnmMNBVCZ"
    "ASDFGHJKLPIUYTREWQ!~^/()_`'][{}|";

static int remap[maxprint];
static int count = 0;
static XpmColor table[maxprint];
static Display *dpy;


int main( int argc, char *argv[] )
{
    const char *mapfile;
    int i;
    int rc;
    int flagNewCmap = 0;
    XpmImage imMap;
    XpmInfo ifMap;

    mapfile = NULL;
    for( i=1; i<argc; ++i )
    {
	if( argv[i][0] == '-' )
	{
	    fprintf( stdout, "ERROR. Unknown option, \"%s\".\n", argv[i] );
	    usage();
	}
	if( mapfile == NULL )
	{
	    mapfile = argv[i];
	    continue;
	}
    }

    if( mapfile == NULL )
    {
	fputs( "ERROR. A colormap Xpm must be given.\n", stdout );
	usage();
    }

    if( (dpy=XOpenDisplay(":0.0")) == NULL )
    {
	fprintf( stdout, "ERROR. Bad XOpenDisplay.\n" );
	exit( 1 );
    }

    if( !getMaster(mapfile,&imMap,&ifMap) )
	return 0; // Exit.

    /*
    ** Read each basename.
    */
    {
	char *p, *d;
	char line[255+1];
	char name[255+1];

	while( 1 )
	{
	    if( fgets(line,255,stdin) == NULL )
		break;

	    line[strlen(line)-1] = '\0';

	    /* Strip leading whitespace. */
	    for( p=line; isspace(*p); ++p );

	    /* Blank line. */
	    if( *p =='\0' )
		continue;

	    for( d=name; !isspace(*p)&&(*p!='\0'); ++p,++d )
		*d = *p;
	    *d = '\0';

	    fprintf( stdout, "Processing \"%s\".\n", name );
	    if( (rc=doMap(name,&imMap)) == 0 )
	    {
		fprintf( stdout, "ERROR. Could not process \"%s\".\n", name );
		break;
	    }

	    if( rc >= 2 )
		flagNewCmap = 1;
	}
    }

    /*
    ** Update colormap file. Create an XpmImage and write it out as
    ** an Xpm file.
    */
    if( flagNewCmap )
    {
	unsigned int *data;
	unsigned int size;
	int i;

	size = (40*7);
	if( (data=(unsigned int*)malloc(size*sizeof(unsigned int))) == NULL )
	{
	    fprintf( stdout, "ERROR. Bad malloc, %s.\n",
		sys_errlist[errno] );
	    return 1; // Exit.
	}

	for( i=0; i<size; ++i )
	{
	    data[i] = 0;
	}
	for( i=0; i<count; ++i )
	{
	    data[i] = i;
	}

	imMap.width = 40;
	imMap.height = 7;
	imMap.cpp = 1;
	imMap.ncolors = count;
	imMap.colorTable = table;
	imMap.data = data;
	if( (rc=XpmWriteFileFromXpmImage((char*)"nmap.xpm",&imMap,&ifMap)) !=
	    XpmSuccess )
	{
	    fprintf( stdout, "ERROR. Bad XpmWriteFileFromXpmImage. %s.\n",
		XpmGetErrorString(rc) );
	    return 0;
	}
    }

    XpmFreeXpmImage( &imMap );
    XCloseDisplay( dpy );

    return 0; // Exit.

}

static int getMaster( const char *mapfile, XpmImage *imMap,
   XpmInfo *ifMap )
{
    int i;
    int rc;

    if( (rc=XpmReadFileToXpmImage((char*)mapfile,
	imMap,ifMap)) != XpmSuccess )
    {
	fprintf( stdout, "ERROR. Bad XpmReadFileToImage. %s.\n",
	    XpmGetErrorString(rc) );
	return 0;
    }

    for( i=0; i<imMap->ncolors; ++i )
    {
	fprintf( stdout, "%2s \"%20s\"\n",
	    imMap->colorTable[i].string,
	    imMap->colorTable[i].c_color );
	table[count].string = strdup(imMap->colorTable[i].string );
	table[count].c_color = strdup(imMap->colorTable[i].c_color );
	table[count].symbolic = NULL;
	table[count].m_color = NULL;
	table[count].g4_color = NULL;
	table[count].g_color = NULL;
	++count;
    }

    return 1;
}

static int doMap( const char *infile, XpmImage *imMap )
{
    int i, j;
    int rc;
    int newColors = 0;
    XpmImage imOut;
    XpmImage imIn;
    XpmInfo ifIn;

    ifIn.valuemask = XpmReturnExtensions|XpmReturnComments;
    if( (rc=XpmReadFileToXpmImage((char*)infile,&imIn,&ifIn)) != XpmSuccess )
    {
	fprintf( stdout, "ERROR. Bad XpmReadFileToImage. %s.\n",
	    XpmGetErrorString(rc) );
	return 0;
    }

    /*
    ** Loop through infile colormap. Find colors not is master colormap.
    ** Add new colors. Create map table between input colormap and master
    ** colormap.
    */
    for( i=0; i<imIn.ncolors; ++i )
    {
	// If it's not a hex specification, look it up.
	if( imIn.colorTable[i].c_color[0] != '#' )
	{
	    char cbuf[15+1];
	    XColor xcolor;

	    if( XParseColor(dpy,DefaultColormap(dpy,DefaultScreen(dpy)),
		imIn.colorTable[i].c_color,&xcolor) != 1 )
	    {
		fprintf( stdout, "ERROR. Bad XParseColor, \"%s\".\n",
		    imIn.colorTable[i].c_color );
	    }
	    else
	    {
		sprintf( cbuf, "#%04x%04x%04x",
		    xcolor.red, xcolor.green, xcolor.blue );
		fprintf( stdout, "  \"%s\" -> \"%s\".\n",
		    imIn.colorTable[i].c_color, cbuf );
		// FUTURE. Can we free the old c_color?
		imIn.colorTable[i].c_color = strdup(cbuf);
	    }
	}

	// Convert hex specifications to lowercase.
	if( imIn.colorTable[i].c_color[0] == '#' )
	{
	    char *p;

	    for( p=imIn.colorTable[i].c_color; *p!='\0'; ++p )
		*p=tolower(*p);
	}

	// FUTURE. Convert 8bit to 16bit.
	if( !strcmp(imIn.colorTable[i].c_color,"#000000") )
	{
	    free( imIn.colorTable[i].c_color );
	    imIn.colorTable[i].c_color = strdup("#000000000000");
	}
	if( !strcmp(imIn.colorTable[i].c_color,"#ffffff") )
	{
	    free( imIn.colorTable[i].c_color );
	    imIn.colorTable[i].c_color = strdup("#ffffffffffff");
	}

	for( j=0; j<count; ++j )
	{
	    if( !strcmp(table[j].c_color,imIn.colorTable[i].c_color) )
		break;
	}
	if( j >= count )
	{
	    const char *signifer;

	    fprintf( stdout, "   Adding \"%s\" from \"%s\".\n",
		imIn.colorTable[i].c_color, infile );
	    if( count >= maxprint )
	    {
		fputs( "ERROR. Exceeding 96 colors.\n", stdout );
		return 0;
	    }
	    ++newColors;
	    if( (signifer=findSignifer(table,count)) == NULL )
	    {
		fputs( "ERROR. No more color tokens.\n", stdout );
		return 0; // Exit.
	    }
	    table[count].string = (char*)signifer;
	    table[count].c_color = strdup(imIn.colorTable[i].c_color);
	    table[count].symbolic = NULL;
	    table[count].m_color = NULL;
	    table[count].g4_color = NULL;
	    table[count].g_color = NULL;
	    remap[i] = count;
	    ++count;
	    continue;
	}
	remap[i] = j;
    }

    /*
    ** Remap signiferes and data of input image to sync with the
    ** master colormap.
    */
    {
	unsigned int idx;
	int i;
	int size;

	size = imIn.width*imIn.height;
	for( i=0; i<size; ++i )
	{
	    idx = imIn.data[i];
	    imIn.data[i] = remap[idx];
	}
    }

    imOut = imIn;
    imOut.colorTable = table;
    imOut.ncolors = count;

    ifIn.valuemask = XpmExtensions|XpmComments;
    if( (rc=XpmWriteFileFromXpmImage((char*)infile,&imOut,&ifIn)) !=
	XpmSuccess )
    {
	fprintf( stdout, "ERROR. Bad XpmWriteFileFromXpmImage. %s.\n",
	    XpmGetErrorString(rc) );
	return 0;
    }

    /*
    ** Neither Xpm Image can be freed since they use a global color table.
    */

    return newColors+1;
}

static void usage( void )
{
    fputs( "usage: applier <colormap>\n", stdout );
    exit( 1 );
}

static const char *findSignifer( XpmColor *table, int count )
{
    int i, j;

    for( i=0; i<maxprint; ++i )
    {
	for( j=0; j<count; ++j )
	{
	    if( table[j].string[0] == printable[i] )
		break;
	}
	if( j >= count )
	{
	    char buf[2];

	    buf[0] = printable[i];
	    buf[1] = '\0';
	    fprintf( stdout, "   New token \"%s\".\n", buf );
	    return strdup( buf );
	}
    }
    return NULL;
}
