/*
** 
**  Copyright (c) 1998 by Larry Plona
** 
**  This distribution  of XPipeman is copyright by Larry Plona.
**  The  package  and  its  component  files  may  be  freely
**  distributed as long as  all copyright notices remain intact
**  and this distribution is not sold  as  commercial software.
** 
**  No warranty of any kind is attached with this software.  It
**  is supplied as is.
** 
** ----------------------------------------------------------------
*/

/*
** Collector
** 
** collector <indir> <outdir>    < filelist
** collector -c <indir> <outdir> < filelist
** 
** Read Xpm file basenames from standard input. Do the following for each.
** If a name is followed by whitespace and an 'N', then the following will
** be modified as described below.
**     o Read all files named <indir>/<basename>-<nn>.xpm where <nn> is a
**       two digit number.
**     o Create a single Xpm file named <outdir>/<basename>.xpm. The method
**       of collection is that used by Arnaud Le Hors' xpmcollect program
**       (i.e. subsquent frames are stored as extensions of type XPMPIXMAP).
**     o Unless the N option was given, the first Xpm in the sequence must
**       contain an extension of type ANIM.
** 
** Currently there is a hardcoded frame count maximum.
** 
** Existing extensions in the initial frame will be preserved.
**
** If the N option is given, an ANIM extension will not be removed as it
** should.
**
** Multiple ANIM extensions or ANIM extensions in subsequent Xpm files will
** not be removed.
** 
** Frame counts (<nn> above) begin with zero. No number may be skipped.
**
** The "-c" options indicates that generated files in <outdir> should be
** removed (no other work is done).
*/

#include<stdio.h>
#include<unistd.h>
#include<stdlib.h>
#include<string.h>
#include<ctype.h>
#include"glob.h"

#include<X11/xpm.h>

/* Maximum number of frames for a given Xpm. */
#define COLL_MAX_FRAMES 32

static void usage( void );
static int first_frame( XpmInfo*, XpmImage*, const char*, int );
static int linecount( char** );

main( int argc, char *argv[] )
{
    char infile[255+1];
    char outfile[255+1];
    char name[255+1];
    char line[255+1];
    char **framesbuf[COLL_MAX_FRAMES];
    char *indir, *outdir;
    char *p, *d;
    int i, rc;
    int frameidx, fextidx;
    int doAnim;
    int flagClean = 0;
    XpmImage image;
    XpmInfo info;
    XpmExtension framesexts[COLL_MAX_FRAMES];
    XpmExtension *newexts;

    indir = outdir = p = NULL;
    for( i=1; i<argc; ++i )
    {
	if( !strcmp(argv[i],"-c") )
	{
	    ++flagClean;
	    continue;
	}
	if( argv[i][0] == '-' )
	{
	    fprintf( stdout, "ERROR. Unknown option, \"%s\".\n", argv[i] );
	    usage();
	}
	if( indir == NULL )
	{
	    indir = argv[i];
	    continue;
	}
	if( outdir == NULL )
	{
	    outdir = argv[i];
	    continue;
	}
    }

    if( indir == NULL )
    {
	fputs( "ERROR. An input path must be given.\n", stdout );
	usage();
    }
    if( outdir == NULL )
    {
	fputs( "ERROR. An output path must be given.\n", stdout );
	usage();
    }

    /*
    ** Read each basename.
    */
    while( 1 )
    {
	if( fgets(line,255,stdin) == NULL )
	    break;

	doAnim = 1;

	line[strlen(line)-1] = '\0';

	/* Strip leading whitespace. */
	for( p=line; isspace(*p); ++p );

	/* Blank line. */
	if( *p =='\0' )
	    continue;

	/* Copy basename. */
	for( d=name; !isspace(*p)&&(*p!='\0'); ++p,++d )
	    *d = *p;
	*d = '\0';

	for( ; isspace(*p); ++p );

	if( *p == 'N' )
	    doAnim = 0;

	if( flagClean )
	{
	    sprintf( outfile, "%s/%s.xpm", outdir, name );
	    unlink( outfile );
	    continue;
	}

	fprintf( stdout, "Processing \"%s\".\n", name );
	frameidx = -1; /* Not counting initial frame. */
	fextidx = 0;   /* Index to framesexts (frames after first). */

	/*
	** Read in each frame: <indir>/<basename>-<nn>.xpm.
	*/
	while( 1 )
	{
	    sprintf( infile, "%s/%s-%02d.xpm", indir, name, (frameidx+1) );
	    if( access(infile,R_OK) == -1 )
	        break;

	    /*
	    ** First frame is special.
	    */
	    if( frameidx == -1 )
	    {
		if( !first_frame(&info,&image,infile,doAnim) )
		    break;
	    }
	    else
	    {
		if( (rc=XpmReadFileToData(infile,&(framesbuf[fextidx])))
		    != XpmSuccess )
		{
		    fprintf( stdout, "ERROR. Bad frame %s. %s.\n",
			infile, XpmGetErrorString(rc) );
		    /* Too many nested loops, just go to next frame. */
		    continue;
		}
		/* Create an extension name. */
		sprintf( infile, "%02d", (frameidx+1) );
		p = (char*)malloc(strlen(infile)+10+1);
		sprintf( p, "XPMPIXMAP %s", infile );
		framesexts[fextidx].name = p;
		framesexts[fextidx].nlines = linecount(framesbuf[fextidx]);
		framesexts[fextidx].lines = framesbuf[fextidx];
		++fextidx;
	    }
	    ++frameidx;
	}
	if( frameidx == -1 )
	{
	    fprintf( stdout, "ERROR. No frames for %s.\n", name );
	    continue;
	}

	fprintf( stdout, "Found %d frames.\n", frameidx );

	/*
	** Write out the new file.
	*/

	/*
	** Allocate space for extensions appearing in the first frame plus
	** the extensions created by adding subsequent frames.
	*/
	newexts = (XpmExtension*)malloc(
	    sizeof(XpmExtension)*(frameidx+info.nextensions) );
	memcpy( newexts, info.extensions,
	    sizeof(XpmExtension)*info.nextensions );
	memcpy( &(newexts[info.nextensions]), framesexts,
	    sizeof(XpmExtension)*frameidx );

	info.extensions = newexts;
	info.nextensions += frameidx; /* First frame is not an extension. */
	info.valuemask = XpmComments | XpmExtensions;
	sprintf( outfile, "%s/%s.xpm", outdir, name );
	if( (rc=XpmWriteFileFromXpmImage(outfile,&image,&info)) != XpmSuccess )
	{
	    fprintf( stdout, "ERROR. Bad XpmWriteFileFromXpmImage. %s.\n",
		XpmGetErrorString(rc) );
	}

	/*
	** Free data associated with the initial frame.
	*/
	XpmFreeXpmImage( &image );
	/* XpmFreeXpmInfo( &info ); */

	/*
	** Free extension data for extensions containing subsequent frames.
	*/
	for( i=0; i<fextidx; ++i )
	{
	    free( framesexts[i].name );
	    free( framesexts[i].lines );
	}
    }

    exit(0);
}

static void usage( void )
{
    fputs( "usage: collector [-c] <indir> <outdir>\n", stdout );
    exit( 1 );
}

/*
** first_frame()
** 
** Read given Xpm file as an XpmImage. Grab Xpm info as well. Verify that
** this file is suitable as the first frame in an Xpm animation file.
*/
static int first_frame( XpmInfo *info, XpmImage *image,
    const char *infile, int doAnim )
{
    int rc;

    info->valuemask = XpmReturnComments | XpmReturnExtensions;

    if( (rc=XpmReadFileToXpmImage((char*)infile,image,info)) != XpmSuccess )
    {
	fprintf( stdout, "ERROR. Bad XpmReadFileToXpmImage on %s. %s.\n",
	    infile, XpmGetErrorString(rc) );
	return 0;
    }

    fprintf( stdout, "Found %d initial extensions.\n", info->nextensions );

    /*
    ** Verify that there is an ANIM extension.
    */
    if( doAnim )
    {
	for( rc=0; rc<info->nextensions; ++rc )
	{
	    if( strcmp(info->extensions[rc].name,"ANIM") == 0 )
		break;
	}
	if( rc >= info->nextensions )
	{
	    fputs( "ERROR. No ANIM extension.\n", stdout );
	    return 0;
	}
    }

    return 1;
}

/*
** linecount()
** 
** Given a pointer to Xpm data (an array of strings), determine the number of
** strings in the array. The first line of Xpm data is "x y ncolors [else]".
*/
static int linecount( char **first )
{
    int y, ncolors;

    sscanf( first[0], "%*d %d %d %*s", &y, &ncolors );

    return y + ncolors + 1;
}
