/*
** masker.
**
** Apply a background pixmap, en masse, to several pixmaps with
** transparent regions.
**
** masker <indir> <outdir> [background.xpm]
**
** For each file in <indir>, apply either <background.xpm> or
** the file named <indir>/bg.xpm. Write the result in <outdir>
** under the same name.
*/
#include <X11/X.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/cursorfont.h>
#include <X11/Core.h>
#include <X11/Shell.h>
#include <X11/Xaw/Box.h>

#include <xpm.h>

#include "stdio.h"

static Display *dpy;
static Colormap mkcolormap();
static int addcolor();

static usage( void )
{
    fputs( "usage: masker <indir> <outdir> [background xpm]\n", stdout );
    exit( 1 );
}

int main( int argc, char **argv )
{
    int rc;
    int i;
    char *p;
    char bgfile[255+1];
    char colorfile[255+1];
    char maskfile[255+1];
    char outfile[255+1];
    char line[255+1];
    char *indir, *outdir;
    XpmAttributes attr;
    XGCValues gcv;
    Pixmap pixHold;
    Pixmap pixBg;
    Pixmap pixMask;
    Pixmap pixOvr;
    Colormap colormap;
    GC gcCopy, gcMask;
    Window win;

    indir = outdir = p = NULL;
    for( i=1; i<argc; ++i )
    {
#if 0
	if( !strcmp("-maskdir",argv[i]) )
	{
	    ++i;
	    if( i >= argc )
	    {
		fputs( "ERROR. Argument needed for -maskdir.\n", stdout );
		exit( 1 );
	    }
	    maskdir = argv[i];
	}
#endif
	if( argv[i][0] == '-' )
	{
	    fprintf( stdout, "ERROR. Unknown option, \"%s\".\n", argv[i] );
	    usage();
	}
	if( indir == NULL )
	{
	    indir = argv[i];
	    continue;
	}
	if( outdir == NULL )
	{
	    outdir = argv[i];
	    continue;
	}
	if( p == NULL )
	{
	    p = argv[i];
	    continue;
	}
    }

    if( indir == NULL )
    {
	fputs( "ERROR. An input path must be given.\n", stdout );
	usage();
    }
    if( outdir == NULL )
    {
	fputs( "ERROR. An output path must be given.\n", stdout );
	usage();
    }

    if( p == NULL )
	sprintf( bgfile, "%s/bg.xpm", indir );
    else
	strcpy( bgfile, p );
    sprintf( colorfile, "%s/colormap.xpm", outdir );

    dpy = XOpenDisplay( NULL );
    win = DefaultRootWindow( dpy );

    gcv.function = GXcopy;
    gcCopy = XCreateGC( dpy, win, GCFunction, &gcv );

    gcv.function = GXcopy;
    gcMask = XCreateGC( dpy, win, GCFunction, &gcv );

    colormap = mkcolormap( colorfile );

    pixHold = XCreatePixmap( dpy, win, 40, 40, 8 );

    /*
    ** Read background.
    */

    attr.valuemask = XpmColormap;
    attr.colormap = colormap;
    if( (rc=XpmReadFileToPixmap(dpy,win,bgfile,
	&pixBg,NULL,&attr)) != XpmSuccess )
    {
	fprintf( stdout, "ERROR. Bad XpmReadFileToPixmap \"%s\", %s.\n",
	    bgfile, XpmGetErrorString(rc) );
	XFreePixmap( dpy, pixHold );
	XFreeGC( dpy, gcCopy );
	XFreeGC( dpy, gcMask );
	XFreeColormap( dpy, pixHold );
	exit( 1 );
    }

    while( 1 )
    {
	if( fgets(line,255,stdin) == NULL )
	    break;

	line[strlen(line)-1] = '\0';

	for( p=line; isspace(*p); ++p );

	if( *p == '\0' )
	    continue;

	sprintf( maskfile, "%s/%s", indir, p );
	sprintf( outfile, "%s/%s", outdir, p );
	fprintf( stdout, "Processing \"%s\".\n", maskfile );

	attr.valuemask = XpmColormap;
	attr.colormap = colormap;
	if( (rc=XpmReadFileToPixmap(dpy,win,maskfile,
	    &pixOvr,&pixMask,&attr)) != XpmSuccess )
	{
	    fprintf( stdout, "ERROR. Bad pixmap, \"%s\".\n", p );
	    break;
	}
	rc = XCopyArea( dpy, pixBg, pixHold, gcCopy, 0, 0, 40, 40, 0, 0 );

	if( pixMask == 0 )
	{
	    fprintf( stdout, "No mask for \"%s\".\n", maskfile );
	    continue;
	}

	gcv.clip_mask = pixMask;
	XChangeGC( dpy, gcMask, GCClipMask, &gcv );

	XCopyArea( dpy, pixOvr, pixHold, gcMask, 0, 0, 40, 40, 0, 0 );

	attr.valuemask = XpmColormap;
	attr.colormap = colormap;
	if( (rc=XpmWriteFileFromPixmap(dpy,outfile,pixHold,NULL,&attr))
	    != XpmSuccess )
	{
	    fprintf( stdout, "WARNING. XpmWriteFileFromPixmap, \"%s\".\n",
		outfile, XpmGetErrorString(rc) );
	}

	XFreePixmap( dpy, pixMask );
	XFreePixmap( dpy, pixOvr );
    }

    XFreePixmap( dpy, pixHold );
    XFreePixmap( dpy, pixBg );
    XFreeGC( dpy, gcCopy );
    XFreeGC( dpy, gcMask );
    XFreeColormap( dpy, pixHold );
    exit( 0 );
}

Colormap mkcolormap( filename )
    char *filename;
{
    int i;
    int rc;
    int sn;
    char *colorname;
    Colormap cmap;
    XVisualInfo visual_info;
    XpmImage xpmimage;
    XpmAttributes attr;

    sn = DefaultScreen(dpy);

    /* Make sure that we are using a color visual. */
    if (!XMatchVisualInfo(dpy,sn,DefaultDepth(dpy,sn),
        PseudoColor,&visual_info))
    {
        if (!XMatchVisualInfo(dpy,sn,
            DefaultDepth(dpy,sn),DirectColor,&visual_info))
        {
            if (!XMatchVisualInfo(dpy,sn,
                DefaultDepth(dpy,sn),TrueColor,&visual_info))
            {
                fputs( "You must have a color display.", stdout );
                exit( 1 );
            }
        }
    }

    cmap = XCreateColormap(dpy,
	RootWindow(dpy,XDefaultScreen(dpy)),
	visual_info.visual, AllocNone);

    attr.valuemask = XpmColormap;
    attr.colormap = cmap;

    rc = XpmReadFileToXpmImage( filename, &xpmimage, NULL );
    if( rc != XpmSuccess )
    {
	fprintf( stdout, "Error. Bad colormap \"%s\".\n", filename );
	fprintf( stdout, "XpmReadFileToXpmImage, \"%s\".\n",
	    XpmGetErrorString(rc) );
	exit( 1 );
    }

    for( i=0; i<xpmimage.ncolors; ++i )
    {
	colorname = xpmimage.colorTable[i].symbolic;
	if( xpmimage.colorTable[i].symbolic == NULL )
	    colorname = xpmimage.colorTable[i].c_color;
	if( xpmimage.colorTable[i].c_color == NULL )
	{
	    fprintf( stdout, "Error. Bad color in colormap.\n" );
	    fprintf( stdout, "XPM index is \"%s\".\n",
		xpmimage.colorTable[i].string );
	    exit( 1 );
	}
	if( !addcolor(cmap,colorname) )
	{
	    fprintf( stdout, "Error. Could not allocate color \"%s\".\n",
		colorname );
	    exit( 1 );
	}
    }

    XpmFreeXpmImage( &xpmimage );
    return cmap;
}

static int addcolor( colormap, colorname )
    Colormap colormap;
    char *colorname;
{
    XColor color;

    if (XParseColor(dpy,colormap,colorname,&color) == 0 )
    {
	return 0;
    }
    if (XAllocColor(dpy, colormap, &color) == 0)
    {
	return 0;
    }
    return 1;
}
