/*
** 
**  Copyright (c) 1998 by Larry Plona
** 
**  This distribution  of XPipeman is copyright by Larry Plona.
**  The  package  and  its  component  files  may  be  freely
**  distributed as long as  all copyright notices remain intact
**  and this distribution is not sold  as  commercial software.
** 
**  No warranty of any kind is attached with this software.  It
**  is supplied as is.
** 
** ----------------------------------------------------------------
*/

#include <stdio.h>
#include <math.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include "DialP.h"

#define DIAL_RADIANS(x) (DIAL_PI*2.0*(x)/360.0)
#define DIAL_MAX_ANGLE 345.0
#define DIAL_MIN_DIAMETER		3

/*
** Widget hooks.
*/
static void hook_initialize( Widget, Widget, ArgList, Cardinal* );
static void hook_destroy( Widget w );
static void hook_resize( Widget w );
static void hook_redisplay( Widget, XEvent*, Region );
static Boolean hook_setValues( Widget, Widget, Widget, ArgList, Cardinal* );
static void hook_setValuesAlmost( Widget, Widget, XtWidgetGeometry*,
    XtWidgetGeometry* );
static XtGeometryResult hook_queryGeometry( Widget, XtWidgetGeometry*, 
    XtWidgetGeometry* );

static void mkDiameter( Widget, unsigned int*, Dimension*, Dimension* );
static void draw_dial( Widget );
static void draw_indicator( Widget, Boolean );
static void draw_arrow( Widget, Boolean );
static void mkArrow( Widget );

/*
** Action procedures.
*/
static void timeout_turn( XtPointer, XtIntervalId* );
static void action_left( Widget, XEvent*, char**, Cardinal* );
static void action_right( Widget, XEvent*, char**, Cardinal* );
static void action_stop( Widget, XEvent*, char**, Cardinal* );

static XtResource resources[] =
{
    {
	XljpNactivateCallback,
	XljpCActivateCallback,
	XtRCallback,
	sizeof(XtCallbackList),
	XtOffsetOf(LjpDialRec,dial.activate_callback),
	XtRImmediate, NULL
    },
    {
	XljpNvalue,
	XljpCValue,
	XtRInt,
	sizeof(int),
	XtOffsetOf(LjpDialRec,dial.value),
	XtRImmediate,
	(XtPointer)0
    },
    {
	XljpNmaxValue,
	XljpCMaxValue,
	XtRInt,
	sizeof(int),
	XtOffsetOf(LjpDialRec,dial.max_val),
	XtRImmediate,
	(XtPointer)100
    },
    {
	XljpNminValue,
	XljpCMinValue,
	XtRInt,
	sizeof(int),
	XtOffsetOf(LjpDialRec,dial.min_val),
	XtRImmediate,
	(XtPointer) 0
    },
    {
	XljpNturnDelay,
	XljpCTurnDelay,
	XtRInt,
	sizeof(int),
	XtOffsetOf(LjpDialRec,dial.turn_delay),
	XtRImmediate,
	(XtPointer) 50
    },
    {
	XljpNdialColor,
	XljpCDialColor,
	XtRPixel,
	sizeof(Pixel),
	XtOffsetOf(LjpDialRec,dial.pixelDial),
	XtRString,
	XtDefaultBackground
    },
    {
	XljpNindicatorColor,
	XljpCIndicatorColor,
	XtRPixel,
	sizeof(Pixel),
	XtOffsetOf(LjpDialRec,dial.pixelIndicator),
	XtRString,
	XtDefaultForeground
    }
};
	  
/* Action table - where procedures are mapped to strings */
static XtActionsRec dial_actions[] =
{
    { "turn-left",	action_left },
    { "turn-right",	action_right },
    { "turn-stop",	action_stop }
};

/* Translation table - where events are mapped to action names */
static char defaultTranslations[] = "\
<Btn1Down>:	turn-left()\n\
<Btn3Down>:	turn-right()\n\
<Btn1Up>:	turn-stop()\n\
<Btn3Up>:	turn-stop()";

LjpDialClassRec ljpDialClassRec =
{
   {
      (WidgetClass) &ljpPrimitiveClassRec,      /* superclass         */
      "Dial",                            	/* class_name         */
      sizeof(LjpDialRec),                   	/* widget_size        */
      NULL,                          		/* class_initialize   */
      NULL,                            		/* class_part_initiali*/
      FALSE,                                    /* class_inited       */
      hook_initialize,                          /* initialize         */
      NULL,                                     /* initialize_hook    */
      XtInheritRealize,                         /* realize            */
      dial_actions,                             /* actions            */
      XtNumber(dial_actions),                   /* num_actions        */
      resources,                                /* resources          */
      XtNumber(resources),                      /* num_resources      */
      NULLQUARK,                                /* xrm_class          */
      TRUE,                                     /* compress_motion    */
      XtExposeCompressMaximal,                  /* compress_exposure  */
      TRUE,                                     /* compress_enterleave*/
      FALSE,                                    /* visible_interest   */
      hook_destroy,                             /* destroy            */
      hook_resize,                              /* resize             */
      hook_redisplay,                 		/* expose             */
      hook_setValues,                           /* set_values         */
      NULL,                                     /* set_values_hook    */
      NULL,                            		/* set_values_almost  */
      NULL,                                     /* get_values_hook    */
      NULL,                                     /* accept_focus       */
      XtVersion,                                /* version            */
      NULL,                                     /* callback_private   */
      defaultTranslations,                      /* tm_table           */
      hook_queryGeometry,             		/* query_geometry     */
      XtInheritDisplayAccelerator,              /* display accel      */
      NULL,                                     /* extension          */
    },
    /* Primitive class. */
    {
	XljpInheritDrawFrame,			/* draw_frame.		*/
	NULL                                    /* extension		*/
    },
    /* Dial. */
    {
	NULL,                                   /* extension          */
    }
};

WidgetClass ljpDialWidgetClass = (WidgetClass)&ljpDialClassRec;

static void hook_initialize( Widget w, Widget neww, ArgList args,
    Cardinal *num_args )
{
    LjpDialWidget dw = (LjpDialWidget)w;
    LjpDialWidget new_dw = (LjpDialWidget)neww;
    Dimension min_width, min_height;
    XGCValues values;
    unsigned long valuemask;

    if( new_dw->dial.max_val <= 0 )
	new_dw->dial.max_val = 1;

    if( new_dw->dial.max_val < new_dw->dial.min_val )
       new_dw->dial.min_val = 0;

    if( new_dw->dial.value > new_dw->dial.max_val )
	new_dw->dial.value = new_dw->dial.max_val;

    if( new_dw->dial.value < new_dw->dial.min_val )
	new_dw->dial.value = new_dw->dial.min_val;

    if( new_dw->dial.turn_delay < 1 )
	new_dw->dial.turn_delay = 1;

    min_width = DIAL_MIN_DIAMETER +
	(2*(new_dw->primitive.margin_width+new_dw->core.border_width));
    min_height = DIAL_MIN_DIAMETER +
	(2*(new_dw->primitive.margin_width+new_dw->core.border_width));
		
    if( dw->core.width == 0 )
	new_dw->core.width = 64 + (2*(new_dw->primitive.margin_width));
    else
    {
	if( new_dw->core.width < min_width )
	    new_dw->core.width = min_width;
    }

    if( dw->core.height == 0 )
	new_dw->core.height = 64 + (2*(new_dw->primitive.margin_height));
    else
    {
	if (new_dw->core.height < min_width)
	    new_dw->core.height = min_height;
    }

    new_dw->dial.orig_width = new_dw->core.width;
    new_dw->dial.orig_height = new_dw->core.height;

    new_dw->dial.move_clockwise = True;
    new_dw->dial.turning = False;
    new_dw->dial.timer_id = 0;
    new_dw->dial.angle_offset = 0;
    new_dw->dial.timer_id = 0;

    mkArrow( neww );
}

static void hook_destroy( Widget w )
{
    LjpDialWidget dw = (LjpDialWidget)w;
   
    if( dw->dial.timer_id )
    {
	XtRemoveTimeOut( dw->dial.timer_id );
	dw->dial.timer_id = 0;
    }

    if( dw->dial.activate_callback != NULL )
	XtRemoveAllCallbacks( w, XljpNactivateCallback );
}

static void hook_resize( Widget w )
{
    mkArrow( w );
}

static void hook_redisplay( Widget w, XEvent *event, Region region )
{
    LjpDialWidget dw = (LjpDialWidget)w;

    draw_dial( w );
}

static Boolean hook_setValues( Widget oldw, Widget w, Widget neww,
    ArgList args, Cardinal *num_args )
{
    LjpDialWidget old_dw = (LjpDialWidget) oldw;
    LjpDialWidget new_dw = (LjpDialWidget) neww;
    Dimension min_width, min_height;
    Boolean redisplay = False;

    if( new_dw->dial.max_val <= 0 )
	new_dw->dial.max_val = 1;

    if( new_dw->dial.min_val >= new_dw->dial.max_val )
       new_dw->dial.min_val = 0;

    if( (new_dw->dial.max_val != old_dw->dial.max_val ||
         new_dw->dial.value != old_dw->dial.value) &&
         new_dw->dial.value > new_dw->dial.max_val )
    {
	new_dw->dial.value = new_dw->dial.max_val;
	redisplay = True;
    }

    if( (new_dw->dial.min_val != old_dw->dial.min_val ||
         new_dw->dial.value != old_dw->dial.value) &&
         new_dw->dial.value < new_dw->dial.min_val )
    {
	new_dw->dial.value = new_dw->dial.min_val;
	mkArrow( neww );
	redisplay = True;
    }

    if( new_dw->dial.turn_delay < 1 )
	new_dw->dial.turn_delay = 1;

    if( new_dw->core.width == 0 )
	new_dw->core.width = old_dw->core.width;
    if( new_dw->core.height == 0 )
	new_dw->core.height = old_dw->core.height;

#if 0
    if( new_dw->core.background_pixel != old_dw->core.background_pixel ||
        new_dw->dial.pixelDial != old_dw->dial.pixelDial ||
        new_dw->dial.pixelIndicator != old_dw->dial.pixelIndicator ||
        new_dw->dial.pixelBottom != old_dw->dial.pixelBottom ||
        new_dw->dial.pixelTop != old_dw->dial.pixelTop )
    {
	XGCValues values;
	unsigned long valuemask;

	XtReleaseGC( neww, new_dw->dial.gcTop);
	XtReleaseGC( neww, new_dw->dial.gcBottom);
	XtReleaseGC( neww, new_dw->dial.gcClear);
	mkGC( new_dw );
	redisplay = True;
    }
#endif

    min_width = DIAL_MIN_DIAMETER + (2 *(new_dw->primitive.margin_width));
    min_height = DIAL_MIN_DIAMETER + (2 *(new_dw->primitive.margin_height));

    if( new_dw->core.width < min_width )
    {
	new_dw->core.width = min_width;
	redisplay = True;
    }

    if( new_dw->core.height < min_height )
    {
	new_dw->core.height = min_height;
	redisplay = True;
    }

    if( new_dw->primitive.margin_height != old_dw->primitive.margin_height ||
        new_dw->primitive.margin_width != old_dw->primitive.margin_width ||
        new_dw->dial.max_val != old_dw->dial.max_val ||
        new_dw->dial.min_val != old_dw->dial.min_val ||
        new_dw->core.height != old_dw->core.height ||
        new_dw->core.width != old_dw->core.width )
    {
	mkArrow( neww );
	redisplay = True;
    }

    return redisplay;
}

static XtGeometryResult hook_queryGeometry( Widget w,
    XtWidgetGeometry *request, XtWidgetGeometry *reply )
{
    LjpDialWidget dw = (LjpDialWidget)w;

    reply->width = dw->dial.orig_width;
    reply->height = dw->dial.orig_height;
    reply->request_mode = CWWidth|CWHeight;

    if( ((request->request_mode&CWWidth) &&
	request->width==reply->width) &&
	((request->request_mode & CWHeight) &&
	request->height==reply->height) )
    {
	return XtGeometryYes;
    }

    if( (reply->width==dw->core.width) &&
	(reply->height==dw->core.height) )
    {
	return XtGeometryNo; 
    }

    return XtGeometryAlmost; 
}

static void mkDiameter( Widget w, unsigned int *diameter,
   Dimension *padX, Dimension *padY )
{
    LjpDialWidget dw = (LjpDialWidget)w;
    
    if( (Dimension)(dw->core.height-(2*dw->primitive.margin_width)) >=
        (Dimension)(dw->core.width-(2*dw->primitive.margin_height)) )
    {
	*diameter = dw->core.width-(2*(dw->primitive.margin_width));
    }
    else
       *diameter = dw->core.height-(2*(dw->primitive.margin_height));

    if( *diameter < DIAL_MIN_DIAMETER )
	*diameter = DIAL_MIN_DIAMETER;

    if( padX != NULL )
    {
	*padX = dw->core.width;
	*padX -= dw->primitive.margin_width*2;
	*padX -= *diameter;
	*padX /= 2;
    }

    if( padY != NULL )
    {
	*padY = dw->core.height;
	*padY -= dw->primitive.margin_height*2;
	*padY -= *diameter;
	*padY /= 2;
    }

    dw->dial.angle_offset = (DIAL_RADIANS(DIAL_MAX_ANGLE)*dw->dial.value)/ 
 	                   (dw->dial.max_val-dw->dial.min_val);
}

static void mkArrow( Widget w )
{
    LjpDialWidget dw = (LjpDialWidget)w;
    unsigned int diameter;
    unsigned int radius, inner_radius;
    double angle, cosine, sine;
    Position cx, cy;
    Dimension padX, padY;
    XSegment *ptr;
    
    ptr = dw->dial.segments;
    
    mkDiameter( w, &diameter, &padX, &padY );

    cx = (diameter>>1) + dw->primitive.margin_width + padX;
    cy = (diameter>>1) + dw->primitive.margin_height + padY;

    angle = dw->dial.angle_offset;

    radius = ((diameter>>1)*80)/100;

    if( (inner_radius=((((diameter*50)/100)>>1)*110)/100) < 1 )
	inner_radius = 1;

    cosine = cos(angle);
    sine = sin(angle);
    ptr[0].x1 = (short)(cx+radius*sine);
    ptr[0].y1 = (short)(cy-radius*cosine);
    cosine = cos(angle+5);
    sine = sin(angle+5);
    ptr[0].x2  = (short)(cx+inner_radius*sine);
    ptr[0].y2 = (short)(cy-inner_radius*cosine);

    cosine = cos(angle-5);
    sine = sin(angle-5);
    ptr[1].x1 = ptr[0].x1;
    ptr[1].y1 = ptr[0].y1;
    ptr[1].x2  = (short)(cx+inner_radius*sine);
    ptr[1].y2 = (short)(cy-inner_radius*cosine);

    ptr[2].x1 = ptr[0].x2;
    ptr[2].y1 = ptr[0].y2;
    ptr[2].x2 = ptr[1].x2;
    ptr[2].y2 = ptr[1].y2;
}

static void timeout_turn( XtPointer closure, XtIntervalId *id )
{
    Widget w = (Widget)closure;
    LjpDialWidget dw = (LjpDialWidget)w;
    struct ljpGenericCallbackStruct cb;

    if( (dw->dial.value<dw->dial.max_val-1 && dw->dial.move_clockwise) ||
	(dw->dial.value > 1 && !dw->dial.move_clockwise) )
    {
	dw->dial.timer_id = XtAppAddTimeOut(
	    XtWidgetToApplicationContext(w),
	    (unsigned long)dw->dial.turn_delay,
	    timeout_turn, (XtPointer)dw );
	/*
	fprintf( stdout, "TMP dial add %x\n", dw->dial.timer_id );
	*/
    }
    else
    {
	dw->dial.timer_id = 0;
	dw->dial.turning = False;
    }

    draw_indicator( w, False );
    draw_arrow( w, False );

    if( dw->dial.move_clockwise )
	++dw->dial.value;
    else
	--dw->dial.value;

    if( dw->dial.activate_callback )
    {
	cb.reason = XljpCR_VALUE_CHANGED;
	cb.event = NULL;
	cb.value = dw->dial.value;
	XtCallCallbackList( w, dw->dial.activate_callback,
	    (XtPointer)&cb );
    }

    mkArrow( w );

    draw_indicator( w, True );
    draw_arrow( w, True );
}

static void draw_dial( Widget w )
{
    LjpDialWidget dw = (LjpDialWidget)w;
    unsigned int diameter, top_dial_start;
    Dimension padX, padY;

    mkDiameter( w, &diameter, &padX, &padY );

    XSetForeground( XtDisplay(dw), dw->primitive.gcTop,
	dw->dial.pixelDial );
    XFillArc( XtDisplay(w), XtWindow(w), dw->primitive.gcTop, 
	dw->primitive.margin_width+padX, dw->primitive.margin_height+padY,
	diameter, diameter, 0, 360*64 );
    XSetForeground( XtDisplay(dw), dw->primitive.gcTop,
	dw->primitive.pixelTopShadow );
    XDrawArc( XtDisplay(w), XtWindow(w), dw->primitive.gcTop, 
	dw->primitive.margin_width+padX, dw->primitive.margin_height+padY,
	diameter, diameter,
	40*64, 180*64 );
    XDrawArc( XtDisplay(w), XtWindow(w), dw->primitive.gcBottom, 
	dw->primitive.margin_width+padX, dw->primitive.margin_height+padY,
	diameter, diameter, 40*64, -180*64 );

    draw_indicator( w, True );
    draw_arrow( w, True );
}

static void draw_indicator( Widget w, Boolean flagDraw )
{
    LjpDialWidget dw = (LjpDialWidget)w;
    unsigned int diameter;
    unsigned int radius;
    double angle;
    Position cx, cy;
    GC gcOutline, gcIndicator;
    Dimension padX, padY;

    if( flagDraw )
    {
	XSetForeground( XtDisplay(dw), dw->primitive.gcTop,
	    dw->dial.pixelIndicator );
	gcOutline = dw->primitive.gcBottom;
	gcIndicator = dw->primitive.gcTop;
	XSetForeground( XtDisplay(dw), dw->primitive.gcTop,
	    dw->primitive.pixelTopShadow );
    }
    else
	gcOutline = gcIndicator = dw->primitive.gcClear;

    mkDiameter( w, &diameter, &padX, &padY );

    angle = (DIAL_RADIANS(DIAL_MAX_ANGLE)*dw->dial.value)/ 
	(dw->dial.max_val-dw->dial.min_val) ;

    cx = (diameter>>1)+dw->primitive.margin_width;
    cy = (diameter>>1)+dw->primitive.margin_height;

    radius = ((diameter>>1)*93)/100;

    diameter = (diameter*5)/100;
    if( diameter == 0 )
	diameter = 1;

    cx += (Position)(radius*sin(angle))-(diameter>>1);
    cy -= (Position)(radius*cos(angle))+(diameter>>1);

    XDrawArc( XtDisplay(w), XtWindow(w), gcOutline, 
	cx+padX, cy+padY, diameter, diameter, 0, 360*64 );

    XFillArc(XtDisplay(w), XtWindow(w), gcIndicator,
	cx+padX, cy+padY, diameter, diameter, 0, 64*360);
}

static void draw_arrow( Widget w, Boolean flagDraw )
{
    LjpDialWidget dw = (LjpDialWidget)w;
    GC gc1, gc2, gc3;

    /*
    ** Crude attempt at maintaing shadow as triangle rotates. For
    ** this to look more real, the each segment of the triangle
    ** must be broken into two segments.
    */
    if( !flagDraw )
	gc1 = gc2 = gc3 = dw->primitive.gcClear;
    else
    {
	if( dw->dial.angle_offset < 0.75 )
	{
	    gc1 = dw->primitive.gcBottom;
	    gc2 = dw->primitive.gcTop;
	    gc3 = dw->primitive.gcTop;
	}
	else if( dw->dial.angle_offset < 1.5 )
	{
	    gc1 = dw->primitive.gcBottom;
	    gc2 = dw->primitive.gcTop;
	    gc3 = dw->primitive.gcBottom;
	}
	else if( dw->dial.angle_offset < 2.25 )
	{
	    gc1 = dw->primitive.gcTop;
	    gc2 = dw->primitive.gcTop;
	    gc3 = dw->primitive.gcBottom;
	}
	else if( dw->dial.angle_offset < 4.5 )
	{
	    gc1 = dw->primitive.gcTop;
	    gc2 = dw->primitive.gcBottom;
	    gc3 = dw->primitive.gcBottom;
	}
	else if( dw->dial.angle_offset < 5.25 )
	{
	    gc1 = dw->primitive.gcTop;
	    gc2 = dw->primitive.gcBottom;
	    gc3 = dw->primitive.gcTop;
	}
	else
	{
	    gc1 = dw->primitive.gcBottom;
	    gc2 = dw->primitive.gcBottom;
	    gc3 = dw->primitive.gcTop;
	}
    }

    XDrawSegments( XtDisplay(w), XtWindow(w), gc1,
	dw->dial.segments, 1 );
    XDrawSegments( XtDisplay(w), XtWindow(w), gc2,
	&dw->dial.segments[1], 1 );
    XDrawSegments( XtDisplay(w), XtWindow(w), gc3,
	&dw->dial.segments[2], 1 );
}

static void action_left( Widget w, XEvent *event, char **params,
    Cardinal *num_params )
{
    LjpDialWidget dw = (LjpDialWidget)w;
    
    if( dw->dial.value > dw->dial.min_val )
    {
	dw->dial.move_clockwise = False;
	dw->dial.timer_id = XtAppAddTimeOut(
	    XtWidgetToApplicationContext(w),
	    (unsigned long)dw->dial.turn_delay,
	    timeout_turn, (XtPointer)w );
	/*
	fprintf( stdout, "TMP dial add %x\n", dw->dial.timer_id );
	*/
    }
}

static void action_right( Widget w, XEvent *event, char **params,
    Cardinal *num_params )
{
    LjpDialWidget dw = (LjpDialWidget)w;

    if( dw->dial.value < dw->dial.max_val )
    {
	dw->dial.move_clockwise = True;
	dw->dial.timer_id = XtAppAddTimeOut(
	    XtWidgetToApplicationContext(w),
	    (unsigned long)dw->dial.turn_delay,
	    timeout_turn, (XtPointer)w );
	/*
	fprintf( stdout, "TMP dial add %x\n", dw->dial.timer_id );
	*/
    }
}

static void action_stop( Widget w, XEvent *event, char **params,
    Cardinal *num_params )
{
    LjpDialWidget dw = (LjpDialWidget)w;

    if( dw->dial.timer_id )
    {
	XtRemoveTimeOut( dw->dial.timer_id );
	dw->dial.timer_id = 0;
    }

    dw->dial.turning = False;
}
