/*
** 
**  Copyright (c) 1998 by Larry Plona
** 
**  This distribution  of XPipeman is copyright by Larry Plona.
**  The  package  and  its  component  files  may  be  freely
**  distributed as long as  all copyright notices remain intact
**  and this distribution is not sold  as  commercial software.
** 
**  No warranty of any kind is attached with this software.  It
**  is supplied as is.
** 
** ----------------------------------------------------------------
*/

#include <stdio.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include "ScrollbarP.h"

/*
** Widget hooks.
*/
static void hook_initialize( Widget, Widget );
static Boolean hook_setValues( Widget, Widget, Widget, ArgList, Cardinal* );
static void hook_resize( Widget );
static void hook_redisplay( Widget, XEvent*, Region );
static XtGeometryResult hook_queryGeometry( Widget, XtWidgetGeometry*, 
    XtWidgetGeometry* );
static void hook_destroy( Widget );

static void action_press( Widget, XEvent*, char**, Cardinal* );
static void action_release( Widget, XEvent*, char**, Cardinal* );
static void action_slide( Widget, XEvent*, char**, Cardinal* );

static void layout( LjpScrollbarWidget, Boolean );
static void draw_scrollbar( LjpScrollbarWidget );
static print_request( const char*, XtWidgetGeometry* );
static void mkTrough( LjpScrollbarWidget, Position*, Position*,
   Dimension*, Dimension* );

static XtResource resources[] =
{
    {
	XljpNorientation,
	XljpCOrientation,
	XtRInt,
	sizeof(int),
	XtOffsetOf(LjpScrollbarRec,scrollbar.orientation),
	XtRImmediate,
	(XtPointer)XljpOrientVertical
    },
    {
	XljpNactivateCallback,
	XljpCActivateCallback,
	XtRCallback,
	sizeof(XtCallbackList),
	XtOffsetOf(LjpScrollbarRec,scrollbar.activate_callback),
	XtRImmediate, NULL
    },
    {
	XljpNsliderSize,
	XljpCSliderSize,
	XtRDimension,
	sizeof(Dimension),
	XtOffsetOf(LjpScrollbarRec,scrollbar.slider_size),
	XtRImmediate,
	(XtPointer)5
    },
    {
	XljpNvalue,
	XljpCValue,
	XtRInt,
	sizeof(int),
	XtOffsetOf(LjpScrollbarRec,scrollbar.value),
	XtRImmediate,
	(XtPointer)0
    }
};

static XtActionsRec hook_actions[] =
{
    { "press",		action_press },
    { "release",	action_release },
    { "slide",		action_slide }
};

static char hook_translations[] = "\
<Btn1Down>:	press()\n\
<Btn1Up>:	release()\n\
<Btn1Motion>:	slide()";
	  
LjpScrollbarClassRec ljpScrollbarClassRec =
{
   {
      (WidgetClass) &ljpPrimitiveClassRec,      /* superclass         */
      "LjpScrollbar",                         	/* class_name         */
      sizeof(LjpScrollbarRec),                	/* widget_size        */
      NULL,                          		/* class_initialize   */
      NULL,                            		/* class_part_initiali*/
      FALSE,                                    /* class_inited       */
      hook_initialize,                          /* initialize         */
      NULL,                                     /* initialize_hook    */
      XtInheritRealize,                         /* realize            */
      hook_actions,				/* actions            */
      XtNumber(hook_actions),			/* num_actions        */
      resources,                                /* resources          */
      XtNumber(resources),                      /* num_resources      */
      NULLQUARK,                                /* xrm_class          */
      TRUE,                                     /* compress_motion    */
      XtExposeCompressMaximal,                  /* compress_exposure  */
      TRUE,                                     /* compress_enterleave*/
      FALSE,                                    /* visible_interest   */
      hook_destroy,                             /* destroy            */
      hook_resize,                              /* resize             */
      hook_redisplay,	              		/* expose             */
      hook_setValues,                           /* set_values         */
      NULL,                                     /* set_values_hook    */
      NULL,                            		/* set_values_almost  */
      NULL,                                     /* get_values_hook    */
      NULL,                                     /* accept_focus       */
      XtVersion,                                /* version            */
      NULL,                                     /* callback_private   */
      hook_translations,			/* tm_table           */
      hook_queryGeometry,             		/* query_geometry     */
      XtInheritDisplayAccelerator,              /* display accel      */
      NULL,                                     /* extension          */
    },
    /* Primitive class. */
    {
	XljpInheritDrawFrame,			/* draw_frame.		*/
	NULL                                    /* extension		*/
    },
    /* Scrollbar. */
    {
	NULL                                    /* extension          */
    }
};

WidgetClass ljpScrollbarWidgetClass = (WidgetClass)&ljpScrollbarClassRec;

static void hook_initialize( Widget request, Widget new )
{
    LjpScrollbarWidget swnew = (LjpScrollbarWidget)new;

    /*
    fprintf( stdout, "LjpScrollbar 0x%x hook initialize\n", new );
    */

    swnew->primitive.shadow_type = XljpShadowIn;

    if( swnew->core.width == 0 )
        swnew->core.width = 24;

    if( swnew->core.height == 0 )
        swnew->core.height = 100;

    swnew->scrollbar.sliding = False;
    swnew->scrollbar.value = 0;

    mkTrough( swnew, &swnew->scrollbar.xSlider,
	&swnew->scrollbar.ySlider, NULL, NULL );
}

static void hook_redisplay( Widget w, XEvent *event, Region region )
{
    LjpScrollbarWidget sw = (LjpScrollbarWidget)w;

    /*
    fputs( "LjpScrollbar hook redisplay\n", stdout );
    */
    draw_scrollbar( sw );
}

static Boolean hook_setValues( Widget current,  Widget request,  Widget new,
    ArgList args, Cardinal *num_args )
{
    Position x, y, nx, ny;
    Dimension w, h;
    LjpScrollbarWidget swcurr = (LjpScrollbarWidget)current;
    LjpScrollbarWidget swnew = (LjpScrollbarWidget)new;

    if( swnew->scrollbar.value != swcurr->scrollbar.value )
    {
	mkTrough( swnew, &x, &y, &w, &h );
	if( swnew->scrollbar.orientation == XljpOrientHorizontal )
	{
	    nx = x + swnew->scrollbar.value;
	    if( nx > (x+(w-swnew->scrollbar.slider_size)) )
		nx = w-swnew->scrollbar.slider_size;
	    if( nx < x )
		nx = x;
	}
	else
	{
	    ny = y + swnew->scrollbar.value;
	    if( ny > (y+(w-swnew->scrollbar.slider_size)) )
		ny = w-swnew->scrollbar.slider_size;
	    if( ny < y )
		ny = y;
	}
	draw_scrollbar( swnew );
    }

    /*
    ** Size changes override value changes.
    */
    if( (swnew->core.width!=swcurr->core.width) ||
	(swnew->core.height!=swcurr->core.height) ||
	(swnew->scrollbar.slider_size!=swcurr->scrollbar.slider_size) )
    {
	/*
	if( swnew->scrollbar.slider_size!=swcurr->scrollbar.slider_size )
	    fputs( "LjpScrollbar setValues change in slider size.\n", stdout );
	*/
        layout( swnew, True );
    }

    return False;
}

static void hook_destroy( Widget w )
{
    LjpScrollbarWidget sbw = (LjpScrollbarWidget)w;

    if( sbw->scrollbar.activate_callback != NULL )
	XtRemoveAllCallbacks( w, XljpNactivateCallback );
}

static void hook_resize( Widget w )
{
    LjpScrollbarWidget sbw = (LjpScrollbarWidget)w;

    layout( sbw, True );
}

static XtGeometryResult hook_queryGeometry( Widget w,
    XtWidgetGeometry *request, XtWidgetGeometry *reply_return )
{
    Boolean flagAlmost = False;
    LjpScrollbarWidget sbw = (LjpScrollbarWidget)w;
    XtGeometryResult result = XtGeometryYes;

    request->request_mode &= CWWidth | CWHeight;

    if( request->request_mode == 0 )
	return XtGeometryYes;

    if( request->request_mode & CWHeight )
    {
	if( (sbw->scrollbar.orientation==XljpOrientVertical) &&
	    (request->height<(sbw->scrollbar.slider_size+
	    (2*sbw->primitive.margin_height)+
	    (2*sbw->primitive.shadow_thickness))) )
	{
            flagAlmost = True;
            reply_return->height = (request->height<
		(sbw->scrollbar.slider_size+(2*sbw->primitive.margin_height)+
		(2*sbw->primitive.shadow_thickness)));
            reply_return->request_mode &= CWHeight;
	}

	if( (sbw->scrollbar.orientation==XljpOrientHorizontal) &&
	    (request->height<10) )
	{
            flagAlmost = True;
            reply_return->height = 10;
            reply_return->request_mode &= CWHeight;
        }
	else if( (sbw->scrollbar.orientation==XljpOrientVertical) &&
	    (request->height<30) )
	{
            flagAlmost = True;
            reply_return->height = 30;
            reply_return->request_mode &= CWHeight;
	}
        else
            result = XtGeometryYes;
    }

    if( request->request_mode & CWWidth )
    {
	if( (sbw->scrollbar.orientation==XljpOrientHorizontal) &&
	    (request->width<(sbw->scrollbar.slider_size+
	    (2*sbw->primitive.margin_width)+
	    (2*sbw->primitive.shadow_thickness))) )
	{
            flagAlmost = True;
            reply_return->height = (request->width<
		(sbw->scrollbar.slider_size+(2*sbw->primitive.margin_width)+
		(2*sbw->primitive.shadow_thickness)));
            reply_return->request_mode &= CWWidth;
	}
	if( (sbw->scrollbar.orientation==XljpOrientHorizontal) &&
	    (request->width<30) )
	{
            flagAlmost = True;
            reply_return->width = 30;
            reply_return->request_mode &= CWWidth;
        }
	if( (sbw->scrollbar.orientation==XljpOrientVertical) &&
	    (request->width<10) )
	{
            flagAlmost = True;
            reply_return->width = 10;
            reply_return->request_mode &= CWWidth;
	}
        else
            result = XtGeometryYes;
    }

    if( flagAlmost )
	return XtGeometryAlmost;

    return result;
}

static void draw_scrollbar( LjpScrollbarWidget sw )
{
    LjpPrimitiveWidgetClass fc = (LjpPrimitiveWidgetClass)XtClass(sw);
    Position x, y, xTrough, yTrough;
    Dimension w, h;

    /*
    fputs( "LjpScrollbar draw\n", stdout );
    fflush( stdout );
    */

    mkTrough( sw, &x, &y, &w, &h );

    /* Clear trough. */
    XFillRectangle( XtDisplay(sw), XtWindow((Widget)sw),
	sw->primitive.gcClear, x, y, w, h );

    x = sw->scrollbar.xSlider;
    y = sw->scrollbar.ySlider;

    if( sw->scrollbar.orientation == XljpOrientHorizontal )
	w = sw->scrollbar.slider_size;
    else
	h = sw->scrollbar.slider_size;

    XFillRectangle( XtDisplay(sw), XtWindow((Widget)sw),
	sw->primitive.gcTop, x, y, w, h );
    XFillRectangle( XtDisplay(sw), XtWindow((Widget)sw),
	sw->primitive.gcBottom, x+1, y+1, w-1, h-1 );

    XSetForeground( XtDisplay(sw), sw->primitive.gcClear,
	sw->primitive.pixelForeground );
    XFillRectangle( XtDisplay(sw), XtWindow((Widget)sw),
	sw->primitive.gcClear, x+1, y+1, w-2, h-2 );
    XSetForeground( XtDisplay(sw), sw->primitive.gcClear,
	sw->core.background_pixel );

    fc->primitive_class.draw_frame( (Widget)sw );
}

static void layout( LjpScrollbarWidget sw, Boolean doit )
{
    Position xTr, yTr;
    Dimension wTr, hTr;
    XtGeometryResult rc;
    XtWidgetGeometry request, reply;

    if( doit == False )
    {
	mkTrough( sw, &xTr, &yTr, &wTr, &hTr );
	return;
    }

    request.request_mode = CWWidth;
    request.width = sw->core.width;
    rc = XtMakeGeometryRequest( (Widget)sw, &request, &reply );
    /*
    fprintf( stdout, "LjpScrollbar layout width req %d\n", request.width );
    */
    if( rc != XtGeometryYes )
    {
	fputs( "LjpScrollbar geom width request denied\n", stdout );
    }

    request.request_mode = CWHeight;
    request.height = sw->core.height;
    rc = XtMakeGeometryRequest( (Widget)sw, &request, &reply );
    /*
    fprintf( stdout, "LjpScrollbar layout height req %d\n", request.height );
    */
    if( rc != XtGeometryYes )
    {
	fputs( "LjpScrollbar geom height request denied\n", stdout );
    }

    /*
    fprintf( stdout, "LjpScrollbar layed out x%d y%d w%d h%d\n",
	sw->core.x, sw->core.y, sw->core.width, sw->core.height );
    */
}

static void action_press( Widget w, XEvent *event, char **params,
    Cardinal *num_params )
{
    XButtonEvent *be = (XButtonEvent*)event;
    LjpScrollbarWidget sw = (LjpScrollbarWidget)w;
    Dimension wTrough, hTrough;

    mkTrough( sw, NULL, NULL, &wTrough, &hTrough );

    /*
    fprintf( stdout, "LjpScrollbar slider x%d y%d cursor x%d y%d\n",
	sw->scrollbar.xSlider, sw->scrollbar.ySlider,
	be->x, be->y );
    */

    /*
    ** Verify that the press is within the slider.
    */
    if( (be->x<sw->scrollbar.xSlider) ||
	(be->y<sw->scrollbar.ySlider) )
    {
	return;
    }
    if( sw->scrollbar.orientation == XljpOrientHorizontal )
    {
	if( be->x>(sw->scrollbar.xSlider+sw->scrollbar.slider_size) )
	{
	    return;
	}
	if( be->y>(sw->scrollbar.ySlider+hTrough) )
	{
	    return;
	}
    }
    else
    {
	if( be->y>(sw->scrollbar.ySlider+sw->scrollbar.slider_size) )
	{
	    return;
	}
	if( be->x>(sw->scrollbar.xSlider+wTrough) )
	{
	    return;
	}
    }

    sw->scrollbar.sliding = True;
    sw->scrollbar.xPress = be->x;
    sw->scrollbar.yPress = be->y;
    /*
    fprintf( stdout, "LjpScrollbar press x%d y%d\n", be->x, be->y );
    */
}

static void action_release( Widget w, XEvent *event, char **params,
    Cardinal *num_params )
{
    LjpScrollbarWidget sw = (LjpScrollbarWidget)w;

    /*
    fprintf( stdout, "LjpScrollbar release\n" );
    */
    sw->scrollbar.sliding = False;
}

static void action_slide( Widget w, XEvent *event, char **params,
    Cardinal *num_params )
{
    LjpScrollbarWidget sw = (LjpScrollbarWidget)w;
    XMotionEvent *me = (XMotionEvent*)event;
    struct ljpGenericCallbackStruct rec;
    Position xTrough, yTrough;
    Dimension wTrough, hTrough;
    Dimension delta;

    mkTrough( sw, &xTrough, &yTrough, &wTrough, &hTrough );

    if( sw->scrollbar.sliding != True )
	return;

    if( sw->scrollbar.orientation == XljpOrientHorizontal )
    {
	delta = me->x - sw->scrollbar.xPress;
	if( delta == 0 )
	    return;
	sw->scrollbar.xSlider += delta;
	if( sw->scrollbar.xSlider < xTrough )
	{
	    sw->scrollbar.xSlider -= delta;
	    return;
	}
	if( (sw->scrollbar.xSlider+sw->scrollbar.slider_size) >
	    (xTrough+wTrough) )
	{
	    sw->scrollbar.xSlider -= delta;
	    return;
	}
	sw->scrollbar.value = sw->scrollbar.xSlider - xTrough;
    }
    else
    {
	delta = me->y - sw->scrollbar.yPress;
	if( delta == 0 )
	    return;
	sw->scrollbar.ySlider += delta;
	if( sw->scrollbar.ySlider < yTrough )
	{
	    sw->scrollbar.ySlider -= delta;
	    return;
	}
	if( (sw->scrollbar.ySlider+sw->scrollbar.slider_size) >
	    (yTrough+hTrough) )
	{
	    sw->scrollbar.ySlider -= delta;
	    return;
	}
	sw->scrollbar.value = sw->scrollbar.ySlider - yTrough;
    }

    sw->scrollbar.xPress = me->x;
    sw->scrollbar.yPress = me->y;

    draw_scrollbar( sw );

    rec.reason = XljpCR_VALUE_CHANGED;
    rec.event = event;
    rec.value = sw->scrollbar.value;
    if( sw->scrollbar.activate_callback )
    {
	XtCallCallbackList( (Widget)sw, sw->scrollbar.activate_callback,
	    &rec );
    }
}

static void mkTrough( LjpScrollbarWidget sw, Position *xTrough,
    Position *yTrough, Dimension *wTrough, Dimension *hTrough )
{
    Position x, y;
    Dimension w, h;

    x = sw->primitive.margin_width;
    y = sw->primitive.margin_height;
    w = sw->core.width - (2*sw->primitive.margin_width);
    h = sw->core.height - (2*sw->primitive.margin_height);

    switch( sw->primitive.shadow_type )
    {
	case XljpShadowRelief:
	case XljpShadowEngrave:
	    x += 3*sw->primitive.shadow_thickness;
	    h += 3*sw->primitive.shadow_thickness;
	    w -= 6*sw->primitive.shadow_thickness;
	    h -= 6*sw->primitive.shadow_thickness;
	    break;
	default:
	    x += sw->primitive.shadow_thickness;
	    y += sw->primitive.shadow_thickness;
	    w -= 2*sw->primitive.shadow_thickness;
	    h -= 2*sw->primitive.shadow_thickness;
	    break;
    }

    if( xTrough != NULL )
	*xTrough = x;
    if( yTrough != NULL )
	*yTrough = y;
    if( wTrough != NULL )
	*wTrough = w;
    if( hTrough != NULL )
	*hTrough = h;
}
