/*
** 
**  Copyright (c) 1998 by Larry Plona
** 
**  This distribution  of XPipeman is copyright by Larry Plona.
**  The  package  and  its  component  files  may  be  freely
**  distributed as long as  all copyright notices remain intact
**  and this distribution is not sold  as  commercial software.
** 
**  No warranty of any kind is attached with this software.  It
**  is supplied as is.
** 
** ----------------------------------------------------------------
*/

#include "artwork.h"
#include "xpipeman.h"

/*
** These must align with the ARTWORK_SND_* constants.
*/
static const char *tags[] =
{
    "null",
    "levelover",
    "badplace",
    "cross",
    "flowcnt",
    "flowstart",
    "place",
    "gameover",
    "shift",
    NULL
};

/*
** Constructor.
**
** o Load colormap into global Colores from glob data.
** o Set X11 cursor from glob data.
** o Register obstacle TObj ids.
** o Register pipe TObj ids.
** o Register sounds from glob data.
*/
Artwork::Artwork( const char *name, int desired_level )
{
    const char *buf;
    char line[1023+1];
    int buflen;
    int i;
    struct intlevel *il;

    obscnt = 0;

    glob = new Glob( name );
    if( !(glob->initialized()) )
    {
	fprintf( stdout, "ERROR. Failed to open glob \"%s\".\n", name );
	delete glob;
	glob = NULL;
	return;
    }

    /* Read config component from glob. Parse TObj and sounds names. */
    if( (buf=glob->extract("conf",&buflen)) == NULL )
    {
	fprintf( stdout, "ERROR. Malformed glob \"%s\".\n", name );
	delete glob;
	glob = NULL;
	return;
    }
    level = desired_level;
    if( (il=parse_conf(buf,buflen)) == NULL )
    {
	fputs( "ERROR. Can not parse conf component.\n", stdout );
	delete glob;
	glob = NULL;
	return;
    }

    if( (buf=glob->extract("colormap.xpm",&buflen)) != NULL )
    {
	if( !(xface.colores->load(buf)) )
	{
	    fputs( "ERROR. Could not load colormap.\n", stdout );
	    delete glob;
	    free( (void*)buf );
	    glob = NULL;
	    return;
	}
	free( (void*)buf );
    }

    buf = glob->extract("cursor.xpm",&buflen);
    (void)xfaceSetCursor(buf);
    if( buf != NULL )
	free( (void*)buf );

    for( i=0; i<il->obscnt; ++i )
    {
	if( (obstacles[i]=tcache.notify(il->obstacles[i])) == -1 )
	{
	    fputs( "ERROR. Failed to register obstacle.\n", stdout );
	    free_intlevel( il );
	    delete glob;
	    glob = NULL;
	    return;
	}
    }
    obscnt = il->obscnt;

    for( i=0; i<il->pipecnt; ++i )
    {
	if( (pipes[i]=tcache.notify(il->pipes[i])) == -1 )
	{
	    free_intlevel( il );
	    delete glob;
	    glob = NULL;
	    return;
	}
    }
    pipecnt = il->pipecnt;

    if( flagSound )
	sfx_glob( name );
    for( i=0; i<ARTWORK_MAX_SOUNDS; ++i )
    {
	if( il->sounds[i] == NULL )
	    continue;
	sprintf( line, "G:%s", il->sounds[i] );
	if( flagSound )
	    sfx_link( (i+ARTWORK_SND_FIRST)-1, line, 0 );
    }
    if( flagSound )
    {
	sfx_flags( ARTWORK_SND_PLACE, SFX_SFLG_KEEP );
	sfx_flags( ARTWORK_SND_CROSS, SFX_SFLG_KEEP );
	sfx_flags( ARTWORK_SND_BADPLACE, SFX_SFLG_KEEP );
    }

    free_intlevel( il );
}

Artwork::~Artwork()
{
    int i;

    if( glob != NULL )
	delete glob;
    (void)xfaceSetCursor( NULL );
    if( flagSound )
    {
	for( i=ARTWORK_SND_FIRST; i<ARTWORK_MAX_SOUNDS; ++i )
	    sfx_delete( i );
    }
}

struct intlevel *Artwork::parse_conf( const char *buf, int len )
{
    const char *ip, *ep;
    char *p;
    char line[ARTWORK_MAX_CONFLINE+1];
    int lineno = 0;
    int flagStanza = ARTWORK_STANZA_NULL;
    int max, remain;
    int obscnt, pipecnt, synccnt;
    struct intlevel *il;

    if( (il=(struct intlevel*)malloc(sizeof(struct intlevel))) == NULL )
    {
	fprintf( stdout, "ERROR. Bad malloc. %s.\n", sys_errlist[errno] );
	return NULL;
    }

    for( obscnt=0; obscnt<ARTWORK_MAX_SOUNDS; ++obscnt )
	il->sounds[obscnt] = ARTWORK_SND_NULL;
    obscnt = pipecnt = synccnt = 0;

    ep = (buf+len)-1;

    for( ip=buf; ip<=ep; )
    {
	++lineno;

	memset( line, (int)'\0', ARTWORK_MAX_CONFLINE+1 );

	/*
	** Copy one line from buffer ( sgets()! ).
	*/
	remain = (ep-ip)+1;
	max = ARTWORK_MAX_CONFLINE>remain ? remain : ARTWORK_MAX_CONFLINE;
	if( (p=(char*)memccpy(line,ip,(int)'\n',max)) == NULL )
	{
	    fprintf( stdout,
		"ERROR. Conf line %d too long or unterminated.\n",
		lineno );
	    free_intlevel( il );
	    return NULL;
	}

	/* Remove newline in copy. */
	*(p-1) = '\0';

	ip += (p-line); /* For next "sgets()". */

	/* Ignore leading whitespace. */
	for( p=line; isspace(*p); ++p );
	if( *p == '\0' )
	    continue;

	if( p[0] == '#' )
	    continue;

	if( p[0] == '%' )
	{
	    if( !strncmp(p,"%obstacles",10) )
	    {
		flagStanza = ARTWORK_STANZA_OBSTACLES;
		continue;
	    }
	    if( !strncmp(p,"%pipes",6) )
	    {
		flagStanza = ARTWORK_STANZA_PIPES;
		continue;
	    }
	    if( !strncmp(p,"%syncs",6) )
	    {
		flagStanza = ARTWORK_STANZA_SYNCS;
		continue;
	    }
	    if( !strncmp(p,"%sounds",7) )
	    {
		flagStanza = ARTWORK_STANZA_SOUNDS;
		continue;
	    }
	    continue;
	}

	switch( flagStanza )
	{
	    case ARTWORK_STANZA_OBSTACLES:
	    {
		if( obscnt >= ARTWORK_MAX_OBSTACLES )
		{
		    fputs( "WARNING. Too many obstacle tiles.\n", stdout );
		    continue;
		}

		if( (p=parse_identifier(p)) == NULL )
		    break;
		il->obstacles[obscnt] = strdup( p );
		//fprintf( stdout, "TMP obs \"%s\".\n", il->obstacles[obscnt] );
		++obscnt;
		break;
	    }
	    case ARTWORK_STANZA_PIPES:
	    {
		char bufFile[ARTWORK_MAX_CONFLINE+1];
		char bufLev[ARTWORK_MAX_CONFLINE+1];
		int val;

		if( pipecnt >= ARTWORK_MAX_PIPES )
		{
		    fputs( "WARNING. Too many pipe tiles.\n", stdout );
		    continue;
		}

		if( (p=parse_identifier(p)) == NULL )
		    break;

		if( sscanf(p,"%s %s",bufFile,bufLev) != 2 )
		{
		    fprintf( stdout,
			"WARNING. Malformed line, %d.\n", lineno );
		    break;
		}

		if( (bufLev[0]=='-') && (bufLev[1]=='-') )
		    val = -1;
		else
		    val = atoi(bufLev);

		if( val > (level+1) )
		    break;

		il->pipes[pipecnt] = strdup( bufFile );
		/*
		fprintf( stdout, "TMP pipe \"%s\".\n",
		    il->pipes[pipecnt] );
		*/
		++pipecnt;
		break;
	    }
	    case ARTWORK_STANZA_SYNCS:
	    {
		if( synccnt >= ARTWORK_MAX_SYNCS )
		{
		    fputs( "WARNING. Too many syncs.\n", stdout );
		    continue;
		}

		if( (p=parse_identifier(p)) == NULL )
		    break;
		if( (syncs[synccnt]=atoi(p)) < 1 )
		{
		    //fprintf( stdout, "TMP sync \"%s\".\n", p );
		    continue;
		}
		++synccnt;
		break;
	    }
	    case ARTWORK_STANZA_SOUNDS:
	    {
		char bufFile[ARTWORK_MAX_CONFLINE+1];
		char bufTag[ARTWORK_MAX_CONFLINE+1];
		int tag;

		if( (p=parse_identifier(p)) == NULL )
		    break;

		if( sscanf(p,"%s %s",bufFile,bufTag) != 2 )
		{
		    fprintf( stdout,
			"WARNING. Malformed line, %d.\n", lineno );
		    break;
		}

		if( (tag=nameToTag(bufTag)) == ARTWORK_SND_NULL )
		{
		    fprintf( stdout,
			"WARNING. Unrecognized sound, \"%s\".\n", bufTag );
		    break;
		}

		il->sounds[tag] = strdup( bufFile );
		/*
		fprintf( stdout, "TMP sound %d \"%s\".\n",
		    tag, il->sounds[tag] );
		*/
		break;
	    }
	    case ARTWORK_STANZA_NULL:
	    default:
		break;
	}
    }

    il->obscnt = obscnt;
    il->pipecnt = pipecnt;

    return il;
}

char *Artwork::parse_identifier( char *nline )
{
    char *p, *p1;
    char line[ARTWORK_MAX_CONFLINE+1];

    memset( line, (int)'\0', ARTWORK_MAX_CONFLINE+1 );
    // Skip leading whitespace.
    for( p=nline; isspace(*p); ++p );
    // Ignore empty lines.
    if( (*p=='\0') || (*p=='\n') )
	return NULL;
    strncpy( line, p, ARTWORK_MAX_CONFLINE );
    // Remove trailing whitespace.
    for( p1=&nline[ARTWORK_MAX_CONFLINE]; p1>nline; --p1 )
    {
	if( (*p1=='\0') || (isspace(*p1)) || (*p1=='\n') )
	{
	    *p1 = '\0';
	    continue;
	}
	break;
    }
    return p;
}

void Artwork::free_intlevel( struct intlevel *il )
{
    int i;

    for( i=0; i<il->pipecnt; ++i )
	free( il->pipes[i] );
    for( i=0; i<il->pipecnt; ++i )
	free( il->pipes[i] );
    for( i=0; i<il->obscnt; ++i )
	free( il->obstacles[i] );
    for( i=0; i<ARTWORK_MAX_SOUNDS; ++i )
	free( il->sounds[i] );
    free( il );
}

int Artwork::initialized( void )
{
    return (glob==NULL) ? 0 : 1;
}

int Artwork::nameToTag( const char *strTag )
{
    int i;

    for( i=0; tags[i]!=NULL; ++i )
    {
	if( !strcmp(strTag,tags[i]) )
	    return i;
    }
    return ARTWORK_SND_NULL;
}
