/*
** 
**  Copyright (c) 1998 by Larry Plona
** 
**  This distribution  of XPipeman is copyright by Larry Plona.
**  The  package  and  its  component  files  may  be  freely
**  distributed as long as  all copyright notices remain intact
**  and this distribution is not sold  as  commercial software.
** 
**  No warranty of any kind is attached with this software.  It
**  is supplied as is.
** 
** ----------------------------------------------------------------
*/

#include "board.h"

/*
** Constructor.
**
** Set positions of all tiles on screen. The Tiles have not yet been
** assigned TObjs, but the dimensions of each are known in advance.
*/
Board::Board( Display *ndisplay, XVisualInfo *nvinfo, Window nwin,
   int nlocality, TCache *ntcache, Colores *ncolores )
{
    int c, r;

    display = ndisplay;
    vinfo = nvinfo;
    win = nwin;
    locality = nlocality;
    colores = ncolores;
    tcache = ntcache;

    effect = NULL;
    mode = board_mode_title;

    // Set the coordinates for the playfield tiles.
    for( r=0; r<BOARD_HEIGHT; ++r )
    {
	for( c=0; c<BOARD_WIDTH; ++c )
	    (playfield[c][r]).setpos( c*40, r*40, 40, 40 );
    }

    // Set coordinates of tile deck.
    for( r=0; r<10; ++r )
	deck[r].setpos( (BOARD_WIDTH*40)+40, 360-(r*40), 40, 40 );

    // Set coordinates of next tile.
    ntile.setpos( (12*40), 360, 40, 40 );

    // Set coordinates for the flow meter.
    for( r=0; r<3; ++r )
	meter[r].setpos( (BOARD_WIDTH*40), 160-(r*40), 40, 40 );

    // Set coordinates of misc. tiles.
    tbar.setpos( BOARD_WIDTH*40, 0, 40, 80 );
    bbar.setpos( (BOARD_WIDTH*40), 80+(40*3), 40, 160 );
    rbar.setpos( (BOARD_WIDTH*40)+40+40, 0, 40, 400 );

    imago = new Imago( display, vinfo, win, locality );
    imago->resize( (BOARD_WIDTH*40)+(3*40), BOARD_HEIGHT*40 );

    animList = new AnimList( imago );
}

/*
** Destructor.
*/
Board::~Board()
{
    /*
    fputs( "TMP. Board destructor.\n", stdout ); fflush( stdout );
    */
    if( imago != NULL )
	delete imago;
    if( animList != NULL )
	delete animList;
}

/*
** Board::title
**
** Place board in title mode. Use given Xpm buffer as an image to display.
** Begin palette shift. Set animation timeout.
*/
int Board::title( const char *buf )
{
    int rc;
    XImage *fimg;

    externSetAnim( 0, this );

    mode = board_mode_title;
    if( effect != NULL )
    {
	free( effect );
	effect = NULL;
    }

    // Use given Xpm as a colormap.
    if( !colores->load(buf) )
	return 0;

    if( (rc=colores->createImageFromBuffer((char*)buf,&fimg,
	NULL,NULL)) != XpmSuccess )
    {
	fprintf( stdout, "ERROR. Bad XpmCreateImageFromBuffer. \"%s\".\n",
	    XpmGetErrorString(rc) );
	free( (void*)buf );
	return 0;
    }
    free( (void*)buf );

    imago->copy( fimg, 0, 0, 600, 400 );
    imago->flush();
    XDestroyImage( fimg );

    /*
    ** FUTURE. The starting and ending pixels should be determined
    ** from the Xpm file.
    */
    if( vinfo->c_class == PseudoColor )
    {
	struct colores_study_ramp rec;

	rec.pixelBegin = 6;
	rec.pixelEnd = 16;
	rec.seed = 1000;
	rec.maxsteps = 10;
	rec.flagBegin = COLORES_RAMP_KEEP;
	rec.flagRamp = COLORES_RAMP_STATIC;
	if( (effect=(void*)colores->effectRampInit(&rec)) == NULL )
	{
	    fputs( "ERROR. Could not start Ramp effect.\n", stdout );
	    return 0;
	}
    }

    externSetAnim( 100, this );

    return 1;
}

/*
** Board::pause
**
** Place board in pause mode.
*/
int Board::pause( void )
{
    if( mode != board_mode_animate )
	return 0;

    externSetAnim( 0, this );

    mode = board_mode_pause;
    if( effect != NULL )
    {
	free( effect );
	effect = NULL;
    }

    imago->back( 0x03 );
    imago->back( 0x05 );

    imago->clear( 0, 0, 0, 600, 400 );

    /*
    ** The greater the depth, the larger the consumption of bandwidth
    ** used by the Stellae effect.
    */
    if( vinfo->c_class == PseudoColor )
    {
	if( (effect=(void*)imago->effectStellaeInit(100,100,400,200))
	    == NULL )
	{
	    fputs( "ERROR. Could not start Stellae effect.\n", stdout );
	    return 0;
	}
    }
    else
    {
	if( (effect=(void*)imago->effectStellaeInit(150,150,300,100))
	    == NULL )
	{
	    fputs( "ERROR. Could not start Stellae effect.\n", stdout );
	    return 0;
	}
    }

    imago->flush();

    externSetAnim( 100, this );

    return 1;
}

void Board::reset( int sync_interval )
{
    int c, r;

    externSetAnim( 0, this );

    if( mode == board_mode_pause )
    {
	imago->effectStellaeFinis((struct imago_effect_stellae*)effect);
	effect = NULL;
	imago->back( 0x02 );
	imago->back( 0x04 );
	imago->flush();
	mode = board_mode_animate;
    }
    else
    {
	/*
	** Set board up for game play. Enter backdraw mode to begin
	** melt effect (which is driven by the animation interrupt).
	*/
	mode = board_mode_backdraw;

	// Free the title screen effect.
	if( effect != NULL )
	    free( effect );
	effect = NULL;

	metercount = 0;

	animList->clear();
	animList->set_imago( imago );

	set( &tbar, BOARD_TILE_TBAR );
	set( &bbar, BOARD_TILE_BBAR );
	set( &rbar, BOARD_TILE_RBAR );

	// These are never animated.
	tbar.tobj->incanim( imago );
	bbar.tobj->incanim( imago );
	rbar.tobj->incanim( imago );

	placed = NULL;
	backtobj = NULL;

	// Fill playfield with empty TObjs.
	for( r=0; r<BOARD_HEIGHT; ++r )
	{
	    for( c=0; c<BOARD_WIDTH; ++c )
		set( c, r, BOARD_TILE_EMPTY );
	}

	// Reset the Flow Meter.
	for( r=0; r<3; ++r )
	    set( &meter[r], BOARD_TILE_METER );
    }

    externSetAnim( 250, this );
}

/*
** Board::place()
** 
** Return
**    0 - Error.
**    1 - Tile placed.
**    2 - Nothing done.
**    3 - Tile obscures existing tile.
*/
int Board::place( int x, int y )
{
    int obscurum = 0;
    Tile *newtile;

    if( (x>(BOARD_WIDTH*40)) || (y>(BOARD_HEIGHT*40)) )
	return 0;

    newtile = &(playfield[x/40][y/40]);

    if( !newtile->canplace() )
    {
	/*
	fprintf( stdout, "TMP place on non empty/pipe, class %d t%d\n",
	    to->tclass, to->type );
	*/
	return 0;
    }

    // Ignore if tile has already been placed here.
    if( placed == newtile )
	return 2;

    // If already placed, restore previous tobj.
    if( placed != NULL )
	set( placed, backtobj );

    // New current placement.
    placed = newtile;
    backtobj = newtile->tobj;
    if( backtobj->tclass == tobj_tclass_pipe )
	++obscurum;

    // Place in given tile.
    set( newtile, ntile.tobj );

    return obscurum?3:1;
}

void Board::shift( TObj *newtobj )
{
    placed = NULL;
    backtobj = NULL;

    // Next in deck becomes "next tile".
    set( &ntile, deck[0].tobj );

    // Shift deck down one.
    for( int i=0; i<9; ++i )
	set( &deck[i], deck[i+1].tobj );

    // Add new tobj to bottom of deck.
    set( &deck[9], newtobj );
}

/*
** Board::incanim()
** 
** Animation interrupt hander. Returns time to set for next interrupt.
** 
*/
int Board::incanim( void )
{
    int interval = 0;

    /*
    fprintf( stdout, "TMP board incanim mode %d.\n", mode );
    fflush( stdout );
    */

    /*
    ** Break if a image flush is desired, otherwise return. interval
    ** must be set or no further interrupts will occur.
    */
    switch( mode )
    {
	case board_mode_title:
	    if( vinfo->c_class == PseudoColor )
		colores->effectRamp( (struct colores_effect_ramp*)effect );
	    return 500;
	case board_mode_backdraw:
	    imago->back( 0x03 );
	    //fputs( "TMP Doing backdraw for melt\n", stdout );
	    mode = board_mode_initmelt;
	    animList->incanim();
	    imago->flush();
	    interval = 75;
	    break;
	case board_mode_initmelt:
	    //fputs( "TMP initializing melt\n", stdout );
	    effectMelt = imago->effectMeltInit( 0, 0, 600, 400 );
	    imago->back( 0x02 );
	    mode = board_mode_melt;
	    interval = 75;
	    break;
	case board_mode_melt:
	{
	    if( imago->effectMelt(
		(struct imago_effect_melt*)effectMelt) == 2 )
	    {
		imago->effectMeltFinis((struct imago_effect_melt*)effectMelt);
		effectMelt = NULL;
		//fputs( "TMP melt over\n", stdout );
		mode = board_mode_animate;
	    }
	    interval = 75;
	    break;
	}
	case board_mode_animate:
	{
	    //fputs( "TMP bo mode animate\n", stdout );
	    animList->incanim();
	    if( vinfo->c_class == PseudoColor )
		interval = 250;
	    else
		interval = 350;
	    break;
	}
	case board_mode_pause:
	    imago->effectStellae( (struct imago_effect_stellae*)effect );
	    if( vinfo->c_class == PseudoColor )
		interval = 100;
	    else
		interval = 150;
	    break;
    }
    imago->flush();
    return interval;
}

int Board::set( Tile *t, int type )
{
    int rc;
    TObj *to;

    if( (to=tcache->fetch(type)) == NULL )
	return 0;

    if( (rc=t->set(to)) )
	animList->push( t );

    return rc;
}

int Board::set( int col, int row, int type )
{
    int rc;
    Tile *t;
    TObj *to;

    if( (to=tcache->fetch(type)) == NULL )
	return 0;

    t = &playfield[col][row];

    if( (rc=t->set(to)) )
	animList->push( t );

    return rc;
}

int Board::set( Tile *t, TObj *to )
{
    int rc;

    if( (rc=t->set(to)) )
	animList->push( t );

    return rc;
}

int Board::incflow( Tile *t, tobj_dir *flowdir, struct tobj_gain *gain )
{
    int rc;

    rc = t->incflow( imago, flowdir, gain );

    if( rc&TILE_FLOW_DRAW )
	animList->push( t );

    imago->flush();

    return rc;
}

int Board::incmeter( void )
{
    int rc;
    Tile *t;

    while( 1 )
    {
	if( metercount >= 3 )
	    return 3;

	t = &(meter[metercount]);
	rc = t->incmeter(imago);
	switch( rc )
	{
	    case 0:
		//fprintf( stdout, "TMP bo meter error\n" );
		//fflush(stdout);
		return 0;
	    case 1:
		//fprintf( stdout, "TMP bo meter ok\n" );
		//fflush(stdout);
		animList->push( t );
		imago->flush();
		return 1;
	    case 2:
		//fprintf( stdout, "TMP bo meter new tile\n" );
		//fflush( stdout );
		++metercount;
		continue;
	    default:
		fprintf( stdout, "ERROR. Illegal value from incmeter, %d.\n",
		    rc );
		fflush(stdout);
		return 0;
	}
    }
    // NOT REACHED.
}

void Board::boardExpose_cb( Widget w, XtPointer client_data, XEvent *event )
{
    Board *b = (Board*)client_data;
    XExposeEvent *ev = (XExposeEvent*)event;

    b->imago->flush( ev->x, ev->y, ev->width, ev->height );
}
