/*
** 
**  Copyright (c) 1998 by Larry Plona
** 
**  This distribution  of XPipeman is copyright by Larry Plona.
**  The  package  and  its  component  files  may  be  freely
**  distributed as long as  all copyright notices remain intact
**  and this distribution is not sold  as  commercial software.
** 
**  No warranty of any kind is attached with this software.  It
**  is supplied as is.
** 
** ----------------------------------------------------------------
*/

#include "episode.h"

extern int errno;
extern char *sys_errlist[];

Episode::Episode()
{
    levels = NULL;
    levelcnt = 0;
    currlevel = 0;
    artwork = NULL;
    glob = NULL;
    glob_name = NULL;
}

Episode::~Episode()
{
    if( levels != NULL )
	free( levels );
    if( glob != NULL )
	delete glob;
}

int Episode::load( const char *path, const char *name, int flagSound )
{
    const char *buf;
    char globpath[1023+1];
    int buflen;

    if( name == NULL )
	return 0;

    if( glob != NULL )
	delete glob;
    if( levels != NULL )
	free( levels );
    levelcnt = -1;
    this->path = path;

    sprintf( globpath, "%s/%s", path, name );
    glob = new Glob( globpath );
    if( !(glob->initialized()) )
    {
	fprintf( stdout, "ERROR. Failed to open glob \"%s\".\n", globpath );
	delete glob;
	glob = NULL;
	return 0;
    }
    glob_name = strdup(globpath);

    if( flagSound )
    {
	char linkline[1023+1];

	strcpy( linkline, "G:s_splash" );
	if( flagSound )
	{
	    sfx_glob( globpath );
	    sfx_link( EPISODE_SND_SPLASH, linkline, 0 );
	}
    }

    currlevel = -1;
    if( (levels=(struct levelrec*)malloc(
	sizeof(struct levelrec)*EPISODE_MAX_LEVELS)) == NULL )
    {
	fprintf( stdout, "ERROR. Bad malloc. %s.\n", sys_errlist[errno] );
	return 0;
    }
    levelcnt = 0;

    if( (buf=glob->extract("conf",&buflen)) == NULL )
    {
	fprintf( stdout, "ERROR. Malformed glob \"%s\".\n", globpath );
	delete glob;
	glob = NULL;
	return 0;
    }
    if( !parse_conf(buf,buflen) )
    {
	fputs( "ERROR. Can not parse episode conf file.\n", stdout );
	return 0;
    }

    return 1;
}

int Episode::initialized( void )
{
    return (levels==NULL) ? 0 : 1;
}

int Episode::rewind( void )
{
    currlevel = -1;
    if( artwork != NULL )
    {
	delete artwork;
	artwork = NULL;
    }
    return 1;
}

int Episode::new_level( void )
{
    char filename[1023+1];
    ++currlevel;
    if( currlevel >= levelcnt )
	return 2; // No more levels.

    if( artwork != NULL )
    {
	delete artwork;
	artwork = NULL;
    }
    if( glob_name != NULL )
	free( glob_name );

    sprintf( filename, "%s/%s.glob", path, levels[currlevel].artwork );
    artwork = new Artwork( filename, currlevel );
    if( (artwork==NULL) || !artwork->initialized() )
    {
	fprintf( stdout, "ERROR. Unable to load artwork \"%s\".\n",
	    levels[currlevel].artwork );
	return 0;
    }

    return 1;
}

int Episode::parse_conf( const char *buf, int len )
{
    const char *ep = (buf+len)-1;
    const char *ip;
    char *p;
    char line[EPISODE_MAX_CONFLINE];
    int lineno = 0;
    int flagStanza = EPISODE_STANZA_NULL;
    int max, remain;

    for( ip=buf; ip<=ep; )
    {
	++lineno;

	/*
	** Copy one line from buffer ( sgets() ).
	*/
	remain = (ep-ip)+1;
	max = EPISODE_MAX_CONFLINE>remain ? remain : EPISODE_MAX_CONFLINE;
	if( (p=(char*)memccpy(line,ip,(int)'\n',max)) == NULL )
	{
	    fprintf( stdout,
		"ERROR. Conf line %d too long or unterminated.\n",
		lineno );
	    return NULL;
	}

	/* Remove newline in copy. */
	*(p-1) = '\0';

	ip += (p-line); /* For next "sgets()". */

	/* Ignore leading whitespace. */
	for( p=line; isspace(*p); ++p );
	if( *p == '\0' )
	    continue;

	if( p[0] == '#' )
	    continue;

	if( p[0] == '%' )
	{
	    if( !strncmp(p,"%levels",10) )
	    {
		flagStanza = EPISODE_STANZA_LEVELS;
		continue;
	    }
	}

	switch( flagStanza )
	{
	    case EPISODE_STANZA_LEVELS:
	    {
		if( levelcnt >= EPISODE_MAX_LEVELS )
		{
		    fputs( "WARNING. Too many levels.\n", stdout );
		    continue;
		}

		if( !parse_level((const char*)p,&levels[levelcnt]) )
		    break;
		++levelcnt;
		break;
	    }
	    default:
		break;
	}
    }

    return 1;
}

int Episode::parse_level( const char *p, struct levelrec *rec )
{
    const char *fields = "%d %d %d | %d,%d,%d,%d,%d,%d,%d,%d,%d,%d | %s";
    char name[EPISODE_MAX_NAME+1];

    if( sscanf(p,fields,&rec->min_tiles,&rec->flow_rate,&rec->flow_start,
	&rec->obmap[0],&rec->obmap[1],&rec->obmap[2],&rec->obmap[3],
	&rec->obmap[4],&rec->obmap[5],&rec->obmap[6],&rec->obmap[7],
	&rec->obmap[8],&rec->obmap[9],name) != 14 )
    {
	fprintf( stdout, "ERROR. Bad levels line,  \"%s\".\n", p );
	return 0;
    }
    rec->artwork = strdup(name);

    return 1;
}

Glob *Episode::art_glob( void )
{
    return artwork->glob;
}

const char *Episode::art_glob_name( void )
{
    return glob_name;
}

int Episode::obmap( int row )
{
    return levels[currlevel].obmap[row];
}

int Episode::flow_rate( void )
{
    return levels[currlevel].flow_rate;
}

int Episode::flow_start( void )
{
    return levels[currlevel].flow_start;
}

int Episode::min_tiles( void )
{
    return levels[currlevel].min_tiles;
}
