/*
** 
**  Copyright (c) 1998 by Larry Plona
** 
**  This distribution  of XPipeman is copyright by Larry Plona.
**  The  package  and  its  component  files  may  be  freely
**  distributed as long as  all copyright notices remain intact
**  and this distribution is not sold  as  commercial software.
** 
**  No warranty of any kind is attached with this software.  It
**  is supplied as is.
** 
** ----------------------------------------------------------------
*/

#include "game.h"

extern int flagSound;
extern TCache tcache;

static int randint( int );

/*
** These names must be registered first so that the BOARD_TILE_*
** constants are correct.
*/
const char * const Game::rtilenames[] =
{
    "empty.xpm",
    "tbar.xpm",
    "bbar.xpm",
    "rbar.xpm",
    "meter.xpm",
    "start.xpm",
    "death.xpm",
    NULL
};

Game::Game( void )
{
    inFlow = 0;
    mode = game_mode_null;
    flowdir = tobj_dir_null;
    scorefile = NULL;
    createdLevel = 0;
    createdOptions = 0;
    createdScores = 0;
    createdInfo = 0;
}

/*
** Game::new_episode
**
** Initialize an episode. Must be called before new_mode().
**
** This is not a mode because an episode name must be given.
** It is not done in the constructor because we want a single
** game object to persist through out the executable's lifetime.
**
*/
int Game::new_episode( const char *path, const char *name )
{
    if( !episode.load(path,name,flagSound) )
    {
	fprintf( stdout, "ERROR. Unable to load episode \"%s\".\n", name );
	return 0;
    }
    score.reset();
    return 1;
}

/*
** Game::flow()
**
** Called via external timeout interrupt.
*/
int Game::incflow( void )
{
    if( inFlow )
    {
	int rc;
	struct tobj_gain gain;
	Tile *t;
	tobj_dir ndir = flowdir;

	/*
	** If current flow position is out of bounds terminate flow.
	** This is not a sanity check.
	*/
	if( flowcol<0 || flowrow<0 || flowcol>=BOARD_WIDTH ||
	    flowrow>=BOARD_HEIGHT )
	{
	    board->set( ocol, orow, BOARD_TILE_DEATH );
	    //fputs( "TMP gm flow CRASH out of bounds\n", stdout );
	    new_mode( game_signal_crash );
	    return 0;
	}

	t = &board->playfield[flowcol][flowrow];

	/*
	fprintf( stdout, "TMP gm flow t%d flowcnt %d (%d) dir %d\n",
	    t->tobj->type, flowcnt, flowcnt-6, flowdir );
	*/

	rc = board->incflow( t, &ndir, &gain );

	if( gain.newtile )
	    ++lstats.tacta;
	if( gain.cross )
	{
	    ++lstats.crossed;
	    if( flagSound )
		sfx_play( ARTWORK_SND_CROSS, 0 );
	}

	guiSetLabelNumber( game_label_remain,
	    lstats.tacta-episode.min_tiles() );

	if( (rc&TILE_FLOW_CONT) == 0 )
	{
	    board->set( ocol, orow, BOARD_TILE_DEATH );
	    //fputs( "TMP gm flow CRASH\n", stdout );
	    new_mode( game_signal_crash );
	    return 0;
	}

	/*
	** New direction for next flow interrupt.
	*/
	ocol = flowcol;
	orow = flowrow;
	if( ndir != tobj_dir_null )
	{
	    // fprintf( stdout, "TMP gm flow newdir %d\n", ndir );
	    switch( ndir )
	    {
		case tobj_dir_left:
		    --flowcol;
		    break;
		case tobj_dir_right:
		    ++flowcol;
		    break;
		case tobj_dir_top:
		    --flowrow;
		    break;
		case tobj_dir_bottom:
		    ++flowrow;
		    break;
	    }
	    flowdir = ndir;
	}

	return flowrate;
    }

    {
	int rc;

	rc = board->incmeter();
	if( rc == 0 )
	{
	    fprintf( stdout, "ERROR. Bad increment of flow meter.\n" );
	    return 0;
	}
	if( rc == 3 )
	{
	    // fputs( "TMP gm meter begin flow\n", stdout );
	    // fflush(stdout);
	    inFlow = 1;
	    flowrate = episode.flow_rate();
	}
    }

    if( flagSound )
    {
	sfx_play( (inFlow?ARTWORK_SND_FLOWSTART:ARTWORK_SND_FLOWCNT), 0 );
    }

    return flowrate;
}

static int randint( int max )
{
    return ((int)rand()%(max));
}

int randtile( int flag )
{
#if 0
    switch( flag )
    {
	case AN_RAND_WHOLE:
	    return ((int)rand()%(AN_TILE_LASTWHOLE-1)) + 1;
	case AN_RAND_BASIC:
	    return ((int)rand()%(AN_TILE_LASTBASIC-1)) + 1;
	case AN_RAND_OBSTACLE:
	{
	    return ((int)rand()%6) + AN_TILE_OBSTACLE;
	}
    }
#endif
    return 0;
}

/*
** new_mode()
**
** This is an anchor function, it should exit if an error is found.
*/
int Game::new_mode( game_signal sig )
{
    switch( sig )
    {
	case game_signal_start:
	{
	    mode = game_mode_title;
	    break;
	}
	case game_signal_new:
	{
	    if( mode == game_mode_pause )
		return 1;
	    if( flagSound )
		sfx_halt( EPISODE_SND_SPLASH, 0 );
	    if( mode != game_mode_title )
	    {
		mode = game_mode_title;
		break;
	    }
	    episode.rewind();
	    score.reset();
	    mode = game_mode_play;
	    break;
	}
	case game_signal_pause:
	{
	    if( mode == game_mode_pause )
	    {
		board->reset( episode.artwork->syncs[0] );
		prep_intr();
		mode = game_mode_play;
		return 1;
	    }
	    if( mode != game_mode_play )
		return 1;
	    mode = game_mode_pause;
	    break;
	}
	case game_signal_continue:
	{
	    if( mode == game_mode_gameover )
		mode = game_mode_title;
	    else
		mode = game_mode_play;
	    break;
	}
	case game_signal_crash:
	{
	    mode = game_mode_levelover;
	    break;
	}
	case game_signal_quit:
	case game_signal_error:
	{
	    mode = game_mode_quit;
	    break;
	}
    }

    switch( mode )
    {
	case game_mode_play:
	{
	    prep_stop();
	    prep_popdown();

	    if( !prep_level() )
		return 0;

	    if( !prep_board() )
		return 0;

	    if( !prep_intr() )
		return 0;

	    break;
	}
	case game_mode_title:
	{
	    char *buf;
	    int buflen;

	    prep_stop();
	    prep_popdown();

	    /*
	    ** This is redundant since a rewind is done when a start
	    ** signal is given. However, rewinding here will remove
	    ** any artwork cursor left over from the previous game.
	    */
	    episode.rewind();

	    if( (buf=episode.glob->extract("splash.xpm",&buflen)) == NULL )
	    {
		fprintf( stdout, "ERROR. Malformed glob \"%s\".\n",
		    "splash.xpm" );
		return 0;
	    }

	    if( flagSound )
		sfx_play( EPISODE_SND_SPLASH, 0 );

	    if( !board->title(buf) )
		return 0;

	    break;
	}
	case game_mode_levelover:
	case game_mode_gameover:
	{
	    struct score_pars totals;

	    prep_stop();

	    // Add the level statistics into the game statistics.
	    score.accipere( &lstats );

	    if( episode.min_tiles() > lstats.tacta )
		mode = game_mode_gameover;
	    else
		mode = game_mode_levelover;

	    if( mode == game_mode_gameover )
		score.referre( "./xpipescores" );

	    if( !createdLevel )
	    {
		char *title;

		if( (title=episode.glob->extract("title.xpm",NULL)) == NULL )
		{
		    fprintf( stdout, "ERROR. Malformed glob \"%s\".\n",
			"title.xpm" );
		}
		guiCreateDialog( game_dialog_level, (const char*)title, this );
		createdLevel = 1;
	    }

	    if( flagSound )
	    {
		sfx_play(
		    (mode==game_mode_gameover?ARTWORK_SND_GAMEOVER:
		    ARTWORK_SND_LEVELOVER), 0 );
	    }

	    score.fetch( &totals );
	    guiSetLabel( game_label_levelTitle, 
		(mode==game_mode_levelover)?"Level Over":"Game Over" );
	    guiSetStats( episode.currlevel+1, &lstats, &totals );
	    guiDialog( game_dialog_level, 1 );
	    break;
	}
	case game_mode_pause:
	{
	    prep_stop();

	    /*
	    ** A zero return is not always an error. It can occur
	    ** when the board is not able to pause (such as when
	    ** it is in the middle of a melt effect.
	    */
	    if( !board->pause() )
	    {
		prep_intr();
		mode = game_mode_play;
	    }

	    break;
	}
	case game_mode_quit:
	{
	    /* Anything which needs to be cleaned up? */
	    if( board != NULL )
		delete board;
	}
    }
    return 1;
}

int Game::prep_level( void )
{
    srand( (int)time(NULL) );
    if( !(tcache.reset()) )
	return 0;
    if( !(tcache.setMax(128)) )
	return 0;

    /*
    ** Stock tobjs must be registered first since the code
    ** relies on their type numbers always being 0-n.
    */
    for( int i=0; rtilenames[i]!=NULL; ++i )
    {
	if( tcache.notify(rtilenames[i]) == -1 )
	{
	    fputs( "ERROR. Unable to register required tiles.\n",
		stdout );
	    return 0;
	}
    }

    if( !episode.new_level() )
    {
	fputs( "ERROR. Unable to begin new level.\n", stdout );
	new_mode( game_signal_error );
    }

    Score::scoreInitPars( &lstats );
    score.level = episode.currlevel+1;
    lstats.emissa = 2; // Tile in next and start tile.
    lstats.proposita = 1; // Start tile.
    guiSetLabelNumber( game_label_remain, 0-episode.min_tiles() );
    guiSetLabelNumber( game_label_level, episode.currlevel+1 );
    tcache.setGlob( episode.art_glob() );
    inFlow = 0;

    return 1;
}

int Game::prep_intr( void )
{
    guiSetInput( 1, this );
    if( inFlow )
	flowrate = episode.flow_rate();
    else
	flowrate = episode.flow_start()/6;
    guiSetFlow( flowrate, this );
    return 1;
}

int Game::prep_stop( void )
{
    guiSetFlow( 0, this );
    guiSetInput( 0, NULL );
    return 1;
}

int Game::prep_popdown( void )
{
    guiDialog( game_dialog_level, 0 );
    guiDialog( game_dialog_credits, 0 );

    /* Some might like to have the scores up while they are playing. */
    /* guiDialog( game_dialog_scores, 0 ); */

    return 1;
}

/*
** Game::prep_board
** 
** Instantiate deck, place obstacles and place the starting tobj.
** Flow meter, empty tiles and static area are handled by the board
** object.
*/ 
int Game::prep_board( void )
{
    int c, r;
    int *blks = NULL;
    Artwork *art;

    // FUTURE. Syncs should be set dynamicaly through a function.
    board->reset( episode.artwork->syncs[0] );

    art = episode.artwork;

    if( (blks=(int*)malloc(sizeof(int)*art->pipecnt)) == NULL )
    {
	fprintf( stdout, "ERROR. Bad malloc. %s.\n", sys_errlist[errno] );
	return 0;
    }
    for( r=0; r<art->pipecnt; ++r )
	blks[r] = 0;

    // Choose a next tile.
    c = randint( art->pipecnt );
    board->set( &board->ntile, art->pipes[c] );
    ++blks[c];

    // Create a deck of upcoming tiles.
    {
	int *pipes = art->pipes;

	for( r=0; r<10; ++r )
	{
	    while( 1 )
	    {
		c = randint( art->pipecnt );
		if( (art->pipecnt>10) && (blks[c]>=2) )
		    continue;
		board->set( &board->deck[r], pipes[c] );
		break;
	    }
	}
    }

    free( blks );

    // Place obstacles.
    if( art->obscnt > 0 )
    {
	int o;
	int *obstacles = art->obstacles;

	for( r=0; r<BOARD_HEIGHT; ++r )
	{
	    o = episode.obmap(r);
	    if( o != 0 )
	    {
		c = randint( art->obscnt );
		if( !board->set(o,r,obstacles[c]) )
		{
		    fprintf( stdout, "ERROR. Unable to place obstacle.\n" );
		    return 0;
		}
	    }
	}
    }

    // Place starting tile object.
    {
	int coff, roff;
	TObj *t;

	if( (t=tcache.fetch(BOARD_TILE_START)) == NULL )
	{
	    fputs( "ERROR. No start TObj.\n", stdout );
	    return 0;
	}
	coff = tobj_dirs[t->tinfo.start.initial_dir].coff;
	roff = tobj_dirs[t->tinfo.start.initial_dir].roff;

	while( 1 )
	{
	    // c = ((int)rand()%(BOARD_WIDTH-3))+2;
	    // r = ((int)rand()%(BOARD_HEIGHT-4))+1;
	    c = ((int)rand()%(BOARD_WIDTH-2))+1;
	    r = ((int)rand()%(BOARD_HEIGHT-2))+1;
	    if( board->playfield[c][r].tobj->type != BOARD_TILE_EMPTY )
		continue;
	    if( board->playfield[c+coff][r+roff].tobj->type !=
		BOARD_TILE_EMPTY )
	    {
		continue;
	    }

	    /* FUTURE. This code assumes no diagnols. */
	    if( roff == 0 )
	    {
		if( board->playfield[c+coff][r+1].tobj->type !=
		    BOARD_TILE_EMPTY )
		{
		    continue;
		}
		if( board->playfield[c+coff][r-1].tobj->type !=
		    BOARD_TILE_EMPTY )
		{
		    continue;
		}
	    }
	    else
	    {
		if( board->playfield[c+1][r+roff].tobj->type !=
		    BOARD_TILE_EMPTY )
		{
		    continue;
		}
		if( board->playfield[c-1][r+roff].tobj->type !=
		    BOARD_TILE_EMPTY )
		{
		    continue;
		}
	    }

	    /*
	    if( (board->playfield[c][r].tobj->type!=BOARD_TILE_EMPTY) ||
		(board->playfield[c+1][r].tobj->type!=BOARD_TILE_EMPTY) ||
		(board->playfield[c+1][r+1].tobj->type!=BOARD_TILE_EMPTY) ||
		(board->playfield[c][r+1].tobj->type!=BOARD_TILE_EMPTY) )
	    {
		continue;
	    }
	    */
	    break;
	}

	board->start = &(board->playfield[c][r]);
	board->set( board->start, BOARD_TILE_START );
	flowdir = t->tinfo.start.initial_dir;
	flowcol = c;
	flowrow = r;
	/*
	** Initial previous column and row are the same as the next column
	** and row. This prevents the death TObj from being written over
	** the start TObj.
	*/
	ocol = flowcol;
	orow = flowrow;
    }

    return 1;
}

int Game::credits( void )
{
    if( !createdInfo )
    {
	char *title;

	if( (title=episode.glob->extract("title.xpm",NULL)) == NULL )
	{
	    fprintf( stdout, "ERROR. Malformed glob \"%s\".\n",
		"title.xpm" );
	    // Non fatal.
	}

	guiCreateDialog( game_dialog_credits, (const char*)title, this );
	createdInfo = 1;
    }

    guiDialog( game_dialog_credits, 1 );

    return 1;
}

int Game::options( void )
{
    if( !createdOptions )
    {
	char *title;

	if( (title=episode.glob->extract("title.xpm",NULL)) == NULL )
	{
	    fprintf( stdout, "ERROR. Malformed glob \"%s\".\n",
		"title.xpm" );
	    // Non fatal.
	}

	guiCreateDialog( game_dialog_options, (const char*)title, this );
	createdOptions = 1;
    }

    guiDialog( game_dialog_options, 1 );

    return 1;
}

int Game::scores( void )
{
    const char **list;

    if( !createdScores )
    {
	char *title;

	if( (title=episode.glob->extract("title.xpm",NULL)) == NULL )
	{
	    fprintf( stdout, "ERROR. Malformed glob \"%s\".\n",
		"title.xpm" );
	    // Non fatal.
	}

	guiCreateDialog( game_dialog_scores, (const char*)title, this );
	createdScores = 1;
    }

    list = score.summaByList( scorefile );
    guiSetList( game_list_scores, list );

    guiDialog( game_dialog_scores, 1 );

    return 1;
}

int Game::event( game_signal sig )
{
    if( sig == game_signal_flow )
	return incflow();
    else
	new_mode( sig );

    return 1;
}

int Game::event( game_input inp, int x, int y )
{
    switch( inp )
    {
	case game_input_fast:
	{
	    flowrate = 25;
	    break;
	}
	case game_input_place:
	{
	    int rc;

	    rc = board->place( x, y );
	    if( rc == 0 )
	    {
		if( flagSound )
		    sfx_play( ARTWORK_SND_BADPLACE, 0 );
		break;
	    }
	    if( rc == 3 )
		++lstats.obscura;
	    if( rc != 2 )
		++lstats.proposita;
	    if( flagSound )
		sfx_play( ARTWORK_SND_PLACE, 0 );
	    break;
	}
	case game_input_shift:
	{
	    int type;
	    TObj *newtobj;

	    type = episode.artwork->pipes[
		randint(episode.artwork->pipecnt)];

	    newtobj = tcache.fetch( type );
	    board->shift( newtobj );
	    if( flagSound )
		sfx_play( ARTWORK_SND_SHIFT, 0 );
	    ++lstats.emissa;
	    break;
	}
	default:
	    return 0;
    }

    return 1;
}
