/*
** 
**  Copyright (c) 1998 by Larry Plona
** 
**  This distribution  of XPipeman is copyright by Larry Plona.
**  The  package  and  its  component  files  may  be  freely
**  distributed as long as  all copyright notices remain intact
**  and this distribution is not sold  as  commercial software.
** 
**  No warranty of any kind is attached with this software.  It
**  is supplied as is.
** 
** ----------------------------------------------------------------
*/

/*
** This file should not contain xpipeman.h or any XPipeman code.
** It is dependant only on the Glob code.
**
** Right now, the cached sounds have a one-for-one association with
** their application ids. This is ok, because the game currently
** has very few sounds. In the future, the sound driver must be
** smart about cacheing and releasing sounds.
**
** Code for speed.
**
** Protocol
**
** "llll" fields are records lengths, including newline. All commands
** without "llll" fields are padded to 7 bytes (including newline).
** "llll" fields are 4 hex digits. All commands must be at least 7
** bytes long Padding is with spaces.
**
**    %a llll dd xxxx\n       - Set flags. id (2 ASCII). flags (4 ASCII hex).
**    %d llll dd\n            - Unlink and empty id.
**    %e llll\n               - Empty cache and links.
**    %g llll globfile\n      - Set current glob.
**    %h llll\n               - Halt active sounds.
**    %l llll dd filename\n   - Link filename with id (two ASCII).
**    %m llll dd\n	      - Mute on (01) or off (00).
**    %o llll dd\n            - Load id now.
**    %s llll\n               - Sync with current sounds.
**    %v llll dd\n            - Change volume. dd is 0-99.
**    dd    \n                - Play id (two ASCII).
*/

#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#ifdef USE_POLL
#include <poll.h>
#else
#ifdef linux
#include <sys/time.h>
#include <sys/types.h>
#else
#include <sys/select.h>
#endif
#endif
#include <signal.h>
#include "sfx.h"
#include "glob.h"

#ifdef HAVE_OSS
struct srec
{
    unsigned short len;
    unsigned short flags; /* SFX_SFLG_* */
    unsigned char *buf;
    char *filename;
};

struct chan
{
    unsigned char *cur;
    unsigned char *begin;
    unsigned char *end;
    short id; /* Cache id. */
    short time;
};

static int mytime = 0;
static struct chan chans[SFX_MAX_CHANS];
static int fdAudio = -1;
static int fdPipe[2];
static pid_t child;
static struct srec *srecs;
static int mixsize;
static unsigned char *mixbuffer;
static int intr_nosound = SFX_INTR_NOSOUND;
static int intr_quit = SFX_INTR_QUIT;
static float globalVolume = 0.5;
static unsigned int globalMute = 0;
static Glob *glob;

static int child_loop();
static void child_intr( int );
static int child_update();
static unsigned char *child_fetch( Glob*, const char*, int* );
static int child_addsfx( Glob*, short );
static int process_cmds( int );
#endif

extern int errno;
extern char *sys_errlist[];

int sfx_init( void )
{
#if HAVE_OSS
    if( pipe(fdPipe) == -1 )
    {
	fprintf( stdout, "ERROR. Bad pipe. %s.\n",
	    sys_errlist[errno] );
	return 0;
    }
    /*
    fprintf( stdout, "TMP pipe %d %d\n", fdPipe[0], fdPipe[1] );
    fflush( stdout );
    */

    if( (child=fork()) == -1 )
    {
	fprintf( stdout, "ERROR. Bad fork. %s.\n",
	    sys_errlist[errno] );
    }
    if( child == 0 )
    {
	int i;

	/* Sound daemon only reads. */
	/*close( fdPipe[1] );*/

	/*
	** Non blocking reads. If you are having trouble with O_NONBLOCK,
	** try O_NDELAY. See below.
	*/
	i = fcntl( fdPipe[0], F_GETFL );
	fcntl( fdPipe[0], F_SETFL, i|O_NONBLOCK );

	if( (fdAudio=open("/dev/dsp", O_WRONLY)) == -1 )
	{
	    fprintf( stdout, "ERROR. Bad open of audio. %s.\n",
		sys_errlist[errno] );
	    kill( getppid(), intr_nosound );
	    exit( 1 );
	}
	/*
	** Set fragment size right away. Bit map of MMMMSSSS.
	** This setting has a large impact on the quality of
	** the sound. If the number or size of the fragments
	** is too large, the sounds will be out of sync. If
	** too small, the sound will be choppy. The value
	** below represents the smallest non-choppy value for
	** UnixWare 2 running on an EISA 486/66. 0x7fff0008
	** is safe, but a little slow.
	**
	** MMMM is number of fragments or 7FFF for unlimited.
	**
	** SSSS is 2^S for size of fragment.
	** 9 - 512 bytes
	** 8 - 256 bytes
	**
	*/
	i = 0x00100008;
	if( ioctl(fdAudio,SNDCTL_DSP_SETFRAGMENT,&i) == -1 )
	{
	    fprintf( stdout, "ERROR. Bad SETFRAGMENT.\n" );
	    kill( getppid(), intr_nosound );
	    exit( 2 );
	}

	/*
	** Set encoding format, number of channels and sample rate.
	** FUTURE: This should come from the application.
	*/
	i = AFMT_U8;
        if( ioctl(fdAudio,SNDCTL_DSP_SETFMT,&i) == -1 )
	{
	   fprintf( stdout, "ERROR. Bad SNDCTL_DSP_SETFMT. %s.\n",
	     sys_errlist[errno] );
	    kill( getppid(), intr_nosound );
	    exit( 3 );
	}
        i = 0;
        if( ioctl(fdAudio,SNDCTL_DSP_STEREO,&i) == -1 )
	{
	    fprintf( stdout, "ERROR. Bad SNDCTL_DSP_STEREO. %s\n",
		sys_errlist[errno] );
	    kill( getppid(), intr_nosound );
	    exit( 4 );
	}
        i = 11025;
        if( ioctl(fdAudio,SNDCTL_DSP_SPEED,&i) == -1 )
	{
	    fprintf( stdout, "ERROR. Bad SNDCTL_DSP_SPEED. %s\n",
		sys_errlist[errno] );
	    kill( getppid(), intr_nosound );
	    exit( 5 );
	}

	/*
	** Get fragment size being used (hopefully its the same value
	** we set above (but we may be overriden).
	*/
	if( ioctl(fdAudio,SNDCTL_DSP_GETBLKSIZE,&mixsize) == -1 )
	{
	    fprintf( stdout, "ERROR. Bad SNDCTL_DSP_GETBLKSIZE. %s\n",
		sys_errlist[errno] );
	    kill( getppid(), intr_nosound );
	    exit( 6 );
	}
	//fprintf( stdout, "TMP frag is %d\n", mixsize );
	if( (mixbuffer=(unsigned char*)malloc(sizeof(char)*mixsize)) == NULL )
	{
	    fprintf( stdout, "ERROR. Bad malloc. %s\n",
		sys_errlist[errno] );
	    kill( getppid(), intr_nosound );
	    exit( 7 );
	}

	if( (srecs=(struct srec*)malloc(sizeof(struct srec)*SFX_MAX_SOUNDS))
	    == NULL )
	{
	    fprintf( stdout, "ERROR. Bad malloc. %s\n",
		sys_errlist[errno] );
	    kill( getppid(), intr_nosound );
	    exit( 7 );
	}
	for( i=0; i<SFX_MAX_SOUNDS; ++i )
	{
	    srecs[i].buf = NULL;
	    srecs[i].filename = NULL;
	    srecs[i].len = 0;
	    srecs[i].flags = 0;
	}
	child_loop();
	exit( 0 );
    }

    /* Application only writes. */
    /*close( fdPipe[0] );*/

#endif
    return 1;
}

int sfx_delete( short id )
{
#ifdef HAVE_OSS
    char line[1023+1];

    sprintf( line, "%%d %04x %02d\n", 11, id );

    //fprintf( stdout, "TMP sfx_delete %d\n", id ); fflush( stdout );
    if( write(fdPipe[1],line,11) == -1 )
    {
	fprintf( stdout,
	    "ERROR. Bad write of audio pipe. %s.\n",
	    sys_errlist[errno] );
	return 0;
    }
#endif
    return 1;
}

int sfx_finis( void )
{
#ifdef HAVE_OSS
    /*fprintf( stdout, "TMP sfx_finis\n" ); fflush( stdout );*/
    kill( child, SIGTERM );
    close( fdPipe[1] );
    fdPipe[1] = -1;
#endif
    return 1;
}

int sfx_flags( short id, unsigned short flags )
{
#ifdef HAVE_OSS
    char line[1023+1];

    sprintf( line, "%%a %04x %02d %04x\n", 16, id, flags );

    /*fprintf( stdout, "TMP sfx_flags %02d\n", id ); fflush( stdout );*/
    if( write(fdPipe[1],line,16) == -1 )
    {
	fprintf( stdout,
	    "ERROR. Bad write of audio pipe. %s.\n",
	    sys_errlist[errno] );
	return 0;
    }
#endif
    return 1;
}

int sfx_glob( const char *globfile )
{
#ifdef HAVE_OSS
    int len;
    char line[1023+1];

    len = strlen(globfile) + 9;
    sprintf( line, "%%g %04x %s\n", len, globfile );

    //fprintf( stdout, "TMP sfx_glob %s\n", globfile ); fflush( stdout );
    len = strlen( line );
    if( write(fdPipe[1],line,len) == -1 )
    {
	fprintf( stdout,
	    "ERROR. Bad write of audio pipe. %s.\n",
	    sys_errlist[errno] );
	return 0;
    }
#endif
    return 1;
}

int sfx_halt( short id, unsigned short flags )
{
#if 0 /* TMP */
#ifdef HAVE_OSS
    char filebuf[8+1];

    sprintf( filebuf, "%h %04x %02d\n", 11, id );
    if( write(fdPipe[1],filebuf,11) != 11 )
    {
	fprintf( stdout, "ERROR. Bad write of audio pipe (halt). %s.\n",
	    sys_errlist[errno] );
	return 0;
    }
#endif
#endif
    return 1;
}

int sfx_link( short id, const char *soundfile, unsigned short flags )
{
#ifdef HAVE_OSS
    char line[1023+1];
    int len;

    len = strlen(soundfile) + 17;
    sprintf( line, "%%l %04x %02d %04x %s\n", len, id, flags, soundfile );

    //fprintf( stdout, "TMP sfx_link %02d %s\n", id, soundfile );
    //fflush( stdout );
    if( write(fdPipe[1],line,len) == -1 )
    {
	fprintf( stdout,
	    "ERROR. Bad write of audio pipe. %s.\n",
	    sys_errlist[errno] );
	return 0;
    }
#endif
    return 1;
}

int sfx_load( short id )
{
#ifdef HAVE_OSS
    char line[1023+1];

    sprintf( line, "%%o %04x %02d\n", 11, id );

    /*fprintf( stdout, "TMP sfx_load %02d\n", id ); fflush( stdout );*/
    if( write(fdPipe[1],line,11) == -1 )
    {
	fprintf( stdout,
	    "ERROR. Bad write of audio pipe. %s.\n",
	    sys_errlist[errno] );
	return 0;
    }
#endif
    return 1;
}

int sfx_mute( int toggle )
{
    char line[1023+1];

#ifdef HAVE_OSS
    sprintf( line, "%%m %04x %02d\n", 11, toggle );

    if( write(fdPipe[1],line,11) == -1 )
    {
	fprintf( stdout,
	    "ERROR. Bad write of audio pipe. %s.\n",
	    sys_errlist[errno] );
	return 0;
    }
#endif
    return 1;
}

int sfx_play( short id, unsigned short flags )
{
#ifdef HAVE_OSS
    char filebuf[7+1];

    sprintf( filebuf, "%02d    \n", id );

    //fprintf( stdout, "TMP sfx_play %d\n", id ); fflush( stdout );
    if( write(fdPipe[1],filebuf,7) != 7 )
    {
	fprintf( stdout, "ERROR. Bad write of audio pipe (play). %s.\n",
	    sys_errlist[errno] );
	return 0;
    }
#endif
    return 1;
}

int sfx_volume( signed int vol )
{
#ifdef HAVE_OSS
    char filebuf[31+1];

    sprintf( filebuf, "%%v %04x %03d\n", 12, vol );

    if( write(fdPipe[1],filebuf,12) != 12 )
    {
	fprintf( stdout, "ERROR. Bad write of audio pipe (play). %s.\n",
	    sys_errlist[errno] );
	return 0;
    }
#endif
    return 1;
}

#ifdef HAVE_OSS
static int child_loop( void )
{
    int rc;
    int flagBlock = SFX_POLL_BLOCK;
#ifdef USE_POLL
    struct pollfd fdset[2];
    int waitp;
    int fdcnt;
#else
    fd_set fdReads;
    fd_set fdWrites;
    struct timeval *waitp;
    struct timeval waitv;
#endif

    for( rc=0; rc<SFX_MAX_CHANS; ++rc )
    {
	chans[rc].cur = NULL;
	chans[rc].begin = NULL;
	chans[rc].end = NULL;
    }

    signal( SFX_INTR_QUIT, child_intr );

    /*
    ** POLLHACK
    **
    ** Normally, whether poll or select is used, the application waits
    ** indefintely until there is data to read on the command pipe. If,
    ** if one or more sounds are active, the application will wait until
    ** either there is data on the command pipe or the audio device may
    ** be written to. The flagBlock variable indicates if blocking on
    ** the audio device is needed.
    **
    ** When POLLHACK is defined, no blocking ever occurs on the audio
    ** device. When there are active sounds, poll (or select) is told
    ** to return immediately.
    **
    ** POLLHACK is independant of whether poll or select is used.
    */

    while( 1 )
    {
	++mytime;
#ifdef USE_POLL
	fdset[0].fd = fdPipe[0];
	fdset[0].events = POLLIN;
	fdset[0].revents = 0;
	fdcnt = 1;
#ifndef USE_POLLHACK
	if( flagBlock == SFX_POLL_NOWAIT )
	{
	    fdset[1].fd = fdAudio;
	    fdset[1].events = POLLOUT;
	    fdset[1].revents = 0;
	    ++fdcnt;
	}
	waitp = -1;
#else /* USE_POLL & USE_POLLHACK */
	if( flagBlock == SFX_POLL_NOWAIT )
	    waitp = 0;
	else if( flagBlock == SFX_POLL_BLOCK )
	    waitp = -1;
	else
	    waitp = 0; /* 250. Little less than one fragment. */
#endif
	/*
	fprintf( stdout, "TMP pip %d aud %d\n", fdPipe[0], fdAudio );
	fprintf( stdout, "TMP sleep %d %ld %d %d\n", flagBlock,
	    time(NULL), fdcnt, waitp );
	fprintf( stdout, "TMP b fdset[0].fd %d\n", fdset[0].fd );
	fprintf( stdout, "TMP b fdset[0].events %x\n", fdset[0].events );
	fprintf( stdout, "TMP b fdset[0].revents %x\n", fdset[0].revents );
	*/
	rc = poll( fdset, fdcnt, waitp );
	/*
	fprintf( stdout, "TMP a sleep %ld %d\n", time(NULL), rc );
	fprintf( stdout, "TMP a fdset[0].fd %d\n", fdset[0].fd );
	fprintf( stdout, "TMP a fdset[0].events %x\n", fdset[0].events );
	fprintf( stdout, "TMP a fdset[0].revents %x\n", fdset[0].revents );
	*/
#else /* !USE_POLL */
	FD_ZERO(&fdReads);
	FD_ZERO(&fdWrites);
	FD_SET(fdPipe[0],&fdReads);
#ifndef USE_POLLHACK
	if( flagBlock == SFX_POLL_NOWAIT )
	    FD_SET(fdAudio,&fdWrites);
	waitp = NULL;
#else
	if( flagBlock == SFX_POLL_NOWAIT )
	{
	    waitv.tv_sec = 0;
	    waitv.tv_usec = 0;
	    waitp = &waitv;
	}
	else if( flagBlock == SFX_POLL_BLOCK )
	    waitp = NULL;
	else
	{
	    waitv.tv_sec = 0;
	    waitv.tv_usec = 2500; /* Little less than one fragment. */
	    waitp = &waitv;
	}
#endif
	/*
	printf( stdout, "TMP sleep %d %ld %d\n", flagBlock, time(NULL),
	    waitp );
	fprintf( stdout, "TMP pip %d aud %d\n", fdPipe[0], fdAudio );
	*/
	rc = select( FD_SETSIZE, &fdReads, &fdWrites, NULL, waitp );
	/*
	fprintf( stdout, "TMP asleep %ld\n", time(NULL) );
	*/
#endif

	if( rc == -1 )
	{
	    fprintf( stdout, "ERROR. Bad poll/select. %s (%d).\n",
		sys_errlist[errno], errno );
	    break;
	}

	/* If we can read the command pipe, do so. */
	if( rc > 0 )
	{
#ifdef USE_POLL
	    if( fdset[0].revents & POLLIN )
#else
	    if( FD_ISSET(fdPipe[0],&fdReads) )
#endif
	    {
		rc = process_cmds( fdPipe[0] );
		if( rc > 0 )
		    flagBlock = SFX_POLL_NOWAIT;
	    }
	}

	/*
	** If we can write audio data, do so. If POLLHACK is defined,
	** we assume this to be true (if there are active sounds).
	*/
#ifndef USE_POLLHACK
#ifdef USE_POLL
	/*
        ** Without POLLHACK, we check audio only when NOWAIT is set;
	** i.e. when there are active sounds.
        */
	if( (flagBlock==SFX_POLL_NOWAIT) && fdset[1].revents & POLLOUT )
#else
	if( FD_ISSET(fdAudio,&fdWrites) )
#endif
#else
	if( (flagBlock==SFX_POLL_NOWAIT) || (flagBlock==SFX_POLL_TIMED) )
#endif
	{
	    audio_buf_info rec;

	    /*fputs( "TMP sel aud\n", stdout );*/
	    if( ioctl(fdAudio,SNDCTL_DSP_GETOSPACE,&rec) == -1 )
	    {
		fprintf( stdout, "ERROR. Bad GETOSPC. %s.\n",
		    sys_errlist[errno] );
		break;
	    }
	    /*fprintf( stdout, "TMP frg %d\n", rec.fragments );*/
	    /*
	    ** POLLHACK
	    ** When there are no fragments left, wait for the duration
	    ** of one fragment. Otherwise the application will spin until
	    ** a fragment becomes available. Actually the application
	    ** should wait a little less than the duration of one fragment
	    ** in order to avoid underruns.
	    */
	    if( rec.fragments < 1 )
	    {
#ifdef USE_POLLHACK
		flagBlock = SFX_POLL_TIMED; /* Wait for one fragment. */
#endif
		/*fputs( "TMP no frags\n", stdout );*/
	    }
	    else
		flagBlock = child_update();
	}
    }

    kill( getppid(), intr_nosound );
    exit( 0 );
    return 0;
}
#endif

#ifdef HAVE_OSS
int process_cmds( int fd )
{
    int rc;
    int id;
    int flagPlay = 0;
    char line[1023+1];

    while( 1 )
    {
	/*fputs( "TMP sel pip\n", stdout ); fflush( stdout );*/
	line[0] = '\0';
	if( (rc=read(fd,line,7)) == -1 )
	{
	    /* Empty pipe. */
	    if( errno == EAGAIN )
		break;
	    fprintf( stdout, "ERROR. Bad read from cpipe. %s.\n",
		sys_errlist[errno] );
	    return -1;
	}

	/* Empty pipe. Only occurs if O_NDELAY is used for O_NONBLOCK. */
	if( rc == 0 )
	    break;

	line[7] = '\0';
	/*
	fprintf( stdout, "TMP rd %d \"%s\" fr pip\n", rc, line );
	fflush( stdout );
	*/

	if( line[0] == '%' )
	{
	    unsigned int len;

	    sscanf( line, "%%%*c %x ", &len );

	    if( (rc=read(fdPipe[0],&line[7],len-7)) == -1 )
	    {
		fprintf( stdout, "ERROR. Bad read from cpipe. %s.\n",
		    sys_errlist[errno] );
		break;
	    }
	    line[len-1] = '\0';
	}

	if( !strncmp(line,"%a ",3) )
	{
	    unsigned short flags;

	    sscanf( line, "%%a %*x %d %x", &id, &flags );
	    srecs[id].flags = flags;
	    continue; 
	}

	if( !strncmp(line,"%d ",3) )
	{
	    sscanf( line, "%%d %*x %d", &id );
	    if( srecs[id].filename != NULL )
		free( srecs[id].filename );
	    if( srecs[id].buf != NULL )
		free( srecs[id].buf );
	    srecs[id].filename = NULL;
	    srecs[id].buf = NULL;
	    srecs[id].len = 0;
	    continue; 
	}

	if( !strncmp(line,"%e ",3) )
	{
	    for( id=0; id<SFX_MAX_SOUNDS; ++id )
	    {
		if( srecs[id].filename != NULL )
		    free( srecs[id].filename );
		if( srecs[id].buf != NULL )
		    free( srecs[id].buf );
		srecs[id].filename = NULL;
		srecs[id].buf = NULL;
	    }
	    continue; 
	}

	if( !strncmp(line,"%g ",3) )
	{
	    char tmp[1023+1];

	    sscanf( line, "%%g %*x %s", tmp );
	    if( glob != NULL )
		delete glob;
	    /*fprintf( stdout, "TMP nw glob %s\n", tmp );*/
	    glob = new Glob( tmp );
	    continue; 
	}

	if( !strncmp(line,"%h ",3) )
	{
	    for( rc=0; rc<SFX_MAX_CHANS; ++rc )
	    {
		chans[rc].cur = NULL;
		chans[rc].begin = NULL;
		chans[rc].end = NULL;
	    }
	    if( (rc=ioctl(fdAudio,SNDCTL_DSP_RESET,0)) == -1 )
	    {
		fprintf( stdout, "ERROR. Bad SNDCTL_DSP_RESET. %s.\n",
		    sys_errlist[errno] );
	    }
	    continue;
	}

	if( !strncmp(line,"%l ",3) )
	{
	    char tmp[1023+1];
	    unsigned short flags;

	    sscanf( line, "%%l %*x %d %x %s", &id, &flags, tmp );
	    /*
	    fprintf( stdout, "TMP link <%d %x \"%s\">\n", id, flags, tmp );
	    */
	    if( srecs[id].filename != NULL )
		free( srecs[id].filename );
	    if( srecs[id].buf != NULL )
		free( srecs[id].buf );
	    srecs[id].buf = NULL;
	    srecs[id].filename = strdup(tmp);
	    srecs[id].flags = flags;
	    continue; 
	}

	if( !strncmp(line,"%m ",3) )
	{
	    unsigned int toggle;

	    sscanf( line, "%%m %*x %u", &toggle );
	    /*
	    fprintf( stdout, "TMP mute <%d>\n", toggle );
	    */
	    globalMute = toggle;
	    continue; 
	}

	if( !strncmp(line,"%o ",3) )
	{
	    unsigned char *data;
	    unsigned short flags;
	    int len;

	    sscanf( line, "%%o %*x %d", &id, &flags );
	    if( srecs[id].filename == NULL )
		continue;
	    if( (data=child_fetch(glob,srecs[id].filename,&len)) == NULL )
		continue;
	    srecs[id].flags = flags;
	    srecs[id].buf = data;
	    srecs[id].len = len;
	    continue; 
	}

	if( !strcmp(line,"%s") )
	{
	    if( (rc=ioctl(fdAudio,SNDCTL_DSP_SYNC,0)) == -1 )
	    {
		fprintf( stdout, "ERROR. Bad sync of audio. %s.\n",
		    sys_errlist[errno] );
	    }
	    continue; 
	}

	if( !strncmp(line,"%v ",3) )
	{
	    int vol;

	    sscanf( line, "%%v %*x %d", &vol );
	    globalVolume = (float)vol/100.0;
	    /*
	    fprintf( stdout, "TMP volume %d -> %1.02f.\n",
		vol, globalVolume );
	    */

	    continue;
	}

	/* Any line which is not a command must be a file name. */
	if( line[0] == '%' )
	{
	    /*fprintf( stdout, "TMP not r <%s>\n", cmd );*/
	    continue;
	}

	if( globalMute )
	    continue;

	id = atoi(line);
	//fprintf( stdout, "TMP Play %d.\n", id );fflush(stdout);
	child_addsfx(glob,id);
	++flagPlay;
    }

    //fprintf( stdout, "TMP play %d\n", flagPlay );
    return flagPlay;
}
#endif

#ifdef HAVE_OSS
static void child_intr( int signo )
{
    if( signo == intr_quit )
    {
	short id;

	if( glob != NULL )
	{
	    delete glob;
	    glob = NULL;
	}
	for( id=0; id<SFX_MAX_SOUNDS; ++id )
	{
	    if( srecs[id].filename != NULL )
		free( srecs[id].filename );
	    if( srecs[id].buf != NULL )
		free( srecs[id].buf );
	}
	if( srecs != NULL )
	    free( srecs );
	close( fdPipe[0] );
	close( fdAudio );
	exit( 0 );
    }
}
#endif

#ifdef HAVE_OSS
static int child_update()
{
    int i, j, d;
    int flagUse = 0;

    /*
    ** Since we write full audio fragments at a time, some data may
    ** not be valid sound data.
    */
    memset( mixbuffer, 0, mixsize );
    /*fputs( "TMP st\n", stdout );*/
    for( i=0; i<mixsize; ++i )
    {
	d = 0;
	j = SFX_MAX_CHANS;
	flagUse = 0;
	do
	{
	    if( chans[--j].cur != NULL )
	    {
		++flagUse;
		d += *chans[j].cur++ - 128;
	    }
	} while( j );
	if( !flagUse )
	{
	    /*fputs( "TMP fl\n", stdout );*/
	    break;
	}
	if( d > 127 )
	    mixbuffer[i] = 255;
	else if( d < -128 )
	    mixbuffer[i] = 0;
	else
	    mixbuffer[i] = (unsigned char)(d+128);
	mixbuffer[i] *= globalVolume;
	for( j=0 ; j<SFX_MAX_CHANS ; ++j )
	{
	    if( chans[j].cur == NULL )
		continue;
	    if (chans[j].cur >= chans[j].end)
	    {
		int id = chans[j].id;
		/*
		fprintf( stdout, "TMP free slot %d %x %x.\n",
		    j, chans[j].cur,
		    chans[j].end );
		*/
		chans[j].cur = NULL;
		chans[j].begin = NULL;
		chans[j].end = NULL;
		if( !(srecs[id].flags&SFX_SFLG_KEEP) )
		{
		    free( srecs[id].buf );
		    srecs[id].buf = NULL;
		}
	    }
	}
    }
    /*fprintf( stdout, "TMP w %d\n", i );*/
    if( i > 0 )
	write( fdAudio, &(mixbuffer[0]), mixsize );
    if( !flagUse )
    {
	/*fputs( "TMP po\n", stdout );*/
	ioctl( fdAudio, SNDCTL_DSP_POST, 0 );
	return 1;
    }

    return 0;
}
#endif

#ifdef HAVE_OSS
static unsigned char *child_fetch( Glob *glob,
    const char *name, int *length )
{
    unsigned char *buf;
    int fd;
    int len;
    struct stat statrec;

    /*
    fprintf( stdout, "TMP fetch %s\n", name ); fflush( stdout );
    */
    /*
    ** If name is a component of the current glob.
    */
    if( (name[0]=='G') && (name[1]==':') )
    {
	/*fprintf( stdout, "TMP glb %s\n", name ); fflush( stdout );*/
	if( (buf=(unsigned char*)glob->extract(&name[2],length)) == NULL )
	    return NULL;
	return buf;
    }

    /*
    ** If name is a file.
    */

    if( stat(name,&statrec) == -1 )
    {
	fprintf( stdout, "ERROR. Bad stat of \"%s\". %s.\n",
	    name, sys_errlist[errno] );
	return NULL;
    }
    /*
    fprintf( stdout, "TMP st_size %d\n", statrec.st_size );
    fflush( stdout );
    */

    if( (fd=open(name,O_RDONLY)) == -1 )
    {
	fprintf( stdout, "ERROR. Bad open of \"%s\". %s.\n",
	    name, sys_errlist[errno] );
	return NULL;
    }

    len = statrec.st_size;

    /*fprintf( stdout, "TMP len %d\n", len ); fflush( stdout );*/
    if( (buf=(unsigned char*)malloc(len)) == NULL )
    {
	fprintf( stdout, "ERROR. Bad malloc of %d bytes.\n", len );
	return NULL;
    }
    /*fprintf( stdout, "TMP fd %d len %d\n", fd, len ); fflush( stdout );*/

    if( read(fd,buf,len) == -1 )
    {
	fprintf( stdout, "ERROR. Bad read. %s.\n",
	    sys_errlist[errno] );
	free( buf );
	return NULL;
    }
    /*fprintf( stdout, "TMP read %d\n", len ); fflush( stdout );*/

    close( fd );

    *length = len;

    return buf;
}
#endif

#ifdef HAVE_OSS
static int child_addsfx( Glob *glob, short id )
{
  int i;
  int rc = -1;
  int len;
  unsigned char *data;
  int oldest = mytime;
  int oldestnum = 0;

    /*
    ** Load sound data from file if it is not in the cache.
    */
    if( srecs[id].buf == NULL )
    {
	/*fprintf( stdout, "TMP ld %d\n", id ); fflush( stdout );*/
	if( srecs[id].filename == NULL )
	{
	    //fprintf( stdout, "TMP no ln %d\n", id );
	    return 0;
	}
	if( (data=child_fetch(glob,srecs[id].filename,&len)) == NULL )
	    return 0;
	srecs[id].buf = data;
	srecs[id].len = len;
    }
    else
    {
	/*fprintf( stdout, "TMP in %d\n", id ); fflush( stdout );*/
	data = srecs[id].buf;
	len = srecs[id].len;
    }
  for (i=0 ; i<SFX_MAX_CHANS ; i++)
  {
    if (!chans[i].cur)
    {
	/*
	fprintf( stdout, "TMP %s in ch%d %x %x\n",
	    srecs[id].filename, i, data, (data+len) );
	*/
      chans[i].cur = data;
      chans[i].begin = data;
      chans[i].end = (data+len);
      chans[i].time = mytime;
      break;
    }
    else
    {
      if (chans[i].time < oldest)
      {
	/*fprintf( stdout, "TMP. new oldest slot %d.\n", i );*/
        oldestnum = i;
	oldest = chans[i].time;
      }
    }
  }

  /* if no channels were available, kill oldest sound and replace it */
  if (i == SFX_MAX_CHANS)
  {
    /*fprintf( stdout, "TMP no ch. old %d\n", oldestnum );*/
    chans[oldestnum].cur = data;
    chans[oldestnum].begin = data;
    chans[oldestnum].end = chans[oldestnum].begin + len;
    chans[oldestnum].time = mytime;
  }

  return rc;
}
#endif
