/*
** 
**  Copyright (c) 1998 by Larry Plona
** 
**  This distribution  of XPipeman is copyright by Larry Plona.
**  The  package  and  its  component  files  may  be  freely
**  distributed as long as  all copyright notices remain intact
**  and this distribution is not sold  as  commercial software.
** 
**  No warranty of any kind is attached with this software.  It
**  is supplied as is.
** 
** ----------------------------------------------------------------
*/

#include "tcache.h"

/*
** Class TCache. Container class for TObj objects. Handles TObj creation
** from a Glob. Supports searching and naming.
**
** A Glob must be registered.
*/

/*
** fetch()
*/
TObj *TCache::fetch( int type )
{
    const char *buf;

    tnode *p = root;
    TObj *to = NULL;

    if( type >= count )
	return NULL;

    while( p != NULL )
    {
	if( p->tobj->type == type )
	    return p->tobj;
	p = p->next;
    }

    if( (buf=glob->extract(names[type],NULL)) == NULL )
    {
	fprintf( stdout, "ERROR. Bad glob.extract of %s.\n", names[type] );
	return 0;
    }

    to = new TObj( type, colores, buf );
    if( type == -1 )
	return NULL;
    if( (p=(struct tnode*)malloc(sizeof(struct tnode))) == NULL )
    {
	fprintf( stdout, "ERROR. Bad malloc. %s.\n", sys_errlist[errno] );
	return NULL;
    }

    p->next = root;
    p->prev = NULL;
    p->tobj = to;
    if( root != NULL )
	root->prev = p;
    root = p;

    return p->tobj;
}

/*
** remove()
*/
void TCache::remove( int type )
{
    struct tnode *p = root;

    while( p != NULL )
    {
	if( p->tobj->type == type )
	{
	    if( p->prev != NULL )
		p->prev->next = p->next;
	    if( p->next != NULL )
		p->next->prev = p->prev;
	    delete p->tobj;
	    free( p );
	    return;
	}
	p = p->next;
    }
}

int TCache::setMax( int nmax )
{
    if( nmax < 1 )
    {
	if( names != NULL )
	{
	    //fputs( "TMP. Freeing names.\n", stdout );
	    for( int i=0; i<count; ++i )
		free( names[i] );
	    free( names );
	}
	names = NULL;
	max = 0;
	count = 0;
	return 1;
    }
    if( names != NULL )
    {
	//fputs( "TMP Realloc of names\n", stdout );
	if( (names=(char**)realloc(names,sizeof(char*)*nmax)) == NULL )
	    return 0;
	max = nmax;
	return 1;
    }

    if( (names=(char**)malloc(sizeof(char*)*nmax)) == NULL )
	return 0;
    count = 0;
    max = nmax;
    return 1;
}

int TCache::notify( const char *name )
{
    int nid;

    if( names == NULL )
    {
	/*
	fprintf( stdout,
	    "TMP. Can not notify \"%s\". TCache not initialized.\n",
	    name );
	*/
	return -1;
    }

    if( count >= max )
    {
	/*
	fprintf( stdout,
	    "TMP. Can not notify \"%s\". TObjs count at max of %d.\n",
	    max );
	*/
	return -1;
    }

    /*
    fprintf( stdout,
	"TMP. Register \"%s\". %d of %d.\n", name, count, max );
    */

    names[count] = strdup( name );
    nid = count;
    ++count;

    return nid;
}

int TCache::reset( void )
{
    struct tnode *p = root;

    setMax( 0 );
    if( glob != NULL )
	glob = NULL;

    while( p != NULL )
    {
	if( p->prev != NULL )
	{
	    delete p->prev->tobj;
	    free( p->prev );
	}
	if( p->next == NULL )
	{
	    delete p->tobj;
	    free( p );
	    break;
	}
	p = p->next;
    }

    root = NULL;

    return 1;
}

/*
** setGlob()
** 
** Set current current glob for all TObjs.
*/
void TCache::setGlob( Glob *nglob )
{
    glob = nglob;
}

/*
** setColores()
*/
void TCache::setColores( Colores *cmap )
{
    colores = cmap;
}

const char *TCache::name( int type )
{
    if( type > count )
	return NULL;
    return (const char*)(names[type]);
}
