/*
** 
**  Copyright (c) 1998 by Larry Plona
** 
**  This distribution  of XPipeman is copyright by Larry Plona.
**  The  package  and  its  component  files  may  be  freely
**  distributed as long as  all copyright notices remain intact
**  and this distribution is not sold  as  commercial software.
** 
**  No warranty of any kind is attached with this software.  It
**  is supplied as is.
** 
** ----------------------------------------------------------------
*/

#include "tobj.h"

extern int errno;
extern char *sys_errlist[];

#if 0 /* Debug. */
static void dump_extension( XpmAttributes* );
#endif

/*
** The indicies must match the values of enum tobj_dir.
*/
struct tobj_dir_idx tobj_dirs[4] =
{
    { -1,  0 }, // Left.
    {  1,  0 }, // Right.
    {  0, -1 }, // Upward.
    {  0,  1 }  // Downward.
};

TObj::TObj( int ntype, Colores *colores, const char *buf )
{
    type = ntype;
    tclass = tobj_tclass_null;
    framecnt = 0;
    /*
    ** FUTURE. Loading can fail for many reasons yet there is no way
    ** to pass this failure up.
    */
    if( !load(colores,buf) )
	type = -1;
}

TObj::~TObj()
{
    if( frames != NULL )
    {
	for( int i=0; i<framecnt; ++i )
	    XDestroyImage( frames[i] );
    }
}

int TObj::load( Colores *colores, const char *buf )
{
    int i, j;
    int hasAnim = 0;  // Has ANIM extension.
    XImage *imgFirst;
    XpmAttributes attr;
    struct tobj_info_anim anim;

    frames = NULL;

    attr.valuemask = XpmReturnExtensions;
    if( (j=colores->createImageFromBuffer((char*)buf,&imgFirst,
	NULL,&attr)) != XpmSuccess )
    {
	free( (void*)buf );
	fprintf( stdout, "Bad XpmCreateImageFromBuffer. %s.\n",
	    XpmGetErrorString(j) );
	return 0;
    }
    /*
    if( attr.nextensions > 1 )
	dump_extension( &attr );
    */
    framecnt = 1;

    /*
    ** Count the number of additional frames. Space for the frames must
    ** be allocated.
    */
    for( i=0; i<attr.nextensions; ++i )
    {
	if( !strncmp(attr.extensions[i].name,"XPMPIXMAP",9) )
	    ++framecnt;
    }

    if( (frames=(XImage**)malloc(sizeof(XImage**)*framecnt)) == NULL )
    {
	free( (void*)buf );
	fprintf( stdout, "ERROR. Bad malloc. %s.\n", sys_errlist[errno] );
	return 0;
    }
    frames[0] = imgFirst;

    /*
    ** Go through the extensions, this time to read the image data.
    */
    j = 1; /* Use as index into frames. */
    for( i=0; i<attr.nextensions; ++i )
    {
	/*
	** ANIM extension. Note that this extension is not mutually
	** exclusive with some TOBJ classes. When an ANIM extension is
	** encountered, the information is loaded into a temporary
	** structure until its final disposition is determined.
	*/
	if( !strncmp(attr.extensions[i].name,"ANIM",4) )
	{
	    char *line;

	    if( hasAnim )
	    {
		fputs( "WARNING. Multiple ANIM extensions.\n", stdout );
		continue;
	    }
	    hasAnim = 1;
	    anim.series[0] = -1;
	    anim.primus = 0;
	    for( int k=0; k<attr.extensions[i].nlines; ++k )
	    {
		line = attr.extensions[i].lines[k];
		if( !strncmp(line,"series ",7) )
		{
		    if( !parse_series(&anim,line,framecnt) )
		    {
			fprintf( stdout, "ERROR. Bad \"series\" line.\n" );
			return 0;
		    }
		    continue;
		}
		if( !strncmp(line,"primus ",7) )
		{
		    if( !parse_primus(&anim,line,framecnt) )
		    {
			fprintf( stdout, "ERROR. Bad \"primus\" line.\n" );
			return 0;
		    }
		    continue;
		}
		if( !strncmp(line,"mora ",5) )
		{
		    if( !parse_mora(&anim,line,framecnt) )
		    {
			fprintf( stdout, "ERROR. Bad \"mora\" line.\n" );
			return 0;
		    }
		    continue;
		}
	    }
	    if( anim.series[0] == -1 )
		make_series(&anim);
	    continue;
	} /* ANIM extension. */
	if( !strncmp(attr.extensions[i].name,"TOBJ",4) )
	{
	    if( !parse_tobj(&(attr.extensions[i])) )
	    {
		fputs( "ERROR. Can not parse TOBJ extension.\n", stdout );
		return 0;
	    }
	    continue;
	}
	if( !strncmp(attr.extensions[i].name,"XPMPIXMAP",9) )
	{
	    char **data;
	    int rc;

	    data = attr.extensions[i].lines;
	    if( (rc=colores->createImageFromData((const char**)data,
		&frames[j],NULL,NULL)) != XpmSuccess )
	    {
		fprintf( stdout, "ERROR. Bad XpmCreateImageFromData. %s.\n",
		    XpmGetErrorString(rc) );
		return 0;
	    }
	    ++j;
	}
    }

    // XpmFreeExtensions( attr.extensions, attr.nextensions );
    XpmFreeAttributes( &attr );

    // Free copy of glob component.
    free( (void*)buf );

    /*
    ** Some TObj classes require an ANIM extension. Add the collected
    ** extension now.
    */
    switch( tclass )
    {
	case tobj_tclass_obstacle:
	    //fprintf( stdout, "TMP load %d obstacle class\n", type );
	    break;
	case tobj_tclass_start:
	    //fprintf( stdout, "TMP load %d start class\n", type );
	    break;
	case tobj_tclass_null:
	    if( hasAnim )
	    {
		//fprintf( stdout, "TMP load %d anim class\n", type );
		tclass = tobj_tclass_anim;
		tinfo.anim = anim;
	    }
	    else
	    {
		//fprintf( stdout, "TMP load %d picture class\n", type );
		tclass = tobj_tclass_picture;
	    }
	    break;
	case tobj_tclass_pipe:
	    //fprintf( stdout, "TMP load %d pipe class\n", type );
	    break;
	case tobj_tclass_meter:
	    //fprintf( stdout, "TMP load %d meter class\n", type );
	    break;
	default:
	    //fprintf( stdout, "TMP load unknown class, %d\n", tclass );
	    break;
    }

    /*
    fprintf( stdout, "TMP tobj load %d %d frames\n",
	type, framecnt );
    fflush( stdout ); // TMP.
    */
    return 1;
}

void *TObj::initstate( int *tag )
{
    *tag = -1;

    switch( tclass )
    {
	case tobj_tclass_pipe:
	case tobj_tclass_start:
	{
	    struct tobj_state_pipe *s;
	    struct tobj_info_pipe *info;

	    if( tclass == tobj_tclass_start )
		info = &tinfo.start.pipe;
	    else
		info = &tinfo.pipe;

	    if( (s=(struct tobj_state_pipe*)malloc(
		sizeof(struct tobj_state_pipe))) == NULL )
	    {
		fprintf( stdout, "ERROR. Bad malloc. %s.\n",
		    sys_errlist[errno] );
		return NULL;
	    }

	    s->seq_idx = 0;
	    s->seq_id = -1;

	    /*
	    ** Pre flow situation. Either still frame or
	    ** animation sequence.
	    */
	    s->base_id = 0;
	    if( (info->bases[s->base_id].flags&0x01) == 0 )
	    {
		// Animation sequence.
		s->seq_id = info->bases[s->base_id].baseframe;
		s->frame_idx = info->seqs[s->seq_id].frames[0];
		s->seq_idx = 0;
		if( info->seqs[s->seq_id].framecnt > 1 )
		    *tag = (info->seqs[s->seq_id].flags&0x0c)>>2;
/*
		fprintf( stdout,
		    "TMP flow initstate pipe t%d anim s%d sync %d\n",
		    type, s->seq_id, *tag );
*/
	    }
	    else
	    {
		// Static frame.
		s->seq_id = -1;
		s->frame_idx = info->bases[s->base_id].baseframe;
	    }
	    return s;
	}
	case tobj_tclass_meter:
	{
	    struct tobj_state_meter *s;

	    if( (s=(struct tobj_state_meter*)malloc(
		sizeof(struct tobj_state_meter))) == NULL )
	    {
		fprintf( stdout, "ERROR. Bad malloc. %s.\n",
		    sys_errlist[errno] );
		return NULL;
	    }

	    s->layer = 0;

	    if( (tinfo.meter.flags[s->layer]&0x01) == 0 )
	    {
		// Animation sequence.
		s->seq_id = tinfo.meter.ids[s->layer];
		s->seq_idx = 0;
		s->frame_idx = tinfo.meter.seqs[s->seq_id].frames[0];
		if( tinfo.meter.seqs[s->seq_id].framecnt > 1 )
		    *tag = (tinfo.meter.seqs[s->seq_id].flags&0x0c)>>2;
	    }
	    else
	    {
		// Static frame.
		s->seq_id = -1;
		s->frame_idx = tinfo.meter.ids[s->layer];
	    }

	    return s;
	}
	case tobj_tclass_obstacle:
	{
	    struct tobj_state_obstacle *s;
	    struct tobj_info_obstacle *info = &tinfo.obstacle;

	    *tag = -1;

	    if( (s=(struct tobj_state_obstacle*)malloc(
		sizeof(struct tobj_state_obstacle))) == NULL )
	    {
		fprintf( stdout, "ERROR. Bad malloc. %s.\n",
		    sys_errlist[errno] );
		return NULL;
	    }

	    s->base_id = 0;
	    s->seq_id = -1;
	    s->seq_idx = -1;

	    // Set up initial animation sequence or frame.

	    if( (info->bases[s->base_id].flags&0x01) == 1 )
		s->frame_idx = info->bases[s->base_id].baseframe;
	    else
	    {
		s->seq_id = info->bases[s->base_id].baseframe;
		s->seq_idx = 0;
		s->frame_idx = info->seqs[s->seq_id].frames[s->seq_idx];
		if( info->seqs[s->seq_id].framecnt > 1 )
		    *tag = (info->seqs[s->seq_id].flags&0x0c)>>2;
	    }

	    return s;
	}
	case tobj_tclass_anim:
	{
	    struct tobj_state_anim *s;
	    struct tobj_info_anim *ia = &tinfo.anim;

	    if( (s=(struct tobj_state_anim*)malloc(
		sizeof(struct tobj_state_anim))) == NULL )
	    {
		fprintf( stdout, "ERROR. Bad malloc. %s.\n",
		    sys_errlist[errno] );
		return NULL;
	    }
	    if( framecnt>1 &&  ia->primus<0 )
		s->frameidx = ((int)rand()%(framecnt));
	    else
		s->frameidx = ia->primus;
	    if( framecnt > 1 )
		*tag = 0; // Accept animation interrupts.
	    return s;
	}
	case tobj_tclass_picture:
	    /*
	    ** The picture class has no state info.
	    */
	    return (void*)1;
	default:
	    return NULL;
    }
}

XImage *TObj::incanim( void *state_info )
{
    switch( tclass )
    {
	case tobj_tclass_pipe:
	case tobj_tclass_start:
	{
	    struct tobj_state_pipe *s =
		(struct tobj_state_pipe*)state_info;
	    struct tseq *tseq;

	    // Can this happen?
	    if( s->seq_id < 0 )
		return frames[s->frame_idx];

	    tseq = &tinfo.pipe.seqs[s->seq_id];

	    if( (tseq->flags&3) == 2 )
	    {
		if( (s->seq_idx>=tseq->framecnt) ||
		    (tseq->frames[s->seq_idx]==-1) )
		{
		    s->seq_idx = 0;
		}
		s->frame_idx = tseq->frames[s->seq_idx];
		/*
		fprintf( stdout, "TMP to anim t%d idx %d(fr %d)\n",
		    type, s->seq_idx, s->frame_idx );
		*/
		++s->seq_idx;
		return frames[s->frame_idx];
	    }
	    /*
	    fprintf( stdout, "TMP incanim pipe set, anim t%d\n", type );
	    */
	    return frames[s->frame_idx];
	}
	case tobj_tclass_obstacle:
	{
	    struct tobj_state_obstacle *s =
		(struct tobj_state_obstacle*)state_info;
	    struct tseq *tseq;

	    // Static frame.
	    if( s->seq_id < 0 )
		return frames[s->frame_idx];

	    // Animation sequence.

	    tseq = &tinfo.obstacle.seqs[s->seq_id];

	    if( (s->seq_idx>=tseq->framecnt) ||
		(tseq->frames[s->seq_idx]==-1) )
	    {
		s->seq_idx = 0;
	    }
	    s->frame_idx = tseq->frames[s->seq_idx];
	    /*
	    fprintf( stdout, "TMP to obs t%d idx %d(fr %d)\n",
		type, s->seq_idx, s->frame_idx );
	    */
	    ++s->seq_idx;
	    return frames[s->frame_idx];
	}
	case tobj_tclass_meter:
	{
	    struct tobj_state_meter *s =
		(struct tobj_state_meter*)state_info;
	    struct tseq *tseq;

	    // Static frame.
	    if( s->seq_id == -1 )
		return frames[s->frame_idx];

	    // Animation sequence.
	    tseq = &tinfo.meter.seqs[s->seq_id];

	    if( (s->seq_idx>=tseq->framecnt) ||
		(tseq->frames[s->seq_idx]==-1) )
	    {
		s->seq_idx = 0;
	    }
	    s->frame_idx = tseq->frames[s->seq_idx];
	    /*
	    fprintf( stdout, "TMP to anim meter t%d seq %d idx %d\n",
		type, s->seq_id, s->seq_idx );
	    */
	    ++s->seq_idx;
	    return frames[s->frame_idx];
	}
	case tobj_tclass_anim:
	{
	    struct tobj_state_anim *s = (struct tobj_state_anim*)state_info;

	    if( s->frameidx == -1 )
		s->frameidx = tinfo.anim.primus;
	    if( tinfo.anim.series[s->frameidx] == -1 )
		s->frameidx = 0;

/*
if( type == 6 )
{
    fprintf( stdout, "TMP to anim anim t%d frame %d\n",
	type, s->frameidx );
    fprintf( stdout, "TMP \" %d %d %d %d %d %d\n",
	tinfo.anim.series[0], tinfo.anim.series[1],
	tinfo.anim.series[2], tinfo.anim.series[3],
	tinfo.anim.series[4], tinfo.anim.series[5] );
}
*/

	    ++s->frameidx;
	    return frames[tinfo.anim.series[s->frameidx-1]];
	}
	case tobj_tclass_picture:
	    /*
	    fprintf( stdout, "TMP incanim picture class, anim t%d\n", type );
	    */
	    return frames[0];
	default:
	    fprintf( stdout, "ERROR. incanim unknown class %d, anim t%d\n",
		tclass, type );
	    return NULL;
    }
}

/*
** latest()
**
** Return the current frame.
**
** FUTURE. This function is required only because all state info (even
** current frame) is tclass specific. Maybe frame_idx should be part
** of a standard state info header.
*/
XImage *TObj::latest( void *state_info )
{
    switch( tclass )
    {
	case tobj_tclass_pipe:
	case tobj_tclass_start:
	{
	    struct tobj_state_pipe *s =
		(struct tobj_state_pipe*)state_info;

	    return frames[s->frame_idx];
	}
	case tobj_tclass_obstacle:
	{
	    struct tobj_state_obstacle *s =
		(struct tobj_state_obstacle*)state_info;

	    return frames[s->frame_idx];
	}
	case tobj_tclass_meter:
	{
	    struct tobj_state_meter *s =
		(struct tobj_state_meter*)state_info;

	    return frames[s->frame_idx];
	}
	case tobj_tclass_anim:
	{
	    struct tobj_state_anim *s = (struct tobj_state_anim*)state_info;

	    return frames[tinfo.anim.series[s->frameidx]];
	}
	case tobj_tclass_picture:
	    return frames[0];
	default:
	    fprintf( stdout, "ERROR. incanim unknown class %d, anim t%d\n",
		tclass, type );
	    return NULL;
    }
}

XImage *TObj::incflow( tobj_dir *dir, int *tag, struct tobj_state_pipe *s,
   struct tobj_gain *gain )
{
    struct tseq *tseq = NULL;
    tobj_dir olddir = *dir;

    gain->newtile = 0;
    gain->rating = 0;
    gain->cross = 0;
    *tag = -1;

    /*
    if( (tclass==tobj_tclass_pipe) || (tclass==tobj_tclass_start) )
    {
	fprintf( stdout, "TMP to flow t%d base %d seq %d(%d) dir %d\n",
	    type, s->base_id, s->seq_id, s->seq_idx, *dir );
    }
    */

    // Indicate no change in direction.
    *dir = tobj_dir_null;

    if( tclass == tobj_tclass_obstacle )
    {
	struct tobj_state_obstacle *so = (struct tobj_state_obstacle*)s;

	//fprintf( stdout, "TMP to flow obstacle t%d\n", type );

	/* Set up new flow sequence. */
	so->seq_id = tinfo.obstacle.bases[so->base_id].seqs[olddir];
	so->seq_idx = 0;

	/*
	fprintf( stdout, "TMP to flow obs seq %d (from b%02d, dir %d)\n",
	    so->seq_id, so->base_id, olddir );
	*/

	if( so->seq_id == -1 )
	{
	    /* FUTURE. Error. What is the correct response? */
	    so->frame_idx = 0;
	}
	else
	{
	    tseq = &tinfo.obstacle.seqs[so->seq_id];
	    so->frame_idx = tseq->frames[so->seq_idx];
	    if( tseq->framecnt > 1 )
		*tag = (tseq->flags&0x0c)>>2;
	    /*
	    ** Because an obstacle can both terminate flow and show
	    ** a new frame, the sync tag must be set to zero to indicate
	    ** that a new frame is to be shown.
	    */
	    if( *tag == -1 )
		*tag = 0;
	}

	// Indicate flow termination.
	return NULL;
    }

    /*
    ** If this is a start pipe, there is no current direction. Use
    ** the bootstrap direction stored in this TObj itself.
    */
    if( tclass == tobj_tclass_start )
    {
	if( s->base_id == 0 )
	    olddir = tinfo.start.initial_dir;
    }

    if( (tclass!=tobj_tclass_pipe) && (tclass!=tobj_tclass_start) )
    {
	/*
	fprintf( stdout, "TMP to flow not pipe t%d tclass %d\n",
	    type, tclass );
	*/
	// Indicate flow termination.
	return NULL;
    }

    /*
    ** Pipe tile. The possibilities.
    **
    ** o Tile is in an animation sequence. This can only happen when
    **   once per tile. Abort animation. Consult current base for
    **   new flow sequence.
    ** o Flow sequence is in effect. If at end, invoke new base and
    **   move flow to a new tile. Otherwise increment flow in tile.
    ** o Tile has no sequence. Initiate new flow sequence from info
    **   in current base (like the first possibility).
    */

    // Current sequence if any. May be flow or animation sequence.
    if( s->seq_id >= 0 )
	tseq = &tinfo.pipe.seqs[s->seq_id];

    /*
    ** Current tile is being encountered for the first time. It may
    ** have an animation sequence in effect (it will be aborted).
    ** Determine the new flow sequence to set up from current base.
    */
    if( (s->seq_id<0) || (tseq&&((tseq->flags&3)==2)) )
    {
#if 0
	if( (tseq!=NULL) && ((tseq->flags&3)==2) )
	{
	    //fprintf( stdout, "TMP to flow encounter anim, t%d\n", type );
	    /*
	    ** FUTURE. Cheap way to distinguish between an animation
	    ** sequence representing a previous flow and an animation
	    ** sequence representing an empty pipe.
	    */
	    if( s->base_id == 0 )
		gain->newtile = 1;
	    else
	    {
		//fprintf( stdout, "TMP cross t%2d\n", type );
		gain->cross = 1;
	    }
	}
	else
	{
	    gain->newtile = 1;
	    fprintf( stdout, "TMP to flow encounter new tile, t%d\n", type );
	}
#endif
	if( s->base_id == 0 )
	    gain->newtile = 1;
	else
	    gain->cross = 1;

	/* Set up new flow sequence. */
	s->inFlow = 1;
	s->seq_id = tinfo.pipe.bases[s->base_id].seqs[olddir];
	s->seq_idx = 0;

	/*
	fprintf( stdout, "TMP to flow new flow seq %d (base %d dir %d)\n",
	    s->seq_id, s->base_id, olddir );
	*/

	// Direction may not be supported by the current base. Crash.
	if( s->seq_id == -1 )
	{
	    /*
	    fprintf( stdout, "TMP to flow no new dir for dir %d in t%d\n",
	       olddir, type );
	    */
	    // Indicate flow termination.
	    return NULL;
	}

	tseq = &tinfo.pipe.seqs[s->seq_id];

	// Continue.
    }

    /*
    fprintf( stdout,
	"TMP to flow base %d seq %d seqidx %d dir %d\n",
	s->base_id, s->seq_id, s->seq_idx, olddir );
    */

    /*
    ** TObj frame index. Store for use by incanim() and other routines
    ** which want to redraw the current frame.
    */
    s->frame_idx = tseq->frames[s->seq_idx];

    // If this is the last frame in the sequence.
    if( s->seq_idx >= tseq->framecnt )
    {
	/*
	** One of three things can happen at the end of a sequence.
	**
	** o Switch to new base. Flow moves to new tile.
	** o Start a new flow sequence in the current tile.
	*/

	if( (tseq->flags&0x03) == 2 )
	{
	    fprintf( stdout, "ERROR. Anim sequence in incflow().\n" );
	    return NULL;
	}

	/*
	** New flow sequence.
	*/
	if( (tseq->flags&3) == 1 )
	{
	    s->seq_id = tseq->newbase;
	    tseq = &tinfo.pipe.seqs[s->seq_id];
	    s->frame_idx = tseq->frames[0];
	    s->seq_idx = 0;

	    //fprintf( stdout, "TMP to flow 2nd seq %d\n", s->seq_id );
	    return frames[s->frame_idx];
	}

	/*
	** flags&3 == 0. A new base occupies the tile.
	*/
	*dir = tseq->newdir;
	new_tobjbase( tseq->newbase, s, tag );
	return frames[s->frame_idx];
    }

    /*
    fprintf( stdout, "TMP to flow inc frame %d (%d of seq %d)\n",
	tseq->frames[s->seq_idx], s->seq_idx, s->seq_id );
    */
    ++s->seq_idx;

    return frames[s->frame_idx];
}

int TObj::incmeter( int *tag, struct tobj_state_meter *s,
    XImage **fimg )
{
    struct tseq *tseq;

    *tag = -1;

    /*
    ** Each layer is associated with a sequence or frame. Sequences are
    ** animation sequences and are not cycled by this interrupt. This
    ** function merely sets up a sequence or frame based on the current
    ** layer.
    */

    ++s->layer;

    /*
    ** Indicate that this TObj has exhausted its layers.
    */
    if( s->layer > 2 )
	return 2;

    // Static frame.
    if( (tinfo.meter.flags[s->layer]&0x01) == 1 )
    {
	s->seq_id = -1;
	s->frame_idx = tinfo.meter.ids[s->layer];
	*fimg = frames[s->frame_idx];
	return 1;
    }

    // Set up an animation sequence.

    s->seq_id = tinfo.meter.ids[s->layer];
    s->seq_idx = 0;
    tseq = &(tinfo.meter.seqs[s->seq_id]);

    /*
    fprintf( stdout, "TMP to meter layer %d seq %d frame %d\n",
	s->layer, s->seq_id, s->seq_idx );
    */

    s->frame_idx = tseq->frames[s->seq_idx];

    if( tseq->framecnt > 1 )
	*tag = (tseq->flags&0x0c)>>2;

    *fimg = frames[s->frame_idx];

    return 1;
}

int TObj::parse_tobj( XpmExtension *e )
{
    char *line;
    int k;

    {
	char c;
	char *line = e->lines[0];

	if( sscanf(line,"c %c\n",&c) != 1 )
	{
	    fprintf( stdout, "ERROR. Bad TOBJ. \"%s\".\n", line );
	    return 0;
	}
	switch( c )
	{
	    case 'm':
		tclass = tobj_tclass_meter;
		tinfo.meter.seqcnt = 0;
		break;
	    case 'o':
		tclass = tobj_tclass_obstacle;
		tinfo.obstacle.basecnt = 0;
		tinfo.obstacle.seqcnt = 0;
		break;
	    case 'p':
		tclass = tobj_tclass_pipe;
		tinfo.pipe.basecnt = 0;
		tinfo.pipe.seqcnt = 0;
		break;
	    case 's':
		tclass = tobj_tclass_start;
		tinfo.start.initial_dir = tobj_dir_right;
		tinfo.start.pipe.basecnt = 0;
		tinfo.start.pipe.seqcnt = 0;
		break;
	    default:
		fprintf( stdout, "ERROR. Bad TOBJ class, '%c'.\n", c );
		tclass = tobj_tclass_null;
		return 0;
	}
    }

    /*
    ** FUTURE. Many of these records only apply to certain classes.
    */
    for( k=1; k<e->nlines; ++k )
    {
	line = e->lines[k];

	if( (line[0]=='#') || (line[0]=='\0') )
	    continue;

	if( line[0] == 'b' )
	{
	    int *cnt;
	    struct tbase *tb;

	    if( tclass == tobj_tclass_pipe )
	    {
		tb = &(tinfo.pipe.bases[tinfo.pipe.basecnt]);
		cnt = &(tinfo.pipe.basecnt);
	    }
	    else if( tclass == tobj_tclass_obstacle )
	    {
		tb = &(tinfo.obstacle.bases[tinfo.obstacle.basecnt]);
		cnt = &(tinfo.obstacle.basecnt);
	    }
	    else
	    {
		tb = &(tinfo.start.pipe.bases[tinfo.start.pipe.basecnt]);
		cnt = &(tinfo.start.pipe.basecnt);
	    }

	    if( !parse_tobjbase(line,tb) )
	    {
		fprintf( stdout, "ERROR. Bad baseline, \"%s\".\n", line );
		return 0;
	    }

	    /*
	    fprintf( stdout,
		"TMP base b%02d %02d %03d | %02d %02d %02d %02d\n",
		*cnt,
		tb->flags,
		tb->baseframe,
		tb->seqs[0],
		tb->seqs[1],
		tb->seqs[2],
		tb->seqs[3] );
	    */
	    *cnt = *cnt+1;
	    continue;
	}

	if( line[0] == 'd' )
	{
	    char sDir[4];
	    tobj_dir dir;

	    if( sscanf(line,"d %03s\n", sDir) != 1 )
	    {
		fprintf( stdout, "ERROR. Bad TOBJ, line %d, \"%s\"\n",
		    k, line );
		return 0;
	    }
	    if( !parse_dir(sDir,&dir) )
	    {
		fprintf( stdout,
		    "ERROR. Bad TOBJ start direction, \"%s\".\n",
		    sDir );
		return 0;
	    }
	    tinfo.start.initial_dir = dir;
	}

	if( line[0] == 'm' )
	{
	    char str1[3];
	    char str2[3];
	    char str3[3];
	    int *flags = tinfo.meter.flags;
	    int *ids = tinfo.meter.ids;

	    if( sscanf(line,
		"m %02x %*[fs]%02s %02x %*[fs]%02s %02x %*[fs]%02s\n",
		&flags[0],str1,&flags[1],str2,&flags[2],str3) != 6 )
	    {
		fprintf( stdout, "ERROR. Bad TOBJ, line %d, \"%s\"\n",
		    k, line );
		return 0;
	    }

	    if( !parse_twodigit(str1,&(ids[0])) )
	    {
		fprintf( stdout,
		    "ERROR. Bad TOBJ, line %d, \"%s\". Bad id, \"%s\".\n",
		    k, line, str1 );
		return 0;
	    }

	    if( !parse_twodigit(str2,&(ids[1])) )
	    {
		fprintf( stdout,
		    "ERROR. Bad TOBJ, line %d, \"%s\". Bad id, \"%s\".\n",
		    k, line, str2 );
		return 0;
	    }

	    if( !parse_twodigit(str3,&(ids[2])) )
	    {
		fprintf( stdout,
		    "ERROR. Bad TOBJ, line %d, \"%s\". Bad id, \"%s\".\n",
		    k, line, str3 );
		return 0;
	    }

	    continue;
	}

	if( line[0] == 'o' )
	{
	    unsigned int flags;

	    if( sscanf(line,"o %02x\n",&flags) != 1 )
	    {
		fprintf( stdout, "ERROR. Bad TOBJ, line %d, \"%s\"\n",
		    k, line );
		return 0;
	    }

	    tinfo.obstacle.flags = flags;
	    continue;
	}

	if( line[0] == 's' )
	{
	    int *cnt;
	    struct tseq *ts;

	    if( tclass == tobj_tclass_pipe )
	    {
		ts = &(tinfo.pipe.seqs[tinfo.pipe.seqcnt]);
		cnt = &(tinfo.pipe.seqcnt);
	    }
	    else if( tclass == tobj_tclass_meter )
	    {
		ts = &(tinfo.meter.seqs[tinfo.meter.seqcnt]);
		cnt = &(tinfo.meter.seqcnt);
	    }
	    else if( tclass == tobj_tclass_obstacle )
	    {
		ts = &(tinfo.obstacle.seqs[tinfo.obstacle.seqcnt]);
		cnt = &(tinfo.obstacle.seqcnt);
	    }
	    else
	    {
		ts = &(tinfo.start.pipe.seqs[tinfo.start.pipe.seqcnt]);
		cnt = &(tinfo.start.pipe.seqcnt);
	    }
	    if( !parse_tobjseq(line,ts) )
	    {
		fprintf( stdout, "ERROR. Bad sequence line, \"%s\".\n",
		    line );
		return 0;
	    }
	    /*
	    fprintf( stdout,
		"TMP to seq %d. fl %02x-%02d. dir %d. %d frames\n",
		*cnt,
		ts->flags,
		ts->newbase,
		ts->newdir,
		ts->framecnt );
	    */
	    *cnt = *cnt+1;
	    continue;
	}
    }
    return 1;
}

int TObj::parse_series( struct tobj_info_anim *a, const char *line,
    int count )
{
    char *p;
    char nline[ANIM_MAX_SEQLINE+1];
    int i, val;

    strncpy( nline, line, ANIM_MAX_SEQLINE );
    nline[ANIM_MAX_SEQLINE] = '\0';

    /* Skip white space between key and value. */
    for( p=nline+4; isspace(*p)&&*p!='\0'; ++p );

    if( *p == '\0' )
	return 1;

    p = strtok( nline+4, "," );
    for( i=0; i<TOBJ_MAX_SEQUENCE; ++i )
    {
	val = (unsigned char)atoi(p);
	if( val > count )
	{
	    a->series[0] = -1;
	    /*
	    fprintf( stdout, "TMP non existant frame in series. %d/%d\n",
		val, count );
	    */
	    return 0;
	}
	a->series[i] = val;
	if( (p=strtok(NULL,",")) == NULL )
	    break;
    }
    a->series[i+1] = -1;

    /*
    for( i=0; a->series[i]!=-1; ++i )
	fprintf( stdout, "TMP TMP series[0] %d\n", a->series[i], i );
    */
    return 1;
}

int TObj::parse_primus( struct tobj_info_anim *a, const char *line,
    int count )
{
    char *p;
    char nline[ANIM_MAX_SEQLINE+1];
    int val;

    strncpy( nline, line, ANIM_MAX_SEQLINE );
    nline[ANIM_MAX_SEQLINE] = '\0';

    /* Skip white space between key and value. */
    for( p=nline+7; isspace(*p)&&*p!='\0'; ++p );

    if( *p == '\0' )
	return 1;

    val = (unsigned char)atoi(p);
    if( (val<0) || (val>count) )
    {
	a->primus = 0;
	/*
	fprintf( stdout, "TMP non existant frame as primus. %d/%d\n",
	    val, count );
	*/
	return 1;
    }

#if TOBJ_DOPRIMUS
    if( val > 0 )
	a->primus = val-1;
    else
    {
	a->primus = ((int)rand()%(count));
fprintf( stdout, "TMP primus is %d (\"%s\").\n", a->primus, p );
    }
#else
    a->primus = val-1;
#endif

    return 1;
}

int TObj::parse_mora( struct tobj_info_anim *a, const char *line, int count )
{
    char *p;
    char nline[ANIM_MAX_SEQLINE+1];
    int val;

    strncpy( nline, line, ANIM_MAX_SEQLINE );
    nline[ANIM_MAX_SEQLINE] = '\0';

    /* Skip white space between key and value. */
    for( p=nline+7; isspace(*p)&&*p!='\0'; ++p );

    if( *p == '\0' )
	return 1;

    val = (unsigned char)atoi(p);
    if( val < 0 )
    {
	a->mora = 0;
	//fprintf( stdout, "TMP negative delay.\n" );
	return 1;
    }

    a->mora = val;

    return 1;
}

void TObj::make_series( struct tobj_info_anim *a )
{
    a->series[0] = -1;

    if( framecnt < 1 )
	return;

    for( int i=0; i<framecnt; ++i )
	a->series[i] = i;
    a->series[framecnt] = -1;
}

#if 0 /* Debug. */
static void dump_extension( XpmAttributes *a )
{
    char **data;
    int i, j;

    fprintf( stdout, "TMP nextensions %d.\n", a->nextensions );
    fflush( stdout );
    for( i=0; i<a->nextensions; ++i )
    {
	fprintf( stdout, "TMP extension %d. lines %d.\n",
	    i, a->extensions[i].nlines );
	data = a->extensions[i].lines;
	for( j=0; j<a->extensions[i].nlines; ++j )
	{
	    fprintf( stdout, "TMP Line %02d \"%s\".\n",
		j, data[j] );
	    fflush( stdout );
	}
    }
}
#endif

int TObj::parse_twodigit( const char *string, int *val )
{
    if( string[0]=='-' && string[1]=='-' )
    {
	*val = -1;
	return 1;
    }

    if( !isdigit(string[0]) || !isdigit(string[1]) )
	return 0;
    *val = atoi( string );
    if( *val < 0 )
	return 0;
    return 1;
}

int TObj::parse_dir( const char *string, tobj_dir *dir )
{
    if( (string[0]!=string[1]) || (string[0]!=string[2]) )
    {
	//fprintf( stdout, "TMP parse_dir \"%s\"\n", string );
	return 0;
    }
    switch( string[0] )
    {
	case '<':
	    *dir = tobj_dir_left;
	    break;
	case '>':
	    *dir = tobj_dir_right;
	    break;
	case '^':
	    *dir = tobj_dir_top;
	    break;
	case 'v':
	    *dir = tobj_dir_bottom;
	    break;
	case '-':
	    *dir = tobj_dir_null;
	    break;
	default:
	    return 0;
    }
    return 1;
}

int TObj::parse_tobjbase( const char *line, struct tbase *tb )
{
    char sL[3], sR[3], sT[3], sB[3];
    int flags;
    int baseframe;
    int seqid;

    if( sscanf(line,"b %02x %*[bfs]%02d | %02s %02s %02s %02s",
	&flags,&baseframe,sL,sR,sT,sB) != 6 )
    {
	fprintf( stdout, "ERROR. Bad TOBJ base line \"%s\"\n",
	    line );
	return 0;
    }
    tb->flags = flags;
    tb->baseframe = baseframe;
    /* Parse the four directions. */
    if( !parse_twodigit(sL,&seqid) )
	return 0;
    tb->seqs[0] = seqid;
    if( !parse_twodigit(sR,&seqid) )
	return 0;
    tb->seqs[1] = seqid;
    if( !parse_twodigit(sT,&seqid) )
	return 0;
    tb->seqs[2] = seqid;
    if( !parse_twodigit(sB,&seqid) )
	return 0;
    tb->seqs[3] = seqid;
    return 1;
}

int TObj::parse_tobjseq( const char *line, struct tseq *ts )
{
    int flags = 0;
    int framecnt = 0;
    int val;
    const char *p;
    char sDir[4];
    tobj_dir dir;

    if( sscanf(line,"s %02x %*[sb]%02d %03s |",&flags,&val,sDir) != 3 )
    {
	fprintf( stdout, "ERROR. Bad TOBJ sequence line, \"%s\".\n", line );
	return 0;
    }

    if( !parse_dir(sDir,&dir) )
    {
	fprintf( stdout, "ERROR. Bad TOBJ sequence direction, \"%s\".\n",
	    sDir );
	return 0;
    }

    ts->newdir = dir;
    ts->newbase = val;
    ts->flags = flags;

    p = &line[14];
    framecnt = 0;
    while( 1 )
    {
	if( *p == '\0' )
	    break;
	if( *p == '\n' )
	    break;
	if( !isspace(*p) )
	{
	    fprintf( stdout,
		"ERROR. Bad TOBJ sequence line, \"%s\". '%c' not a space.\n",
		line, *p );
	    return 0;
	}
	++p;
	if( *p == '#' )
	{
	    break;
	}
	if( !isdigit(*p) )
	{
	    fprintf( stdout,
		"ERROR. Bad TOBJ sequence line, \"%s\". '%c' not a digit.\n",
		line, *p );
	    return 0;
	}
	sDir[0] = *p;
	++p;
	if( !isdigit(*p) )
	{
	    fprintf( stdout,
		"ERROR. Bad TOBJ sequence line, \"%s\". '%c' not a digit.\n",
		line, *p );
	    return 0;
	}
	sDir[1] = *p;
	++p;
	sDir[2] = '\0';
	val = atoi(sDir);

	if( val < 0 )
	{
	    fprintf( stdout, "ERROR. Bad frame in TOBJ sequence, \"%s\".\n",
		line );
	    return 0;
	}
	ts->frames[framecnt] = val;
	++framecnt;
    }

    if( framecnt == 0 )
    {
	fprintf( stdout, "ERROR. No frames in sequence, \"%s\".\n", line );
	return 0;
    }
    ts->framecnt = framecnt;

    return 1;
}

int TObj::new_tobjbase( int base_id, struct tobj_state_pipe *s, int *tag )
{
    struct tbase *nb;

    *tag = -1;

    // Clear out current sequence.
    s->seq_id = -1;
    s->seq_idx = -1;

    // New base.
    s->base_id = base_id;
    nb = &tinfo.pipe.bases[base_id];

    if( (nb->flags&0x01) )
    {
	// New base comes with a single frame.
	s->frame_idx = nb->baseframe;
	/*
	fprintf( stdout, "TMP to flow frame %d for new base %d\n",
	    s->frame_idx, s->base_id );
	*/
    }
    else
    {
	/*
	** New base comes with an animation sequence. Flow is over for
	** this TObj.
	*/
	struct tseq *tseq;

	s->inFlow = 0;
	s->seq_id = nb->baseframe;
	s->seq_idx = 0;

	tseq = &tinfo.pipe.seqs[s->seq_id];

	s->frame_idx = tseq->frames[0];

	if( tseq->framecnt > 1 )
	    *tag = (tseq->flags&0x0c)>>2;

	/*
	fprintf( stdout, "TMP to flow begin anim s%d for new b%d\n",
	    s->seq_id, s->base_id );
	*/
    }

    return 1;
}
