/*
** 
**  Copyright (c) 1998 by Larry Plona
** 
**  This distribution  of XPipeman is copyright by Larry Plona.
**  The  package  and  its  component  files  may  be  freely
**  distributed as long as  all copyright notices remain intact
**  and this distribution is not sold  as  commercial software.
** 
**  No warranty of any kind is attached with this software.  It
**  is supplied as is.
** 
** ----------------------------------------------------------------
*/

#ifndef TOBJ_H
#define TOBJ_H 1

/*
** C++ TObj class.
**
** Instances of this class represent an set of images and rules for
** invoking them. Each Tile object contains an object of this class.
** All images and rules are parsed from a single Xpm buffer.
**
** Each object maintains state information regarding the current
** application of rules and the current image to be displayed.
** Member functions can be invoked to advance the image.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/xpm.h>
#include <colores.h>

/*
** XPM ANIM related information.
*/
#define ANIM_MAX_SEQLINE 127
/* */

#define TOBJ_MAX_SEQUENCE 16

#define TOBJ_MAX_BASES 16
#define TOBJ_MAX_SEQS  32
#define TOBJ_MAX_PSEQ  16

struct tobj_gain
{
    unsigned
	newtile : 1, // First time for this tile.
	rating  : 3, // Rating of the tile.
	cross   : 1; // Is flow being crossed (reported only once).
};

enum tobj_tclass
{
    tobj_tclass_null = 0,
    tobj_tclass_pipe,
    tobj_tclass_obstacle,
    tobj_tclass_start,
    tobj_tclass_meter,
    tobj_tclass_anim,
    tobj_tclass_picture
};

/*
** Values must be extact, used as indices.
*/
enum tobj_dir
{
    tobj_dir_err    = -2,
    tobj_dir_null   = -1,
    tobj_dir_left   =  0,
    tobj_dir_right  =  1,
    tobj_dir_top    =  2,
    tobj_dir_bottom =  3
};

/*
** Data from ANIM extension.
*/
struct tobj_info_anim
{
    char series[16];
    int primus;
    int mora;
};

/*
** Data from TOBJ sequence record. These records appear in several
** TOBJ classes but its exact use differs.
*/
struct tseq
{
    int flags;
    int newbase;
    tobj_dir newdir;
    int frames[TOBJ_MAX_PSEQ];
    int framecnt;
};

/*
**  TOBJ base record.
*/
struct tbase
{
    int flags;
    int baseframe; // Can be frame or seq id.
    int seqs[4];   // Left, Right, Top, Bottom
};

struct tobj_info_pipe
{
    struct tbase bases[TOBJ_MAX_BASES];
    struct tseq seqs[TOBJ_MAX_SEQS];
    int basecnt;
    int seqcnt;
};

struct tobj_info_start
{
    struct tobj_info_pipe pipe;
    tobj_dir initial_dir;
};

struct tobj_info_meter
{
    int flags[3]; // Flow states 1, 2 and 3.
    int ids[3];   // Flow states 1, 2 and 3.
    struct tseq seqs[TOBJ_MAX_SEQS];
    int seqcnt;
};

struct tobj_info_obstacle
{
    int flags; // Obstacle properties.
    struct tbase bases[TOBJ_MAX_BASES];
    struct tseq seqs[TOBJ_MAX_SEQS];
    int basecnt;
    int seqcnt;
};

/*
** Data from various extensions.
*/
union tobj_tinfo
{
    struct tobj_info_pipe pipe;
    struct tobj_info_start start;
    struct tobj_info_meter meter;
    struct tobj_info_obstacle obstacle;
    struct tobj_info_anim anim;
};

/*
** The following are structures containing state information for TObjs
** used in Tiles.
*/

/* TObj containing a ANIM style animation. */
struct tobj_state_anim
{
    int frameidx; 
};

struct tobj_state_pipe
{
    int inFlow;
    int base_id;
    int frame_idx; // Index into frames.
    int seq_idx;   // Index into sequence.
    int seq_id;
};

struct tobj_state_meter
{
    int layer;
    int frame_idx; // Index into frames.
    int seq_idx;   // Index into sequence.
    int seq_id;
};

struct tobj_state_obstacle
{
    int base_id;
    int frame_idx; // Index into frames.
    int seq_idx;   // Index into sequence.
    int seq_id;    // Current sequence.
};

struct tobj_dir_idx
{
    int coff;
    int roff;
};

extern struct tobj_dir_idx tobj_dirs[4];

/* End of Tile state structures. */

class TObj
{
    XImage **frames; // Frames from XPIXMAP extensions.
    int parse_tobj( XpmExtension* );
    int parse_twodigit( const char*, int* );
    int parse_tobjbase( const char*, struct tbase* );
    int parse_tobjseq( const char*, struct tseq* );
    int parse_series( struct tobj_info_anim*, const char*, int );
    int parse_primus( struct tobj_info_anim*, const char*, int );
    int parse_mora( struct tobj_info_anim*, const char*, int );
    int parse_dir( const char*, tobj_dir* );
    int new_tobjbase( int base_id, struct tobj_state_pipe*, int* );
    void make_series( struct tobj_info_anim* );
    int load( Colores*, const char* );
public:
    int type;
    tobj_tclass tclass;
    union tobj_tinfo tinfo;
    int framecnt;    // Size of frames. Count of XPIXMAP extensions.
    TObj( int, Colores*, const char* );
    ~TObj(void);
    void *initstate( int* );
    XImage *incanim( void* );
    XImage *latest( void* );
    XImage *incflow( tobj_dir*, int*, struct tobj_state_pipe*,
       struct tobj_gain* );
    int incmeter( int*, struct tobj_state_meter*, XImage** );
    const char *name( void );
};

#endif
