/*
** 
**  Copyright (c) 1998 by Larry Plona
** 
**  This distribution  of XPipeman is copyright by Larry Plona.
**  The  package  and  its  component  files  may  be  freely
**  distributed as long as  all copyright notices remain intact
**  and this distribution is not sold  as  commercial software.
** 
**  No warranty of any kind is attached with this software.  It
**  is supplied as is.
** 
** ----------------------------------------------------------------
*/

/*
** xface.cc
**
** XPipeman specific code. Creates Xljp widget structure. Provides callbacks
** for Game object (which is hiden from Xljp and Intrinsics). Also provides
** X access to the Board, Artwork, TObj and other objects (which have some
** knowledge of X).
**
** Some routines set up callbacks which in turn, call member functions
** to power the Game object engine. Other routines are registered with
** the Game object and invoked by it.
**
** In addition to handling X, these functions also provide most of the
** communication with the Sfx (sound) library.
*/

#include "xface.h"
#include "icon.h"
#include "backing.h"

/*
** The playfield is that of the GameWindow child. It is a blank widget,
** on which the game board is implemented by application code.
*/
#define XFACE_HEIGHT_PLAYFIELD (10*40)
#define XFACE_WIDTH_PLAYFIELD ((12*40)+(3*40))

/*
** FUTURE. Figure out an interface between xface and main.cc and
** the Game object.
*/
extern const char *levelsdir;
extern const char *scorefile;
extern const char *language;
extern const char * S_DefaultScorefile;
extern const char * S_DefaultLevelsdir;
extern int flagSound;

static XtResource x11resources[] =
{
    {
	"scorefile", "Scorefile", XtRString, sizeof(String),
	(Cardinal)&scorefile, XtRString, (caddr_t)S_DefaultScorefile
    }
};

static XrmOptionDescRec x11options[] =
{
    { "-scorefile", "scorefile", XrmoptionSepArg, NULL },
    { "-levelsdir", "levelsdir", XrmoptionSepArg, NULL },
    { "-lang", "lang", XrmoptionSepArg, NULL }
};

static int mute_state = 0;
static int pause_state = 0;
static struct XljpTitulusChunk titulusScores;

static int create_popup_level( const char*, Game* );
static int create_popup_credits( const char*, Game* );
static int create_popup_scores( const char*, Game* );
static int create_popup_options( const char*, Game* );
extern "C" {
static int xfaceXlibError( Display*, XErrorEvent* );
}
static void xfaceXtError( String message );
static Pixmap getTile( const char* );
static Widget mkButton( const char*, Widget, const char* );
static XFontStruct *font_available( int );
static char *itor( int );

/* X Callbacks. */
static void unmap_cb( Widget, XtPointer, XEvent*, Boolean* );
static void popdown_cb( Widget, XtPointer, struct ljpButtonCallbackStruct* );
static void continue_cb( Widget, XtPointer, struct ljpButtonCallbackStruct* );
static void anim_cb( XtPointer, XtIntervalId* );
static void flow_cb( XtPointer, XtIntervalId* );
static void button_cb( Widget, XtPointer, XEvent*, Boolean* );
static void key_cb( Widget, XtPointer, XEvent*, Boolean* );
static void quit_cb( Widget, XtPointer, struct ljpButtonCallbackStruct* );
static void start_cb( Widget, XtPointer, struct ljpButtonCallbackStruct* );
static void credits_cb( Widget, XtPointer, struct ljpButtonCallbackStruct* );
static void scores_cb( Widget, XtPointer, struct ljpButtonCallbackStruct* );
static void pause_cb( Widget, XtPointer, struct ljpButtonCallbackStruct* );
static void mute_cb( Widget, XtPointer, struct ljpButtonCallbackStruct* );
static void options_cb( Widget, XtPointer, struct ljpButtonCallbackStruct* );
static void volume_cb( Widget, void*, struct ljpGenericCallbackStruct* );

/*
** Create the Xljp widget tree. Install callbacks which in turn will
** power the Game object.
*/
int xfaceInit( char *displayname, int *argc, char *argv[], Game *g )
{
    char *str;
    Display *dpy;
    Pixmap pixBacking;
    nl_catd cd;

    /*
    ** Open message catalog. Used mostly for button and label text. To
    ** eventually be used for error messages and dialog text.
    **
    ** At some point this should work off of LANG and other standard
    ** environment variables. Currently, the game is not fully
    ** internationalised, so we only change from "C" if the user
    ** explicitly says so on the command line.
    */
    {
	char path[1023+1];

	sprintf( path, "%s/locale/%s/LC_MESSAGES/xpipeman", 
	    levelsdir, language );
	if( (cd=catopen(path,0)) == (nl_catd)-1 )
	{
	    fprintf( stdout,
		"ERROR. Failed to open message catalog, \"%s\".\n",
		path );
	    return 0;
	}
	xface.cd = cd;
    }

    /*
    ** Open display.
    */
    if( (dpy=XOpenDisplay(displayname)) == NULL )
    {
	fprintf( stdout, "ERROR. Can not open display \"%s\".\n",
	    displayname );
	catclose( cd );
	return 0;
    }
    xface.display = dpy;

    XSetErrorHandler( xfaceXlibError );

    /*
    ** Verify visual.
    */
    {
	int rc;
	int screen_num;

	screen_num = DefaultScreen(dpy);

	while( 1 ) // Dummy loop.
	{
	    rc = XMatchVisualInfo( dpy, screen_num, 8, PseudoColor,
		&xface.vinfo );
	    if( rc )
	    {
		break;
	    }

	    rc = XMatchVisualInfo( dpy, screen_num, 16, TrueColor,
		&xface.vinfo );
	    if( rc )
	    {
		break;
	    }

	    rc = XMatchVisualInfo( dpy, screen_num, 24, TrueColor,
		&xface.vinfo );
	    if( rc )
	    {
		break;
	    }

	    rc = XMatchVisualInfo( dpy, screen_num, 32, TrueColor,
		&xface.vinfo );
	    if( rc )
	    {
		break;
	    }

	    fprintf( stdout, "ERROR. Unsupported visual.\n" );
	    return 0;
	}
    }

    /*
    ** Create colormap and then a Colores object to manage it.
    */
    if( xface.vinfo.c_class == PseudoColor )
    {
	xface.colormap = XCreateColormap( dpy,
	    DefaultRootWindow(dpy), xface.vinfo.visual, AllocAll );
	xface.colores = new Colores( dpy, xface.colormap, &xface.vinfo );
	/*
	** As a hack, the Colores class has a member to load XPipeman's
	** standard colors. We ought to supply these explicitly or load
	** them through a pixmap in the episode.
	*/
	xface.colores->loadStandard();
	xface.pxBlack = 0;
	xface.pxWhite = 1;
	xface.pxLGray = 2;
	xface.pxDGray = 3;
    }
    else
    {
	XColor color;

	xface.colormap = XCreateColormap( dpy,
	    DefaultRootWindow(dpy), xface.vinfo.visual, AllocNone );

	xface.colores = new Colores( dpy, xface.colormap, &xface.vinfo );

	color.red = 0x0000;
	color.green = 0x0000;
	color.blue = 0x0000;
	XAllocColor( dpy, xface.colormap, &color );
	xface.pxBlack = color.pixel;

	color.red = 0xffff;
	color.green = 0xffff;
	color.blue = 0xffff;
	XAllocColor( dpy, xface.colormap, &color );
	xface.pxWhite = color.pixel;

	color.red = 0xcb2b;
	color.green = 0xcb2b;
	color.blue = 0xcb2b;
	XAllocColor( dpy, xface.colormap, &color );
	xface.pxLGray = color.pixel;

	color.red = 0xab2b;
	color.green = 0xab2b;
	color.blue = 0xab2b;
	XAllocColor( dpy, xface.colormap, &color );
	xface.pxDGray = color.pixel;
    }

    /*
    ** Initialize Intrinsics.
    */
    {
	Arg args[16];

	XtToolkitInitialize();
	xface.app = XtCreateApplicationContext();
	XtDisplayInitialize( xface.app, dpy, argv[0], "XPipeman", x11options,
	    XtNumber(x11options), argc, argv );
	XtSetArg( args[0], XtNinput, True );
	XtSetArg( args[1], XtNbackground, xface.pxBlack );
	XtSetArg( args[2], XtNinitialState, ZoomState );
	if( (xface.wShell=XtAppCreateShell(argv[0],"XPipeman",
	    applicationShellWidgetClass,dpy,args,3)) == NULL )
	{
	    fputs( "ERROR. Bad XtAppCreateShell.\n", stdout );
	    catclose( cd );
	    return 0;
	}
    }

    /* We have our own Xt error handler (for no reason in particular). */
    XtAppSetErrorHandler( xface.app, xfaceXtError );

    XtVaSetValues( xface.wShell,
	XtNcolormap, xface.colormap,
	NULL );

    /*
    ** When game is iconized, popdown all Larry widgets. This really
    ** ought to be handled by the Larry widget code.
    */
    XtAddEventHandler( xface.wShell, StructureNotifyMask, 0,
	unmap_cb, NULL );

    /* Find and load fonts appropriate for various situations. */
    if( (xface.fontButtons=font_available(XFACE_FONT_BUTTONS)) == NULL )
    {
	fputs( "ERROR. No button font found.\n", stdout );
	xfaceFinis();
	return 0;
    }
    if( (xface.fontFixedText=font_available(XFACE_FONT_FIXED)) == NULL )
    {
	fputs( "ERROR. No button font found.\n", stdout );
	xfaceFinis();
	return 0;
    }
    if( (xface.fontTitle=font_available(XFACE_FONT_TITLES)) == NULL )
    {
	fputs( "ERROR. No button font found.\n", stdout );
	xfaceFinis();
	return 0;
    }
    if( (xface.fontText=font_available(XFACE_FONT_TEXT)) == NULL )
    {
	fputs( "ERROR. No button font found.\n", stdout );
	xfaceFinis();
	return 0;
    }

    /*
    ** Build the tree of Xljp widgets. The top widget is a GameWindow.
    ** A GameWindow has three children, a top menu bar, a playfield
    ** and a bottom menu bar.
    **
    ** Although no artwork has been loaded yet, we know that certain
    ** pixel values are constant. The XFACE_PIXEL_* defines represent
    ** those pixels. All of the widgets must use only those colors.
    */
    {
	int i;
	Arg args[16];
	Widget w;
	Widget wUpper, wLower;

	/*
	** Create a pixmap from harcoded data. Use it as the background
	** of the GameWindow widget. This background will occupy unused
	** area created by growing the application window (the GameWindow
	** widget does not grow its children).
	**
	** This ought to be a feature of the episode.
	*/
	if( (i=xface.colores->createPixmapFromData((const char**)backing_xpm,
	    &pixBacking,NULL,NULL)) != XpmSuccess )
	{
	    fprintf( stdout, "ERROR. Bad XpmCreatePixmapFromData. \"%s\".\n",
		XpmGetErrorString(i) );
	    catclose( cd );
	    return 0;
	}

	i = 0;
	XtSetArg( args[i], XtNforeground, xface.pxWhite ); ++i;
	XtSetArg( args[i], XtNbackground, xface.pxLGray ); ++i;
	XtSetArg( args[i], XtNcolormap, xface.colormap ); ++i;
	XtSetArg( args[i], XtNbackgroundPixmap, pixBacking ); ++i;
	XtSetArg( args[i], XljpNorientation, XljpOrientVertical ); ++i;
	XtSetArg( args[i], XljpNmarginWidth, 8 ); ++i;
	XtSetArg( args[i], XljpNmarginHeight, 8 ); ++i;
	XtSetArg( args[i], XljpNmarginTop, 8 ); ++i;
	xface.wTop = XtCreateManagedWidget(
	    "top", ljpGameWindowWidgetClass, xface.wShell, args, i );

	/*
	** Upper menu bar is within a frame. This menu bar contains
	** a few misc labels.
	*/

	i = 0;
	XtSetArg( args[i], XtNbackground, xface.pxLGray ); ++i;
	XtSetArg( args[i], XtNcolormap, xface.colormap ); ++i;
	XtSetArg( args[i], XljpNshadowType, XljpShadowRelief ); ++i;
	XtSetArg( args[i], XljpNshadowThickness, 2 ); ++i;
	XtSetArg( args[i], XljpNtopShadowColor, xface.pxWhite ); ++i;
	XtSetArg( args[i], XljpNbottomShadowColor, xface.pxBlack ); ++i;
	w = XtCreateManagedWidget(
	    "upper_menubar_frame", ljpFrameWidgetClass, xface.wTop, args, i );

	i = 0;
	XtSetArg( args[i], XtNforeground, xface.pxWhite ); ++i;
	XtSetArg( args[i], XtNbackground, xface.pxLGray ); ++i;
	XtSetArg( args[i], XtNcolormap, xface.colormap ); ++i;
	XtSetArg( args[i], XljpNorientation, XljpOrientHorizontal ); ++i;
	XtSetArg( args[i], XljpNhalignment, XljpAlignBeginning ); ++i;
	XtSetArg( args[i], XljpNspacing, 2 ); ++i;
	XtSetArg( args[i], XljpNmarginWidth, 2 ); ++i;
	XtSetArg( args[i], XljpNmarginHeight, 2 ); ++i;
	wUpper = XtCreateManagedWidget(
	    "upper_menubar", ljpRowWidgetClass, w, args, i );

	/* Game title. */
	i = 0;
	XtSetArg( args[i], XtNborderWidth, 1 ); ++i;
	XtSetArg( args[i], XtNforeground,xface.pxLGray ); ++i;
	XtSetArg( args[i], XtNbackground,xface.pxBlack ); ++i;
	XtSetArg( args[i], XtNborderColor,xface.pxWhite ); ++i;
	XtSetArg( args[i], XljpNlabel, "XPipeman3" ); ++i;
	XtSetArg( args[i], XljpNfont, xface.fontButtons ); ++i;
	XtSetArg( args[i], XljpNmarginWidth, 4 ); ++i;
	XtSetArg( args[i], XljpNmarginHeight, 4 ); ++i;
	w = XtCreateManagedWidget( "title_label",
	    ljpLabelWidgetClass, wUpper, args, i );
	/* Bug in LjpPrimitive Class initialization. */
	XtVaSetValues( w,
	    XtNborderWidth, 1,
	    XtNborderColor, xface.pxWhite,
	    NULL );

	/* Label showing remaining required tile count for level. */

	str = catgets( cd, 1, XFACE_MSG_REMAIN, NULL );
	i = 0;
	XtSetArg( args[i], XljpNlabel, str ); ++i;
	XtSetArg( args[i], XtNborderWidth, 0 ); ++i;
	XtSetArg( args[i], XljpNmarginLeft, 6 ); ++i;
	XtSetArg( args[i], XljpNmarginRight, 0 ); ++i;
	XtSetArg( args[i], XljpNmarginWidth, 0 ); ++i;
	XtSetArg( args[i], XljpNhalignment, XljpAlignEnd ); ++i;
	XtSetArg( args[i], XtNforeground, xface.pxBlack ); ++i;
	XtSetArg( args[i], XtNbackground, xface.pxLGray ); ++i;
	XtSetArg( args[i], XljpNfont, xface.fontButtons ); ++i;
	XtCreateManagedWidget( "remain_label",
	    ljpLabelWidgetClass, wUpper, args, i );

	/* Make sure font is fixed width and string is padded. */
	i = 0;
	if( strcmp(language,"la") == 0 )
	    XtSetArg( args[i], XljpNlabel,"+LXXXIII");
	else
	    XtSetArg( args[i], XljpNlabel,"+00");
	++i;
	XtSetArg( args[i], XtNborderWidth, 0 ); ++i;
	XtSetArg( args[i], XljpNmarginLeft, 0 ); ++i;
	XtSetArg( args[i], XljpNmarginWidth, 0 ); ++i;
	XtSetArg( args[i], XljpNhalignment, XljpAlignBeginning ); ++i;
	XtSetArg( args[i], XtNforeground, xface.pxBlack ); ++i;
	XtSetArg( args[i], XtNbackground, xface.pxLGray ); ++i;
	XtSetArg( args[i], XljpNfont, xface.fontFixedText ); ++i;
	xface.wRemain = XtCreateManagedWidget( "remain_value",
	    ljpLabelWidgetClass, wUpper, args, i );

	/* Label showing current level. */

	str = catgets( cd, 1, XFACE_MSG_LEVEL, NULL );

	i = 0;
	XtSetArg( args[i], XljpNlabel, str ); ++i;
	XtSetArg( args[i], XtNborderWidth,0); ++i;
	XtSetArg( args[i], XljpNmarginWidth, 0 ); ++i;
	XtSetArg( args[i], XljpNmarginLeft, 6 ); ++i;
	XtSetArg( args[i], XljpNmarginRight, 0 ); ++i;
	XtSetArg( args[i], XljpNhalignment, XljpAlignEnd ); ++i;
	XtSetArg( args[i], XtNforeground,xface.pxBlack); ++i;
	XtSetArg( args[i], XtNbackground,xface.pxLGray); ++i;
	XtSetArg( args[i], XljpNfont, xface.fontButtons ); ++i;
	XtCreateManagedWidget( "level_label",
	    ljpLabelWidgetClass, wUpper, args, i );

	/* Make sure font is fixed width and string is padded. */
	i = 0;
	if( strcmp(language,"la") == 0 )
	    XtSetArg( args[i], XljpNlabel, "XXXIII" );
	else
	    XtSetArg( args[i], XljpNlabel, "00" );
	++i;
	XtSetArg( args[i], XtNborderWidth, 0 ); ++i;
	XtSetArg( args[i], XljpNmarginWidth, 0 ); ++i;
	XtSetArg( args[i], XljpNmarginLeft, 0 ); ++i;
	XtSetArg( args[i], XljpNhalignment, XljpAlignBeginning ); ++i;
	XtSetArg( args[i], XtNforeground, xface.pxBlack ); ++i;
	XtSetArg( args[i], XtNbackground, xface.pxLGray ); ++i;
	XtSetArg( args[i], XljpNfont, xface.fontFixedText ); ++i;
	xface.wLevel = XtCreateManagedWidget( "level_value",
	    ljpLabelWidgetClass, wUpper, args, i );

	/*
	** Playfield. Use an empty widget. The application will manage
	** this area.
	*/

	i = 0;
	XtSetArg( args[i], XtNheight, XFACE_HEIGHT_PLAYFIELD ); ++i;
	XtSetArg( args[i], XtNwidth, XFACE_WIDTH_PLAYFIELD ); ++i;
	XtSetArg( args[i], XtNborderWidth, 0 ); ++i;
	XtSetArg( args[i], XtNforeground, xface.pxWhite ); ++i;
	XtSetArg( args[i], XtNbackground, xface.pxLGray ); ++i;
	xface.wPlayfield = XtCreateManagedWidget(
	    "playfield", widgetClass, xface.wTop, args, i );

	/*
	** Lower menubar contains buttons.
	*/

	i = 0;
	XtSetArg( args[i], XtNborderWidth, 0 ); ++i;
	XtSetArg( args[i], XtNforeground, xface.pxBlack ); ++i;
	XtSetArg( args[i], XtNbackground, xface.pxLGray ); ++i;
	XtSetArg( args[i], XtNcolormap, xface.colormap ); ++i;
	XtSetArg( args[i], XtNbackgroundPixmap, pixBacking ); ++i;
	XtSetArg( args[i], XljpNorientation, XljpOrientHorizontal ); ++i;
	XtSetArg( args[i], XljpNhalignment, XljpAlignBeginning ); ++i;
	XtSetArg( args[i], XljpNspacing, 4 ); ++i;
	XtSetArg( args[i], XljpNmarginWidth, 2 ); ++i;
	XtSetArg( args[i], XljpNmarginHeight, 2 ); ++i;
	wLower = XtCreateManagedWidget(
	    "lower_box", ljpRowWidgetClass, xface.wTop, args, i );

	str = catgets( cd, 1, XFACE_MSG_QUIT, NULL );
	w = mkButton( "quit_button", wLower, str );
	XtAddCallback( w, XljpNactivateCallback,
	    (XtCallbackProc)quit_cb, (void*)g );

	str = catgets( cd, 1, XFACE_MSG_NEWGAME, NULL );
	w = mkButton( "newgame_button", wLower, str );
	XtAddCallback( w, XljpNactivateCallback,
	    (XtCallbackProc)start_cb, (void*)g );

	str = catgets( cd, 1, XFACE_MSG_SCORES, NULL );
	w = mkButton( "score_button", wLower, str );
	XtAddCallback( w, XljpNactivateCallback, (XtCallbackProc)scores_cb,
	    (void*)g );

	str = catgets( cd, 1, XFACE_MSG_PAUSE, NULL );
	xface.wPause = mkButton( "pause_button", wLower, str );
	XtAddCallback( xface.wPause, XljpNactivateCallback,
	    (XtCallbackProc)pause_cb, (void*)g );

	str = catgets( cd, 1, XFACE_MSG_CREDITS, NULL );
	w = mkButton( "credits_button", wLower, str );
	XtAddCallback( w, XljpNactivateCallback,
	    (XtCallbackProc)credits_cb, (void*)g );

	str = catgets( cd, 1, XFACE_MSG_MUTE, NULL );
	xface.wMute = mkButton( "mute_button", wLower, str );
	XtAddCallback( xface.wMute, XljpNactivateCallback,
	    (XtCallbackProc)mute_cb, (void*)g );

	str = catgets( cd, 1, XFACE_MSG_OPTIONS, NULL );
	w = mkButton( "options_button", wLower, str );
	XtAddCallback( w, XljpNactivateCallback, (XtCallbackProc)options_cb,
	    (void*)g );
    }

    /* Read the app-defaults and local defaults. */
    XtGetApplicationResources( xface.wShell, 0, x11resources, 
	XtNumber(x11resources), NULL, 0 );

    /* Send colormap to X server. */
    if( xface.vinfo.c_class == PseudoColor )
	XInstallColormap( dpy, xface.colormap );

    /* Create the windows. */
    XtRealizeWidget( xface.wShell );

    /*
    ** Cause wTop to remain centered when shell is resized (particularly
    ** when maximized).
    */
#if 0
    {
	XSetWindowAttributes attr;

	attr.win_gravity = CenterGravity;
	XChangeWindowAttributes( xface.display, XtWindow(xface.wTop),
	    CWWinGravity, &attr );
    }
#endif

    xface.wPlayfield = xface.wPlayfield;
    xface.winBoard = XtWindow(xface.wPlayfield);

    /*
    ** Set icon and title. These are sent to the window manager via
    ** the ApplicationShell widget.
    */
    {
	int rc;
	Pixmap pix;

	if( (rc=xface.colores->createPixmapFromData(
	    (const char**)xpipeman_xpm,&pix,NULL,NULL)) != XpmSuccess )
	{
	    fprintf( stdout, "ERROR. Bad XpmCreatePixmapFromData. \"%s\".\n",
		XpmGetErrorString(rc) );
	    catclose( cd );
	    return 0;
	}
	XtVaSetValues( xface.wShell,
	    XtNiconPixmap, pix,
	    XtNiconName, "XPipe",
	    XtNtitle, "XPipeman",
	    NULL );
    }

    /*
    ** Graphics contexts.
    */
    {
	XGCValues gcv;
	gcv.foreground = 0;
	gcv.background = 1;
	gcv.function = GXcopy;
	xface.gc = XCreateGC(dpy, xface.winBoard, 
	    GCForeground | GCBackground | GCFunction, &gcv);
	gcv.foreground = 1;
	xface.gcClear = XCreateGC(dpy, xface.winBoard,
	    GCForeground | GCBackground | GCFunction, &gcv);
    }

    /*
    ** The scores popup displays the text using a LjpTitulus widget.
    ** The first chunk (below) is always the same.
    */
    str = catgets( cd, 1, XFACE_MSG_SCORES, NULL );
    titulusScores.alignment = XljpAlignCenter;
    titulusScores.font = xface.fontTitle->fid;
    titulusScores.format = strdup(str); /* Dynamic alloc required. */
    titulusScores.color = xface.pxBlack;
    titulusScores.next = NULL;

    /*
    XtAddEventHandler( xface.wShell, ResizeRedirectMask, 0,
	xfaceResize_cb, NULL );
    */

    /* Popups. Created on demand. */
    xface.wLCredits = NULL;
    xface.wLLevel = NULL;
    xface.wLScores = NULL;
    xface.wLOptions = NULL;
    xface.inputClosure = NULL;

    /* Volume dial starts at 50%. */
    if( flagSound )
	sfx_volume( 50 );

    return 1;
}

/*
** Called when application is iconized. Unmap all of the Larry windows
** (which ought to unmap themselves).
*/
static void unmap_cb( Widget w, XtPointer client_data,
    XEvent *event, Boolean *flag )
{
    if( event->type != UnmapNotify )
	return;
    if( xface.wLCredits != NULL )
	XtPopdown( xface.wLCredits );
    if( xface.wLLevel != NULL )
	XtPopdown( xface.wLLevel );
    if( xface.wLScores != NULL )
	XtPopdown( xface.wLScores );
}

/*
** Registered with Game object. Called to create a dialog (all them
** are implemented as Larry windows.
*/
int xfaceCreateDialog( enum game_dialog dialog, const char *title, Game *g )
{
    switch( dialog )
    {
	case game_dialog_credits:
	    create_popup_credits( title, g );
	    break;
	case game_dialog_level:
	    create_popup_level( title, g );
	    break;
	case game_dialog_options:
	    create_popup_options( title, g );
	    break;
	case game_dialog_scores:
	    create_popup_scores( title, g );
	    break;
	default:
	    return 0;
    }

    return 1;
}

/*
** create_popup_credits
**
** Create a Larry window and place XPipeman credits in it.
*/
static int create_popup_credits( const char *title, Game *g )
{
    char *str;
    int i;
    const int winWidth = 350;
    Arg args[16];
    Widget wBox, w;
    Pixmap pix;
    struct XljpTitulusChunk *c;
    struct XljpTitulusChunk default_chunk;

    if( xface.wLCredits != NULL )
	return 1;

    /*
    ** Make a pixmap from the Xpm buffer.
    */
    pix = getTile( title );
    free( (void*)title );

    i = 0;
    XtSetArg( args[i], XtNwidth, winWidth ); ++i;
    XtSetArg( args[i], XtNbackground, xface.pxLGray ); ++i;
    XtSetArg( args[i], XljpNtopShadowColor, xface.pxWhite ); ++i;
    XtSetArg( args[i], XljpNbottomShadowColor, xface.pxBlack ); ++i;
    XtSetArg( args[i], XtNcolormap, xface.colores->x_colormap ); ++i;
    XtSetArg( args[i], XtNborderWidth, 4 ); ++i;
    XtSetArg( args[i], XtNborderColor, xface.pxBlack ); ++i;
    xface.wLCredits = XtCreatePopupShell( "credits_popup",
	ljpLarryWidgetClass, xface.wShell, args, i );

    /*
    ** Shell container.
    */
    i = 0;
    XtSetArg( args[i], XljpNorientation, XljpOrientVertical ); ++i;
    XtSetArg( args[i], XljpNspacing, 2 ); ++i;
    XtSetArg( args[i], XtNforeground, xface.pxBlack ); ++i;
    XtSetArg( args[i], XtNbackground, xface.pxLGray ); ++i;
    XtSetArg( args[i], XtNborderWidth, 0 ); ++i;
    XtSetArg( args[i], XtNbackgroundPixmap, pix ); ++i;
    wBox = XtCreateManagedWidget(
	"credits_column",
	ljpRowWidgetClass,
	xface.wLCredits,
	args, i );

    /*
    ** Title.
    */
    str = catgets( xface.cd, 1, XFACE_MSG_CREDITS, NULL );
    i = 0;
    XtSetArg( args[i], XljpNlabel, str ); ++i;
    XtSetArg( args[i], XtNborderWidth, 0 ); ++i;
    XtSetArg( args[i], XtNforeground, xface.pxBlack ); ++i;
    XtSetArg( args[i], XtNbackground, xface.pxLGray ); ++i;
    XtSetArg( args[i], XljpNshadowType, XljpShadowNone ); ++i;
    XtSetArg( args[i], XljpNfont, xface.fontTitle ); ++i;
    XtSetArg( args[i], XtNbackgroundPixmap, pix ); ++i;
    XtCreateManagedWidget( "credits_title", ljpLabelWidgetClass, wBox,
	args, i );

    /*
    ** Text.
    */
    i = 0;
    XtSetArg( args[i], XtNborderWidth, 0 ); ++i;
    XtSetArg( args[i], XtNforeground, xface.pxBlack ); ++i;
    XtSetArg( args[i], XtNbackground, xface.pxLGray ); ++i;
    XtSetArg( args[i], XtNcolormap, xface.colormap ); ++i;
    XtSetArg( args[i], XljpNshadowType, XljpShadowRelief ); ++i;
    XtSetArg( args[i], XljpNshadowThickness, 2 ); ++i;
    XtSetArg( args[i], XljpNtopShadowColor, xface.pxWhite ); ++i;
    XtSetArg( args[i], XljpNbottomShadowColor, xface.pxBlack ); ++i;
    XtSetArg( args[i], XljpNmarginLeft, 5 ); ++i;
    XtSetArg( args[i], XljpNmarginRight, 5 ); ++i;
    XtSetArg( args[i], XljpNmarginTop, 5 ); ++i;
    XtSetArg( args[i], XljpNmarginBottom, 5 ); ++i;
    w = XtCreateManagedWidget( "credits_text", ljpTitulusWidgetClass, wBox,
	args, i );

    default_chunk.color = xface.pxBlack;
    default_chunk.alignment = XljpAlignBeginning;
    default_chunk.line_spacing = 2;
    default_chunk.font = xface.fontText->fid;
    str = catgets( xface.cd, 1, XFACE_MSG_AUTHORS, NULL );
    c = ljpTitulusMkChunksFromString( str, &default_chunk );
    ljpTitulusAddChunk( w, c );

    str = catgets( xface.cd, 1, XFACE_MSG_POPDOWN, NULL );
    w = mkButton( "credits_popdown", xface.wLCredits, str );
    XtAddCallback( w, XljpNactivateCallback,
	(XtCallbackProc)popdown_cb, NULL );

    return 1;
}

static int create_popup_level( const char *title, Game *g )
{
    char *str;
    const int winWidth = 250;
    int i;
    Arg args[16];
    Widget wBox;
    Widget w;
    Pixmap pix;

    if( xface.wLLevel != NULL )
	return 1;

    pix = getTile( title );
    free( (void*)title );

    i = 0;
    XtSetArg( args[i], XtNwidth, winWidth ); ++i;
    XtSetArg( args[i], XtNbackground, xface.pxLGray ); ++i;
    XtSetArg( args[i], XljpNtopShadowColor, xface.pxWhite ); ++i;
    XtSetArg( args[i], XljpNbottomShadowColor, xface.pxBlack ); ++i;
    XtSetArg( args[i], XtNcolormap, xface.colores->x_colormap ); ++i;
    XtSetArg( args[i], XtNborderWidth, 4 ); ++i;
    xface.wLLevel = XtCreatePopupShell( "level_popup", ljpLarryWidgetClass,
	xface.wShell, args, i );

    /*
    ** Shell container.
    */
    i = 0;
    XtSetArg( args[i], XljpNorientation, XljpOrientVertical ); ++i;
    XtSetArg( args[i], XljpNspacing, 2 ); ++i;
    XtSetArg( args[i], XljpNmarginHeight, 4 ); ++i;
    XtSetArg( args[i], XtNforeground, xface.pxBlack ); ++i;
    XtSetArg( args[i], XtNbackground, xface.pxLGray ); ++i;
    XtSetArg( args[i], XtNborderWidth, 0 ); ++i;
    XtSetArg( args[i], XtNbackgroundPixmap, pix ); ++i;
    XtSetArg( args[i], XtNcolormap, xface.colores->x_colormap ); ++i;
    wBox = XtCreateManagedWidget(
	"level_column",
	ljpRowWidgetClass,
	xface.wLLevel,
	args, i );

    /*
    ** Title
    */
    i = 0;
    XtSetArg( args[i], XljpNlabel, "ITER TERMINVM EST" ); ++i;
    XtSetArg( args[i], XtNbackgroundPixmap, pix ); ++i;
    XtSetArg( args[i], XtNborderWidth, 0 ); ++i;
    XtSetArg( args[i], XljpNshadowType, XljpShadowNone ); ++i;
    XtSetArg( args[i], XtNforeground, xface.pxBlack ); ++i;
    XtSetArg( args[i], XtNbackground, xface.pxLGray ); ++i;
    XtSetArg( args[i], XtNcolormap, xface.colores->x_colormap ); ++i;
    /*
    ** Because some languages require phrases for "Game Over" and
    ** "Level Over", use a smaller point size for title.
    */
    XtSetArg( args[i], XljpNfont, xface.fontButtons ); ++i;
    xface.wLevelTitle = XtCreateManagedWidget(
	"level_title",
	ljpLabelWidgetClass,
	wBox,
	args, i );

    /*
    ** Text
    */
    i = 0;
    XtSetArg( args[i], XtNborderWidth, 0 ); ++i;
    XtSetArg( args[i], XtNforeground, xface.pxWhite ); ++i;
    XtSetArg( args[i], XtNbackground, xface.pxLGray ); ++i;
    XtSetArg( args[i], XljpNshadowType, XljpShadowIn ); ++i;
    XtSetArg( args[i], XljpNshadowThickness, 2 ); ++i;
    XtSetArg( args[i], XljpNtopShadowColor, xface.pxWhite ); ++i;
    XtSetArg( args[i], XljpNbottomShadowColor, xface.pxBlack ); ++i;
    XtSetArg( args[i], XljpNmarginLeft, 5 ); ++i;
    XtSetArg( args[i], XljpNmarginRight, 5 ); ++i;
    XtSetArg( args[i], XljpNmarginTop, 5 ); ++i;
    XtSetArg( args[i], XljpNmarginBottom, 5 ); ++i;
    XtSetArg( args[i], XtNcolormap, xface.colores->x_colormap ); ++i;
    xface.wLevelText = XtCreateManagedWidget(
	"level_text",
	ljpTitulusWidgetClass,
	wBox,
	args, i );

    str = catgets( xface.cd, 1, XFACE_MSG_CONTINUE, NULL );
    w = mkButton( "level_popdown", xface.wLLevel, str );

    XtAddCallback( w, XljpNactivateCallback, (XtCallbackProc)continue_cb, g );

    return 1;
}

/*
** create_popup_scores
**
** Create a Larry window and place XPipeman scores in it.
*/
int create_popup_scores( const char *title, Game *g )
{
    char *str;
    int i;
    const int winWidth = 200;
    Arg args[24];
    Widget w;

    if( xface.wLScores != NULL )
	return 1;

    free( (void*)title );

    i = 0;
    XtSetArg( args[i], XtNwidth, winWidth ); ++i;
    XtSetArg( args[i], XtNbackground, xface.pxLGray ); ++i;
    XtSetArg( args[i], XljpNtopShadowColor, xface.pxWhite ); ++i;
    XtSetArg( args[i], XljpNbottomShadowColor, xface.pxBlack ); ++i;
    XtSetArg( args[i], XtNcolormap, xface.colores->x_colormap ); ++i;
    XtSetArg( args[i], XtNborderWidth, 0 ); ++i;
    xface.wLScores = XtCreatePopupShell( "score_popup", ljpLarryWidgetClass,
	xface.wShell, args, i );

    str = catgets( xface.cd, 1, XFACE_MSG_POPDOWN, NULL );
    w = mkButton( "scores_popdown", xface.wLScores, str );
    XtAddCallback( w, XljpNactivateCallback,
	(XtCallbackProc)popdown_cb, NULL );

    i = 0;
    XtSetArg( args[i], XtNborderWidth, 0 ); ++i;
    XtSetArg( args[i], XtNforeground, xface.pxWhite ); ++i;
    XtSetArg( args[i], XtNbackground, xface.pxLGray ); ++i;
    XtSetArg( args[i], XtNwidth, winWidth-4 ); ++i;
    XtSetArg( args[i], XtNheight, 300 ); ++i;
    XtSetArg( args[i], XljpNmarginRight, 0 ); ++i;
    XtSetArg( args[i], XljpNmarginLeft, 0 ); ++i;
    XtSetArg( args[i], XljpNshadowType, XljpShadowRelief ); ++i;
    XtSetArg( args[i], XljpNshadowThickness, 3 ); ++i;
    XtSetArg( args[i], XljpNtopShadowColor, xface.pxWhite ); ++i;
    XtSetArg( args[i], XljpNbottomShadowColor, xface.pxBlack ); ++i;
    XtSetArg( args[i], XtNcolormap, xface.colores->x_colormap ); ++i;
    xface.wScoresText = XtCreateManagedWidget(
	"scores_text",
	ljpTitulusWidgetClass,
	xface.wLScores,
	args, i );

    ljpTitulusAddChunk( xface.wScoresText, &titulusScores );

    return 1;
}

/*
** create_popup_options
**
** Create a Larry window and place XPipeman scores in it.
*/
static int create_popup_options( const char *title, Game *g )
{
    char *str;
    int i;
    const int winWidth = 200;
    Arg args[16];
    Pixmap pix;
    Widget wBox;
    Widget w;

    if( xface.wLOptions != NULL )
	return 1;

    /*
    ** Make a pixmap from the Xpm buffer.
    */
    pix = getTile( title );
    free( (void*)title );

    i = 0;
    XtSetArg( args[i], XtNwidth, winWidth+12 ); ++i;
    XtSetArg( args[i], XtNbackground, xface.pxLGray ); ++i;
    XtSetArg( args[i], XljpNtopShadowColor, xface.pxWhite ); ++i;
    XtSetArg( args[i], XljpNbottomShadowColor, xface.pxBlack ); ++i;
    XtSetArg( args[i], XtNcolormap, xface.colores->x_colormap ); ++i;
    XtSetArg( args[i], XtNborderWidth, 0 ); ++i;
    XtSetArg( args[i], XtNbackgroundPixmap, pix ); ++i;
    xface.wLOptions = XtCreatePopupShell( "options_popup", ljpLarryWidgetClass,
	xface.wShell, args, i );

    /*
    ** Frame.
    */
    i = 0;
    XtSetArg( args[i], XtNbackground, xface.pxLGray ); ++i;
    XtSetArg( args[i], XtNcolormap, xface.colormap ); ++i;
    XtSetArg( args[i], XljpNshadowType, XljpShadowRelief ); ++i;
    XtSetArg( args[i], XljpNshadowThickness, 2 ); ++i;
    XtSetArg( args[i], XtNbackground, 2 ); ++i;
    XtSetArg( args[i], XljpNtopShadowColor, 1 ); ++i;
    XtSetArg( args[i], XljpNbottomShadowColor, 0 ); ++i;
    w = XtCreateManagedWidget( "options_frame", ljpFrameWidgetClass,
	xface.wLOptions, args, i );

    /*
    ** Shell container.
    */
    i = 0;
    XtSetArg( args[i], XljpNorientation, XljpOrientVertical ); ++i;
    XtSetArg( args[i], XljpNspacing, 2 ); ++i;
    XtSetArg( args[i], XtNbackground, 2 ); ++i;
    XtSetArg( args[i], XtNbackgroundPixmap, pix ); ++i;
    wBox = XtCreateManagedWidget( "options_column", ljpRowWidgetClass,
	w, args, i );

    /*
    ** Title
    */
    str = catgets( xface.cd, 1, XFACE_MSG_OPTIONS, NULL );
    i = 0;
    XtSetArg( args[i], XljpNlabel, str ); ++i;
    XtSetArg( args[i], XtNbackgroundPixmap, pix ); ++i;
    XtSetArg( args[i], XtNborderWidth, 0 ); ++i;
    XtSetArg( args[i], XtNforeground, 0 ); ++i;
    XtSetArg( args[i], XtNbackground, 2 ); ++i;
    XtSetArg( args[i], XljpNfont, xface.fontTitle ); ++i;
    XtCreateManagedWidget(
	"options_title",
	ljpLabelWidgetClass,
	wBox,
	args, i );

    i = 0;
    XtSetArg( args[i], XtNborderWidth, 0 ); ++i;
    XtSetArg( args[i], XtNinternalWidth, 0 ); ++i;
    XtSetArg( args[i], XtNbackground, xface.pxLGray ); ++i;
    XtSetArg( args[i], XljpNdialColor, xface.pxLGray ); ++i;
    XtSetArg( args[i], XtNvalue, 50 ); ++i;
    XtSetArg( args[i], XljpNindicatorColor, xface.pxWhite ); ++i;
    XtSetArg( args[i], XljpNtopShadowColor, xface.pxWhite ); ++i;
    XtSetArg( args[i], XljpNbottomShadowColor, xface.pxBlack ); ++i;
    XtSetArg( args[i], XtNwidth, winWidth ); ++i;
    XtSetArg( args[i], XtNbackgroundPixmap, pix ); ++i;
    XtSetArg( args[i], XtNheight, 70 ); ++i;
    w = XtCreateManagedWidget(
	"volume_dial",
	ljpDialWidgetClass,
	wBox,
	args, i );
    XtAddCallback( w, XljpNactivateCallback,
	(XtCallbackProc)volume_cb,
	(XtPointer)NULL );

    str = catgets( xface.cd, 1, XFACE_MSG_VOLUME, NULL );
    i = 0;
    XtSetArg( args[i], XljpNlabel, str ); ++i;
    XtSetArg( args[i], XtNborderWidth, 0 ); ++i;
    XtSetArg( args[i], XtNforeground, xface.pxWhite ); ++i;
    XtSetArg( args[i], XtNbackground, xface.pxBlack ); ++i;
    XtSetArg( args[i], XljpNfont, xface.fontButtons ); ++i;
    XtSetArg( args[i], XljpNmarginTop, 3 ); ++i;
    XtSetArg( args[i], XljpNmarginBottom, 3 ); ++i;
    XtSetArg( args[i], XljpNmarginLeft, 3 ); ++i;
    XtSetArg( args[i], XljpNmarginRight, 3 ); ++i;
    XtCreateManagedWidget(
	"volume_label",
	ljpLabelWidgetClass,
	wBox,
	args, i );

    w = mkButton( "options_popdown", xface.wLOptions, "Popdown" );
    XtAddCallback( w, XljpNactivateCallback,
	(XtCallbackProc)popdown_cb, NULL );

    return 1;
}

void xfaceFinis( void )
{
    if( xface.wShell != NULL )
	XtDestroyWidget( xface.wShell);
    /*
    XFreeGC( xface.display, xface.gc);
    XFreeGC( xface.display, xface.gcClear );
    */
    if( xface.display != NULL )
	XCloseDisplay( xface.display );
    catclose( xface.cd );
}

/*
** Registered with Game object. Called to set a list of strings.
*/
int xfaceSetList( enum game_list list, const char *vec[] )
{
    struct XljpTitulusChunk default_chunk;
    struct XljpTitulusChunk *c;

    switch( list )
    {
	case game_list_scores:
	{
	    default_chunk.color = 0;
	    default_chunk.alignment = XljpAlignCenter;
	    default_chunk.line_spacing = 2;
	    default_chunk.font = xface.fontFixedText->fid;

	    ljpTitulusFreeChunks( titulusScores.next );
	    titulusScores.next = NULL;
	    if( vec != NULL )
	    {
		c = ljpTitulusMkChunks( vec, &default_chunk );
		ljpTitulusAddChunk( xface.wScoresText,  c );
	    }

	    return 1;
	}
    }

    return 0;
}

/*
** Registered with Board object. Called when animation timeout should
** be enabled.
*/
int xfaceSetAnim( int t, Board *b )
{
    // If a timeout is in effect, remove it.
    if( xface.i_anim )
    {
	/*
	fputs( "TMP setAnim remove timeout.\n", stdout );
	fflush( stdout );
	*/
	XtRemoveTimeOut( xface.i_anim );
	xface.i_anim = 0;
    }

    if( t > 0 )
    {
	xface.i_anim = XtAppAddTimeOut( xface.app, t, anim_cb, (void*)b );
    }

    return t;
}

/*
** Registered with Game object. Called when button and keyboard events
** should be generated for the Game object.
*/
int xfaceSetInput( int flag, Game *g )
{
    if( !flag )
    {
	if( xface.wPlayfield != NULL )
	{
	    XtRemoveEventHandler( xface.wPlayfield, ButtonPressMask, 0,
		button_cb, xface.inputClosure );
	    XtRemoveEventHandler( xface.wPlayfield, KeyPressMask, 0,
		key_cb, xface.inputClosure );
	}
	xface.inputClosure = NULL;
	return 1;
    }

    xface.inputClosure = g;

    XtAddEventHandler( xface.wPlayfield, ButtonPressMask, 0,
	button_cb, (void*)g );
    XtAddEventHandler( xface.wPlayfield, KeyPressMask, 0,
	key_cb, (void*)g );

    return 1;
}

/*
** Registered with Game object. Called when flow timeouts should be
** generated for Game object.
*/
int xfaceSetFlow( unsigned int t, Game *g )
{
    if( t == 0 )
    {
	if( xface.i_flow )
	{
	    XtRemoveTimeOut( xface.i_flow );
	    xface.i_flow = 0;
	}
	return 1;
    }

    if( t > 0 )
	xface.i_flow = XtAppAddTimeOut( xface.app, t, flow_cb, (void*)g );

    return 1;
}

/*
** Registered with Game object. Called to set numerical text.
*/
int xfaceSetLabelNumber( enum game_label label, unsigned int value )
{
    char buf[31+1];
    Widget w = NULL;

    switch( label )
    {
	case game_label_remain:
	    if( strcmp(language,"la") == 0 )
		sprintf( buf, "%-s", itor(value) );
	    else
		sprintf( buf, "%+02d", value );
	    w = xface.wRemain;
	    break;
	case game_label_level:
	    if( strcmp(language,"la") == 0 )
		sprintf( buf, "%-s", itor(value) );
	    else
		sprintf( buf, "%02d", value );
	    w = xface.wLevel;
	    break;
	default:
	    return 0;
    }

    if( w == NULL )
	return 1;

    XtVaSetValues( w, XljpNlabel, buf, NULL );

    return 1;
}

/*
** Registered with Game object. Called to set string labels.
*/
int xfaceSetLabel( enum game_label label, const char *value )
{
    char *str;
    Widget w;

    switch( label )
    {
	case game_label_levelTitle:
	    w = xface.wLevelTitle;
	    break;
	case game_label_level:
	    w = xface.wLevel;
	    break;
	case game_label_remain:
	    w = xface.wRemain;
	    break;
	default:
	    return 0;
    }

    if( !strcmp("Level Over",value) )
	str = catgets( xface.cd, 1, XFACE_MSG_LEVELOVER, NULL );
    else if( !strcmp("Game Over",value) )
	str = catgets( xface.cd, 1, XFACE_MSG_GAMEOVER, NULL );

    if( w == NULL )
	return 1;

    XtVaSetValues( w, XljpNlabel, str, NULL );

    return 1;
}

int xfaceSetCursor( const char *buf )
{
    int rc;
    Pixmap pix;
    XColor fg, bg;
    XpmAttributes attr;

    if( xface.cursor != 0 )
	XFreeCursor( xface.display, xface.cursor );
    xface.cursor = 0;

    if( buf == NULL )
    {
	XUndefineCursor( xface.display, XtWindow(xface.wPlayfield) );
	return 1;
    }

    attr.valuemask = XpmHotspot|XpmReturnPixels;
    if( (rc=xface.colores->createPixmapFromBuffer(buf,NULL,&pix,&attr))
	!= XpmSuccess )
    {
	fputs( "ERROR. Bad cursor.\n", stdout );
	fprintf( stdout,
	    "ERROR. Bad XpmCreatePixmapFromBuffer, \"%s\".\n",
	    XpmGetErrorString(rc) );
	return 0;
    }

    fg.pixel = attr.pixels[0];
    XQueryColor( xface.display, xface.colormap, &fg );
    bg.pixel = WhitePixel(xface.display,DefaultScreen(xface.display));
    XQueryColor( xface.display, DefaultColormap( xface.display,
	DefaultScreen(xface.display)), &bg );

    xface.cursor = XCreatePixmapCursor( xface.display, pix, pix,
	&fg, &bg, attr.x_hotspot, attr.y_hotspot);

    XpmFreeAttributes( &attr );
    XFreePixmap( xface.display, pix );

    XDefineCursor( xface.display, XtWindow(xface.wPlayfield), xface.cursor );

    return 1;
}

int xfaceSetStats( int level, struct score_pars *lstats,
   struct score_pars *totals )
{
    char *msg;
    char str[2047+1];
    char str2[2047+1];
    struct XljpTitulusChunk *cstats, *c;
    struct XljpTitulusChunk default_chunk;

    str[0] = '\0';

    msg = catgets(xface.cd,1,XFACE_MSG_LEVEL,"Level");
    sprintf( str2, "%s %d\n", msg, level );
    strcat( str, str2 );

    msg = catgets(xface.cd,1,XFACE_MSG_SCORE,"Score");
    sprintf( str2, "%s %d\n", msg, lstats->score );
    strcat( str, str2 );

    msg = catgets(xface.cd,1,XFACE_MSG_DISPENSED,"Dispensed");
    sprintf( str2, "%s %d\n", msg, lstats->emissa );
    strcat( str, str2 );

    msg = catgets(xface.cd,1,XFACE_MSG_PLACED,"Placed");
    sprintf( str2, "%s %d\n", msg, lstats->proposita );
    strcat( str, str2 );

    msg = catgets(xface.cd,1,XFACE_MSG_OBSCURED,"Obscura");
    sprintf( str2, "%s %d\n", msg, lstats->obscura );
    strcat( str, str2 );

    msg = catgets(xface.cd,1,XFACE_MSG_TOUCHED,"Touched");
    sprintf( str2, "%s %d\n", msg, lstats->tacta );
    strcat( str, str2 );

    msg = catgets(xface.cd,1,XFACE_MSG_CROSSED,"Crossed");
    sprintf( str2, "%s %d\n", msg, lstats->crossed );
    strcat( str, str2 );

    msg = catgets(xface.cd,1,XFACE_MSG_TOTAL,"Total");
    sprintf( str2, "%s\n", msg );
    strcat( str, str2 );

    msg = catgets(xface.cd,1,XFACE_MSG_SCORE,"Score");
    sprintf( str2, "%s %d\n", msg, totals->score );
    strcat( str, str2 );

    msg = catgets(xface.cd,1,XFACE_MSG_DISPENSED,"Dispensed");
    sprintf( str2, "%s %d\n", msg, totals->emissa );
    strcat( str, str2 );

    msg = catgets(xface.cd,1,XFACE_MSG_TOUCHED,"Touched");
    sprintf( str2, "%s %d\n", msg, totals->tacta );
    strcat( str, str2 );

    default_chunk.color = xface.pxBlack;
    default_chunk.alignment = XljpAlignBeginning;
    default_chunk.line_spacing = 2;
    default_chunk.font = xface.fontFixedText->fid;

    cstats = ljpTitulusMkChunksFromString( str, &default_chunk );

    c = ljpTitulusIndexChunk( cstats, 0 );
    c->font = xface.fontButtons->fid;
    c->alignment = XljpAlignCenter;

    c = ljpTitulusIndexChunk( cstats, 7 );
    c->font = xface.fontButtons->fid;
    c->alignment = XljpAlignCenter;

    ljpTitulusReplaceChunks( xface.wLevelText, cstats );

    return 1;
}

/*
** Registered with Game object. Called to popup or down a dialog.
*/
int xfaceDialog( enum game_dialog dialog, int popup )
{
    Widget w = NULL;

    switch( dialog )
    {
	case game_dialog_credits:
	    w = xface.wLCredits;
	    break;
	case game_dialog_level:
	    w = xface.wLLevel;
	    break;
	case game_dialog_options:
	    w = xface.wLOptions;
	    break;
	case game_dialog_scores:
	    w = xface.wLScores;
	    break;
	default:
	    return 0;
    }

    if( w == NULL )
	return 1;

    if( popup )
	XtPopup( w, XtGrabNone );
    else
	XtPopdown( w );

    return 1;
}

/*
** X11 interrupt routine for animation timeout. Invoke Board object
** member.
*/
void anim_cb( XtPointer closure, XtIntervalId *id )
{
    int t;
    Board *b = (Board*)closure;

    t = b->incanim();

    if( t > 0 )
    {
	xface.i_anim = XtAppAddTimeOut( xface.app, t, anim_cb, (void*)b );
    }
    /*
    fprintf( stdout, "TMP animCb %d 0x%x\n", t, xface.i_anim );
    fflush( stdout );
    */
}

/*
** X11 interrupt routine for mouse button. Call Game object member.
*/
static void button_cb( Widget w, XtPointer closure, XEvent *event,
    Boolean *flag )
{
    Game *g = (Game*)closure;
    XButtonEvent *ev = (XButtonEvent*)event;
    enum game_input inp;

    switch( ev->button )
    {
	case Button1:
	    inp = game_input_place;
	    break;
	case Button3:
	    inp = game_input_shift;
	    break;
	default:
	    return;
    }

    g->event( inp, ev->x, ev->y );
}

/*
** X11 interrupt routine for key events.
*/
static void key_cb( Widget w, XtPointer closure, XEvent *event,
    Boolean *flag )
{
    Game *g = (Game*)closure;
    XKeyEvent *ev = (XKeyEvent*)event;

    if( XKeycodeToKeysym(xface.display,ev->keycode,ev->state) == XK_Tab )
	g->event( game_input_fast, ev->x, ev->y );
}

/*
** Xljp button callback. Popdown the Larry widget which is the parent
** of the button.
*/
static void popdown_cb( Widget w, XtPointer client_data,
    struct ljpButtonCallbackStruct *rec )
{
    if( rec->reason != XljpCR_RELEASE )
	return;

    XtPopdown( XtParent(w) );
}

/*
** Xljp button callback. Popdown the Level dialog and invoke
** Game object member.
*/
static void continue_cb( Widget w, XtPointer closure,
    struct ljpButtonCallbackStruct *rec )
{
    Game *g = (Game*)closure;

    if( rec->reason != XljpCR_RELEASE )
	return;

    g->event( game_signal_continue );
}

/*
** X11 interrupt routine for flow meter and flow timeout.
*/
static void flow_cb( XtPointer closure, XtIntervalId *id )
{
    int t;
    Game *g = (Game*)closure;

    t = g->event( game_signal_flow );

    if( t > 0 )
	xface.i_flow = XtAppAddTimeOut( xface.app, t, flow_cb, (void*)g );
}

void quit_cb( Widget, XtPointer closure, struct ljpButtonCallbackStruct *rec )
{
    Game *g = (Game*)closure;

    if( rec->reason != XljpCR_RELEASE )
	return;

    g->event( game_signal_quit );

    if( flagSound )
	sfx_finis();

    xfaceFinis();

    exit( 0 );
}

void start_cb( Widget, XtPointer closure,
    struct ljpButtonCallbackStruct *rec )
{
    Game *g = (Game*)closure;

    if( rec->reason != XljpCR_RELEASE )
	return;
    g->event( game_signal_new );
}

void credits_cb( Widget, XtPointer closure,
    struct ljpButtonCallbackStruct *rec )
{
    Game *g = (Game*)closure;

    if( rec->reason != XljpCR_RELEASE )
	return;
    g->credits();
}

void scores_cb( Widget, XtPointer closure,
    struct ljpButtonCallbackStruct *rec )
{
    Game *g = (Game*)closure;

    if( rec->reason != XljpCR_RELEASE )
	return;
    g->scores();
}

void options_cb( Widget, XtPointer closure,
    struct ljpButtonCallbackStruct *rec )
{
    Game *g = (Game*)closure;

    if( rec->reason != XljpCR_RELEASE )
	return;
    g->options();
}

void pause_cb( Widget, XtPointer closure,
    struct ljpButtonCallbackStruct *rec )
{
    Game *g = (Game*)closure;

    if( rec->reason != XljpCR_RELEASE )
	return;

    pause_state ^= 1;
    if( mute_state )
    {
	XtVaSetValues( xface.wMute,
	    XtNforeground, xface.pxWhite,
	    XtNbackground, xface.pxBlack,
	    NULL );
    }
    g->event( game_signal_pause );
}

void mute_cb( Widget, XtPointer closure,
    struct ljpButtonCallbackStruct *rec )
{
    if( rec->reason != XljpCR_RELEASE )
	return;
    mute_state ^= 1;
    if( flagSound )
	sfx_mute( mute_state );
    if( mute_state )
    {
	XtVaSetValues( xface.wMute,
	    XtNforeground, xface.pxWhite,
	    XtNbackground, xface.pxBlack,
	    NULL );
    }
    else
    {
	XtVaSetValues( xface.wMute,
	    XtNforeground, xface.pxBlack,
	    XtNbackground, xface.pxLGray,
	    NULL );
    }
}

void volume_cb( Widget w, void *closure,
    struct ljpGenericCallbackStruct *cb )
{
    if( flagSound )
	sfx_volume( cb->value );
    //fprintf( stdout, "TMP vol %d/%d\n", cb->value, cb->value/6 );
    //fflush( stdout );
}

/*
** Xlib error callback.
**
** Format output in same manner as the rest of the code.
*/
static int xfaceXlibError( Display *dpy, XErrorEvent *e )
{
    char *use = NULL;
    char string[256];

    XGetErrorText( dpy, e->error_code, string, 255 );

    /*
    ** Request codes can be looked up in "Xproto.h".
    */

    switch( e->request_code )
    {
	case X_CreateWindow:
	    use = "Bad X_CreateWindow";
	    break;
	case X_StoreColors:
	    use = "Bad X_StoreColors";
	    break;
	case X_GetGeometry:
	    use = "Bad X_GetGeometry";
	    break;
	case X_CopyArea:
	    use = "Bad X_CopyArea";
	    break;
    }

    if( use != NULL )
    {
	fprintf( stdout, "Warning. Xlib. %s. Value %lu.\n"
	    "   %u: %s.\n",
	    use, e->resourceid,
	    e->error_code, string );
	fflush( stdout );
	return 1;
    }

    fprintf( stdout, "Warning. Xlib. %u: %s.\n"
	"   Request %u. Resource id %lu. Serial %lu. Minor %u.\n",
	e->error_code, string,
	e->request_code, e->resourceid, e->serial, e->minor_code );
    fflush( stdout );
    return 1;
}

static void xfaceXtError( String message )
{
    fprintf( stdout, "Warning. Xt. %s.\n", message );
    exit( 1 );
}

static Pixmap getTile( const char *buf )
{
    int rc;
    Pixmap pix;

    if( xface.pixTile != 0 )
	return xface.pixTile;

    if( (rc=xface.colores->createPixmapFromBuffer(buf,&pix,NULL,NULL))
	!= XpmSuccess )
    {
	fprintf( stdout, "ERROR. Bad XpmCreatePixmapFromBuffer. \"%s\".\n",
	    XpmGetErrorString(rc) );
	return 0;
    }
    free( (void*)buf );

    xface.pixTile = pix;

    return pix;
}

static Widget mkButton( const char *name, Widget wParent, const char *label )
{
    int i;
    Arg args[16];
    Widget w;

    i = 0;
    XtSetArg( args[i], XljpNlabel, label ); ++i;
    XtSetArg( args[i], XljpNarmColor, xface.pxDGray ); ++i;
    XtSetArg( args[i], XtNforeground, xface.pxBlack ); ++i;
    XtSetArg( args[i], XtNbackground, xface.pxLGray ); ++i;
    XtSetArg( args[i], XljpNtopShadowColor, xface.pxWhite ); ++i;
    XtSetArg( args[i], XljpNbottomShadowColor, xface.pxBlack ); ++i;
    XtSetArg( args[i], XljpNshadowThickness, 2 ); ++i;
    XtSetArg( args[i], XljpNshadowType, XljpShadowOut ); ++i;
    XtSetArg( args[i], XljpNfont, xface.fontButtons ); ++i;
    XtSetArg( args[i], XljpNmarginLeft, 4 ); ++i;
    XtSetArg( args[i], XljpNmarginRight, 4 ); ++i;
    XtSetArg( args[i], XljpNmarginTop, 4 ); ++i;
    XtSetArg( args[i], XljpNmarginBottom, 4 ); ++i;
    w = XtCreateManagedWidget(
	name, ljpButtonWidgetClass, wParent, args, i );

    return w;
}

static XFontStruct *font_available( int base )
{
    char *str;
    int i;
    int flagShow = 0;
    XFontStruct *fs = NULL;

    if( strcmp(language,"gr") == 0 )
	flagShow = 1;

    for( i=0; i<4; ++i )
    {
	if( (str=catgets(xface.cd,2,base+i,NULL)) == NULL )
	    break;
	if( str[0] == '\0' )
	    break;
	if( (fs=XLoadQueryFont(xface.display,str)) != NULL )
	{
	    if( flagShow )
		fprintf( stdout, "Found font: \"%s\".\n", str );
	    break;
	}
	if( flagShow )
	    fprintf( stdout, "No font: \"%s\".\n", str );
    }

    if( fs == NULL )
    {
	if( base == XFACE_FONT_FIXED )
	    str = "fixed";
	else
	    str = "variable";
	fprintf( stdout, "No font found, trying \"%s\".\n", str );
	fs = XLoadQueryFont( xface.display, str );
    }

    return fs;
}

/*
** Convert an integer into a printable string of Roman Numerals. Handles
** -3333 to 3333. Places a plus or minus sign in front. Zero is represented
** by three centered dots (ISO 8859-1 character 183).
*/
static char *itor( int v )
{
    static char buf[31+1];
    int sm, sc, sx;
    int idx, carry;

    idx = 0;

    if( v == 0 )
    {
	strcpy( buf, "\267\267\267" );
	return buf;
    }

    if( v < 0 )
    {
	buf[idx] = '-';
	v *= -1;
    }
    else
	buf[idx] = '+';
    ++idx;

    sm = v/1000;
    strncat( buf, "MMM", sm );
    idx += sm;
    v -= sm*1000;

    if( (carry=(v/500)) )
    {
	buf[idx] = 'D';
	++idx;
	v -= 500;
    }

    sc = v/100;
    if( (sc==4) && carry )   /* 900 */
    {
	buf[idx-1] = 'C';
	buf[idx+0] = 'M';
	++idx;
    }
    else if( sc == 4 ) /* 400 */
    {
	buf[idx+0] = 'C';
	buf[idx+1] = 'D';
	idx += 2;
    }
    else
    {
	buf[idx+0] = 'C';
	buf[idx+1] = 'C';
	buf[idx+2] = 'C';
	idx += sc;
    }
    v -= sc*100;

    if( (carry=(v/50)) )
    {
	buf[idx] = 'L';
	++idx;
	v -= 50;
    }

    sx = v/10;
    if( (sx==4) && carry ) /* 90 */
    {
	buf[idx-1] = 'X';
	buf[idx+0] = 'C';
	++idx;
    }
    else if( sx == 4 ) /* 40 */
    {
	buf[idx+0] = 'X';
	buf[idx+1] = 'L';
	idx += 2;
    }
    else
    {
	buf[idx+0] = 'X';
	buf[idx+1] = 'X';
	buf[idx+2] = 'X';
	idx += sx;
    }
    v -= sx*10;

    if( (carry=(v/5)) )
    {
	buf[idx] = 'V';
	++idx; /* Increment 1, 2 or 3 places. */
	v -= 5;
    }

    if( (v==4) && carry ) /* 9 */
    {
	buf[idx-1] = 'I';
	buf[idx+0] = 'X';
	++idx;
    }
    else if( v == 4 ) /* 4 */
    {
	buf[idx+0] = 'I';
	buf[idx+1] = 'V';
	idx += 2;
    }
    else
    {
	buf[idx+0] = 'I';
	buf[idx+1] = 'I';
	buf[idx+2] = 'I';
	idx += v; /* Increment 1, 2 or 3 places. */
    }

    buf[idx] = '\0';

    return buf;
}
